
using namespace std;
#include "termios.h"
#include "sys/time.h"

#include "ARToolKitPlus/TrackerSingleMarkerImpl.h"
#include <map>
#include "GestioneConfigurazione.h"
#include "MatrixFunctions.h"
#include "Matrix3DRotoTraslation.h"
#include "Param3DRotoTraslation.h"
#include "VideoAcquisitionWheelchair.h"
#include "FileRawAcquisition.h"
#include "VisionSystemCalibHeader.h"


void calcMatrixAndWriteFileComplete(int markerId,ARToolKitPlus::TrackerSingleMarker* tracker,FILE *fout){
	Matrix3DRotoTraslation *matrix=new Matrix3DRotoTraslation();
	if(markerId>=0){
		printf("found marker %d\n",markerId);
		bool allZero=true;
		for(int j=0; j<3; j++){//faccio solo le prime tre righe, la quarta è [0 0 0 1]
			for(int i=0; i<4; i++){
				matrix->setValue(j,i,tracker->getModelViewMatrix()[i*4+j]);
				if(!(i==3 && j==3)){
					if(fabs(tracker->getModelViewMatrix()[i*4+j])>1e-6){
						//printf("artoolkit::allzero=false\n");
						allZero=false;
					}
				}
			}
		}
		//se allZero=true, la matrice non è valida! (anche se markerId è >=0 capita)
		if(!allZero){
			//matrice valida
			Param3DRotoTraslation *p=toParam3DRotoTraslation(matrix);
			fprintf(fout,"%d\t%.16f\t%.16f\t%.16f\t%.16f\t%.16f\t%.16f\t%.16f\t%.16f\t%.16f\n",markerId,p->getX(),p->getY(),p->getZ(),p->getSinA(),p->getCosA(),p->getSinB(),p->getCosB(),p->getSinC(),p->getCosC());
			delete p;
		}
	}
	delete matrix;
}

int createLogFile(){
	FILE *fout=fopen(GestioneConfigurazione::getMatrixLogParameterFile().c_str(),"w");
	if(fout==NULL){
		cout<<"ERRORE APERTURA FILE "<<GestioneConfigurazione::getMatrixLogParameterFile().c_str()<<endl;
		return -1;
	}

	const int     width = 640, height = 480;
	/*size_t        numPixels = width*height*bpp;
	size_t        numBytesRead;*/
	struct timeval tv,tv2,r;
	struct timezone tz;
	bool exit=false;
	unsigned int count=0;
	char c;

	MyLogger      logger;
	ARToolKitPlus::TrackerSingleMarker* tracker = new ARToolKitPlus::TrackerSingleMarkerImpl<6,6,6, 1, 8>(width,height);

	//inizializzazione tracker
	tracker->setLogger(&logger);
	tracker->setPixelFormat(ARToolKitPlus::PIXEL_FORMAT_LUM);
	if(!tracker->init(GestioneConfigurazione::getArtoolkitCalibFile().c_str(), 1.0f, 1000.0f)){
		printf("ERROR: artoolkit init\n");
		delete tracker;
		return -1;	
	}
	// define size of the marker
	tracker->setPatternWidth(GestioneConfigurazione::getArtoolkitTagDimension());
	tracker->setBorderWidth( 0.250f);
	tracker->activateAutoThreshold(true);
	tracker->setUndistortionMode(ARToolKitPlus::UNDIST_LUT);
	tracker->setPoseEstimator(ARToolKitPlus::POSE_ESTIMATOR_RPP);
	tracker->setMarkerMode( ARToolKitPlus::MARKER_ID_SIMPLE);
	
	VideoAcquisitionWheelchair * camera;
	camera=new FileRawAcquisition(GestioneConfigurazione::getImageNamePrefix(),GestioneConfigurazione::getImageNameSubfix(),GestioneConfigurazione::getImageNameFormat());

	camera->Start();
	set_keypress();
	unsigned char * img;
	gettimeofday(&tv,&tz);
	while(!exit){
		c=getchar();
		if(c==27)exit=true;
		img=camera->CaptureFrame();
		if(img==NULL){
			printf("END!\n");
			exit =true;
		}
		if(!exit){
			ARToolKitPlus::ARMarkerInfo* markerInfo;
			int numMarkers;
			//così trova il marker migliore
			int markerId = tracker->calc(img,-1,true, &markerInfo,&numMarkers);
			calcMatrixAndWriteFileComplete(markerId,tracker,fout);

			//ora tratto gli altri marker trovati
			vector<int> tagIdentificati;
			for(int j = 0; j < numMarkers; j++){
				if(markerInfo[j].id!=-1){
					bool exist=false;
					for(int i=0;i<tagIdentificati.size();i++){
						if(tagIdentificati[i]==markerInfo[j].id){
							exist=true;
							break;
						}
					}
					if(!exist && markerInfo[j].id!=markerId){
						printf("altro id trovato: %d\n",markerInfo[j].id);
						tagIdentificati.push_back(markerInfo[j].id);
					}
				}
			}
			
			for(int i=0;i<tagIdentificati.size();i++){
				//ricerca esplicita del tag
				int markerId = tracker->calc(img,tagIdentificati[i]);
				calcMatrixAndWriteFileComplete(markerId,tracker,fout);
			}

			
						
			printf("COUNT=%d\n",count);
			count++;
		}
		
	}
	gettimeofday(&tv2,&tz);//temporizzazione del ciclo di controllo
	timersub(&tv2,&tv,&r);
	printf("tempo trascorso: %u.%.6u\n",r.tv_sec,r.tv_usec);
	double fps=((double)count)/((double)((double)r.tv_sec+(double)r.tv_usec/(double)1000000.0));
	printf("fps: %lf\n",fps);

	reset_keypress();

	fclose(fout);
	camera->Stop();
	delete tracker;
	delete camera;
	return 0;	
}

int main (int argc,char *argv[]){
	if(GestioneConfigurazione::init("config/config.ini",argc,argv)!=0)exit(0);
	cout<<"analisi immagini..."<<endl;
	if(createLogFile()==0){
		cout<<"file "<<GestioneConfigurazione::getMatrixLogParameterFile() << " creato.\n";
	}
	else{
		cout<<"errore.\n";
	}
return 0;
}
