/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Finder.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef FINDER_H
#define FINDER_H

#include <DCDT_Member.h>
#include <DCDT_Channel_IP.h>
#include <DCDT_Channel_Serial.h>
#include <DCDT_LinkRx.h>

#define STARTING_TIMER_VAL 7000000

typedef struct _DCDT_LinkTableElem {
  DCDT_LinkRx *LinkPtr;
  int status;
  int remoteID;
  bool freeSlot;
  DCDT_Mutex mtx;
} DCDT_LinkTableElem;

typedef struct _ChannelTableElem {
  bool active;
  bool dynamic;
  DCDT_Channel *channel;
} ChannelTableElem;

/*! \brief This is the system member that discover and talk with others Agora.
 *
 *  The finder is responsable of the communications links with the external world, outside the local machine. It tries to establish static communications links with external agora defined in the configuration file and waits for communications on the multicast port to find others agora on the local network.
 */
class DCDT_Finder : public DCDT_Member {
 public:
  DCDT_Finder(DCDT_Agora* agora, FinderData *fd);
  ~DCDT_Finder();
  
  void Init();

  CommData* CreateLink(int r_id, int status, CommData *l_cd, CommData *r_cd, unsigned int profile);
  inline void SetLinkStatus(int LinkID, int LinkStatus);
  inline void SetLinkRemoteAgoraID(int LinkID, int id);

  int IsPresent(int id);
  inline int GetFirstFreeLinkTableSlot();
  inline void DeleteLinkTableSlot(int LinkID);

  int ActivateLink(DCDT_LinkRx *);
  void CloseLink(int id);
  void DoYourJob(int par = 0);

 protected:

 private:
  void StaticLinks();

  FinderData *finder_data;
  DCDT_LinkTableElem LinkTable[MAXLINKS];

  DCDT_TIME 
    // When the previous finder cycle was ended
    old_ts, 
    passed;
  DCDT_TIME
    /// Finder start time
    start_ts,
    /// Finder actual time
    new_ts,
    /// When the last notify ( HS ) message was sent
    notify_ts;

  bool starting_timer;
  bool linked,
    /// Flag: true -> something was received in the current Finder cycle
    received;
  short port;
  int remote_AgoraID;
  char from[INET_ADDRSTRLEN];

  /// Informations on the actual message that the Finder is processing
  CommData
    /// Interface via which the messages was received
    *localCD,
    /// Source address informations for the received message
    *remoteCD;

  /// Header of the message that the Finder is processing
  HSMsgHeader *hs_header;
  int linknum, i;
  ChannelTableElem dynamic_channels[MAXCHANNELS];

};

inline void DCDT_Finder::SetLinkStatus(int LinkID, int LinkStatus)
{
  LinkTable[LinkID].status = LinkStatus;
};

inline void DCDT_Finder::SetLinkRemoteAgoraID(int LinkID, int id)
{
  LinkTable[LinkID].remoteID = id;
};

inline int DCDT_Finder::GetFirstFreeLinkTableSlot()
{
  for( int i = 1; i < MAXLINKS; i++) {
    LinkTable[i].mtx.lock();
    if ( LinkTable[i].freeSlot) {
      LinkTable[i].freeSlot = false;
      LinkTable[i].mtx.unlock();
      return i;
      }
    LinkTable[i].mtx.unlock();
    }
  return 0;
};

inline void DCDT_Finder::DeleteLinkTableSlot(int LinkID)
{
  LinkTable[LinkID].mtx.lock();

  LinkTable[LinkID].freeSlot = true;

  LinkTable[LinkID].mtx.unlock();

};

#endif //define
