/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_InnerChannel.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
          (C) 2003   Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_INNERCHANNEL_H
#define DCDT_INNERCHANNEL_H

#include <InnerSocket.h>
#include <DCDT_Channel.h>

#define LOST_TIMER_VAL  5000000
#define SEND_TIMER_VAL  3000000
#define RECV_TIMER_VAL  2000000
#define CONN_TIMER_VAL 10000000

class InnerCommData;

/*! \brief Channel used for local communications.
 */
class DCDT_InnerChannel : public DCDT_Channel {
 public:
  DCDT_InnerChannel(int AgoraID);
  virtual ~DCDT_InnerChannel();

  inline void Open(int link_status = 0);
  void Reopen(int link_status){link_status=0;};
  inline void StartingPrepare(InnerCommData *cd, char *to);
  inline void Prepare(CommData *local, CommData *remote);
  void ChooseChannel(int channel){channel=0;};
  inline void WaitConn();
  inline void StartConn();
  inline void Send(const DCDT_Msg *msg);
  inline void SendNotify();
  inline void SendAnswer(CommData* to, CommData* local);
  inline DCDT_Msg* Receive();
  inline void ReceiveHS(HSMsgHeader*& header, CommData*& remote);
  inline void Dispose();
  inline void Close();
  inline void KeepAlive(){};
  inline CommData* GetCommData() {return NULL;};
  inline CommData* GetStartingCD() {return NULL;};
  inline unsigned int ReadStartingProfile() {return 0;};
  inline void SetTimers();
  inline void SetLostTimer(int usec);
  inline void LostTimerOn();
  inline void LostTimerOff();
  inline void SetConnTimer(int usec);
  inline void ConnTimerOn();
  inline void ConnTimerOff();
  inline void SetSendTimer(int usec);
  inline void SendTimerOn();
  inline void SendTimerOff();
  inline void UnblockSend();
  inline void SetReceiveTimer(int usec);
  inline void ReceiveTimerOn();
  inline void ReceiveTimerOff();
  inline void UnblockReceive();
  void Restart(int link_status);
  void Stop();

 private:
  ProfileType profile;
  InnerSocket *sock;
  InnerCommData *localCD, *remoteCD;

  /// Destination file name
  char *inner_name;
};

class InnerCommData : public CommData {
public:
  char* name;

  InnerCommData();
  InnerCommData(char* n);
  InnerCommData(InnerCommData *cd);

  virtual ~InnerCommData() {
    //if ( name )
    //  delete[] name;
  }

  InnerCommData& operator=(const InnerCommData& cd) {
    // +2 because we append a character in DCDT_InnerLinkManager::SetCDs
    name = new char[strlen(cd.name)+2];
    strcpy(name, cd.name);
    return (*this);
  }

  DCDT_Channel* CreateChannel(int AgoraID) {
    return (new DCDT_InnerChannel(AgoraID));
  }
};


inline void DCDT_InnerChannel::Prepare(CommData *local, CommData *remote)
{
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Prepare InnerChannel"));

  if ( localCD )
    delete localCD;

  if (local) {
    localCD = (InnerCommData*)local;
    
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("localCD: %s", localCD->name));
  } else
    localCD = new InnerCommData();

  if ( remoteCD )
    delete remoteCD;

  if (remote) {
    remoteCD = (InnerCommData*)remote;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("remoteCD: %s", remoteCD->name));
  } else
    remoteCD = new InnerCommData();
}

inline void DCDT_InnerChannel::StartingPrepare(InnerCommData *cd, char *to)
{
  if (cd) {
    localCD = new InnerCommData(cd);

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("localCD: %s", localCD->name));
  } else
    localCD = new InnerCommData();// ???

  if (to) {
    inner_name = new char[strlen(to) + 1];
    strcpy(inner_name, to);

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("dest: %s", inner_name));
  }
}

inline void DCDT_InnerChannel::Open(int link_status)
{
  sock->Open();
  sock->Bind(localCD->name);
}

inline void DCDT_InnerChannel::WaitConn()
{
  DCDT_Msg *m;
  char *n;

  sock->SetReceiveTimer(CONN_TIMER_VAL);
  m = sock->Receive();

  if ( m != NULL ) {
    if (m->ReadType() == MT_CONNECT) {
      n = (char*)m->GetPayload();
      
      if (remoteCD) {
	if (remoteCD->name)
	  delete[] remoteCD->name;
	remoteCD->name = new char[strlen(n) + 1];
	strcpy(remoteCD->name, n);
      }
      else
	remoteCD = new InnerCommData(n);
      
      sock->Connect(remoteCD->name);
    }

    delete m;
    sock->SetReceiveTimer(RECV_TIMER_VAL);
  }

}

inline DCDT_Msg* DCDT_InnerChannel::Receive()
{
  return sock->Receive();
}

inline void DCDT_InnerChannel::ReceiveHS(HSMsgHeader*& header, CommData*& remote)
{
  char* data;
  sock->ReceiveHS(header, data);

  remote = new InnerCommData(data);
}

/*! \brief Send \a msg to the connected socket. */
inline void DCDT_InnerChannel::Send(const DCDT_Msg *msg)
{
  sock->Send(msg);
}

/*! \brief Send HS_NOTIFY to the socket identified by \a inner_name.
 *
 *  The payload contains the informations needed to talk with us, in this case what we need is only the filename associated with our socket.
 */
inline void DCDT_InnerChannel::SendNotify()
{
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Sending HS_NOTIFY"));
  /* ??? notify.AgoraID non e' inizializzato e valgrind da' un warning */
  notify.AgoraID=0;
  sock->SendHS(inner_name, &notify, localCD->name);
}

/*! \brief Send HS_ANSWER with a payload containing our socket identifier.*/
inline void DCDT_InnerChannel::SendAnswer(CommData* to, CommData* local)
{
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Sending HS_ANSWER"));
  /* ??? answer.AgoraID non e' inizializzato e valgrind da' un warning */
  answer.AgoraID=0;
  sock->SendHS(((InnerCommData*)to)->name, &answer, ((InnerCommData*)local)->name);
}

inline void DCDT_InnerChannel::StartConn()
{

  // Allocate a new message, it will be deleted at the end of the function,
  // the payload has to be dynamically allocated because il will be freed
  // when the message will be deleted
  DCDT_Msg m(MT_CONNECT, 1);
  m.AllocPayload(strlen(localCD->name)+1, localCD->name);

  sock->Connect(remoteCD->name);
  sock->Send(&m);
}

inline void DCDT_InnerChannel::Dispose()
{
  if ( localCD ) {
    if ( localCD->name )
      delete[] localCD->name;
    
    delete localCD;
  }

  if ( remoteCD ) {
    if ( remoteCD->name )
      delete[] remoteCD->name;

    delete remoteCD;
  }
}

inline void DCDT_InnerChannel::Close()
{
  sock->Close();
}

inline void DCDT_InnerChannel::SetTimers()
{
  sock->SetLostTimer(LOST_TIMER_VAL);
  sock->LostTimerOn();
  sock->SetConnTimer(CONN_TIMER_VAL);
  sock->ConnTimerOn();
  sock->SetSendTimer(SEND_TIMER_VAL);
  sock->SendTimerOn();
  sock->SetReceiveTimer(RECV_TIMER_VAL);
  sock->ReceiveTimerOn();
}

inline void DCDT_InnerChannel::SetLostTimer(int usec)
{
  sock->SetLostTimer(usec);
}

inline void DCDT_InnerChannel::LostTimerOn()
{
  sock->LostTimerOn();
}

inline void DCDT_InnerChannel::LostTimerOff()
{
  sock->LostTimerOff();
}

inline void DCDT_InnerChannel::SetConnTimer(int usec)
{
  sock->SetConnTimer(usec);
}

inline void DCDT_InnerChannel::ConnTimerOn()
{
  sock->ConnTimerOn();
}

inline void DCDT_InnerChannel::ConnTimerOff()
{
  sock->ConnTimerOff();
}

inline void DCDT_InnerChannel::SetSendTimer(int usec)
{
  sock->SetSendTimer(usec);
}

inline void DCDT_InnerChannel::SendTimerOn()
{
  sock->SendTimerOn();
}

inline void DCDT_InnerChannel::SendTimerOff()
{
  sock->SendTimerOff();
}

inline void DCDT_InnerChannel::UnblockSend()
{
  sock->UnblockSend();
}

inline void DCDT_InnerChannel::SetReceiveTimer(int usec)
{
  sock->SetReceiveTimer(usec);
}

inline void DCDT_InnerChannel::ReceiveTimerOn()
{
  sock->ReceiveTimerOn();
}

inline void DCDT_InnerChannel::ReceiveTimerOff()
{
  sock->ReceiveTimerOff();
}

inline void DCDT_InnerChannel::UnblockReceive()
{
  sock->UnblockReceive();
}

inline void DCDT_InnerChannel::Restart(int link_status)
{
  Open(link_status);
}

inline void DCDT_InnerChannel::Stop()
{
  if (sock)
    sock->ForcedClose();
}

#endif
