/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_LinkRx.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef LINKRX_H
#define LINKRX_H

#include <DCDT_Member.h>
#include <DCDT_Channel.h>
#include <DCDT_LinkTx.h>

/*! \brief Member that receive messages on a channel.
 *
 *  This member implements only channel policies for the message reception, the real reception was implements in the DCDT_Channel specified by \a channel.
 *  The main goals for this member are:
 *  - receive suscription mask from the remote Agora
 *  - receive messages from the remote Agora and forward them to the local PostOffice
 */

class DCDT_LinkRx : public DCDT_Member {
 public:
  DCDT_LinkRx(DCDT_Agora *agora, int id, int r_id);
  ~DCDT_LinkRx() {};

  CommData* Prepare(int stat, CommData *l_cd, CommData *r_cd);
  void Init();
  void DoYourJob(int par = 0);
  CommData* Restart(int restart_status, CommData *l_cd, CommData *r_cd);
  CommData* GetCommData();
  void ResetToConnecting(CommData *l_cd, CommData *r_cd);
  void ChangeStatus(int newstat) { status = newstat; };
  int ReadStatus() { return status; };
  int ReadRemoteID() { return remoteID; };
  void SetCDs(CommData *l_cd, CommData *r_cd);
  void Close();

 private:
  /*
    ( ??? )
    creation_status, restart_counter, mymask - initialized but never used
  */
  int linkID, remoteID;
  int creation_status, status, restart_counter, lost_counter;
  
  /// Channel implementation
  DCDT_Channel *channel;
  
  unsigned int mymask[NUMINT_MSGTYPE];
  CommData *localCD, *remoteCD;
  int i;

  /// Last message received
  DCDT_Msg  *msgrcv;

  /// Message subscription mask receveid from the remote Agora
  unsigned int *subs_mask_rcv;

  /// Pointer to the transmission part of this channel 
  DCDT_LinkTx* LinkTxPtr;

};


#endif //define
