/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE_H
#define DCDT_POSTOFFICE_H

#ifdef TEST_PART

typedef struct _TimeTesterElem
{
  // Time of the measure
  DCDT_TIME Data;
  
  // Type of the measure
  int Type_Op;
    
  _TimeTesterElem *prev;
}
TimeTesterElem;

#endif

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <DCDT_Agora.h>
#include <DCDT_Defs.h>
#include <DCDT_Time.h>
#include <DCDT_Msg.h>
#include <DCDT_ctrace.h>

class DCDT_Member;
class DCDT_PostOffice_SLWBU;
class DCDT_PostOffice_SLWDC;
class DCDT_PostOffice_SLWSM;

/// Structure needed to keep track of various member's subscriptions
typedef struct _SubsMsgElem
{
  DCDT_Mutex mtx;
  // Local and Remote are referred to the msg, i.e. to the its publisher

  /// Suscribers for local generated messages
  unsigned int SubscribersIfLocal[NUMINT_MEMBERS];

  /// Subscribers for remote generated messages
  unsigned int SubscribersIfRemote[NUMINT_MEMBERS];
}
SubsMsgTypeElem;

// Postoffice types
#define POSTOFFICE_SLWB   1
#define POSTOFFICE_SLWDC  2
#define POSTOFFICE_SLWSM  3

// Receive types
#define BLOCKING true
#define NOBLOCKING false

// Message types
#define MSG_ALL_TYPES 0

/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice
{
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;

 public:
  DCDT_PostOffice (DCDT_Agora * agora, int type);
  ~DCDT_PostOffice();

  void AddMember(int Mem_ID);
  void AddSysMember(int Mem_ID);
  void DeleteMemberTableEntry(int Mem_ID );
  void AddMsg (DCDT_Msg * MP, int Mem_ID, DCDT_RequestType Type);
  const DCDT_Msg * ReadNextMsg ( int MemID, bool wait );
  const DCDT_Msg * ReadNextMsg ( int MemID, int type, bool wait );
  const DCDT_Msg * ReadLastMsg ( int MemID, int type, bool wait );
  void CleanupList ();
  void SubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType=DCDT_ALL_MSG );
  void UnSubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType=DCDT_ALL_MSG);
  void UnSubscribeAll( int Member_ID );
  
  //inline void UpdateSubscribersMask (int MemID);
  bool ReadSubscribedMsgMask(unsigned int *mask, int MemID);
  bool UpdateSubs(int MemID);
  void SetActMsgDeletable (int Member_ID);
  
 private:
  void LogMsg( DCDT_Msg *msg );

  DCDT_Agora *myAgora;
  int PO_Type;
  
  DCDT_PostOffice_SLWBU *PostOffice_SLWB;
  DCDT_PostOffice_SLWDC *PostOffice_SLWDC;
  DCDT_PostOffice_SLWSM *PostOffice_SLWSM;
};

#endif
