/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_InnerSocket.h

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef INNERSOCKET_H
#define INNERSOCKET_H

#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <iostream>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <stdlib.h>
#include <DCDT_Defs.h>
#include <DCDT_Msg.h>
#include <DCDT_ctrace.h>


class InnerSocket
{
 public:
  InnerSocket()
    {
      sockfd = opened = bound = connected = tconn_flag = tlost_flag = tsend_flag = trecv_flag = 0;
      tconn_value.tv_sec = tconn_value.tv_usec = tlost_value.tv_sec = tlost_value.tv_usec = 0;
      tsend_value.tv_sec = tsend_value.tv_usec = trecv_value.tv_sec = trecv_value.tv_usec = 0;
      unblock_send_flag = unblock_recv_flag = 0;
      name = NULL;
    };
   
  ~InnerSocket()
    {
      if ( name )
	delete[] name;
    };
  
  inline void Open();
  inline void Bind(char *n);
  inline void Connect(char *n);
  inline void Send(const DCDT_Msg *msg);
  inline DCDT_Msg* Receive();
  inline void SendHS(char *to, HSMsgHeader *header, char* payload);
  inline void ReceiveHS(HSMsgHeader*& header, char*& payload);
  inline void Close();
  inline void ForcedClose();
  inline void SetLostTimer(int usec);
  inline void LostTimerOn();
  inline void LostTimerOff();
  inline void SetConnTimer(int usec);
  inline void ConnTimerOn();
  inline void ConnTimerOff();
  inline void SetSendTimer(int usec);
  inline void SendTimerOn();
  inline void SendTimerOff();
  inline void UnblockSend();
  inline void SetReceiveTimer(int usec);
  inline void ReceiveTimerOn();
  inline void ReceiveTimerOff();
  inline void UnblockReceive();

protected:
  int sockfd, opened, bound, connected, tconn_flag, tlost_flag, tsend_flag, trecv_flag;
  int unblock_send_flag, unblock_recv_flag;
  struct timeval tconn_value, tlost_value, tsend_value, trecv_value;
  char *name;
};

inline void InnerSocket::Open()
{
  if (opened)
    return;

  if ((sockfd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
    perror ( "InnerSocket::Open - Unable to open socket" );
    throw ChannelError(errno);
  }

  opened = 1;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerSocket successfully opened"));
}

inline void InnerSocket::Bind(char *n)
{
  struct sockaddr_un addr;

  if (!opened || bound || !n)
    return;

  if (name)
    delete[] name;

  name = new char[strlen(n) + 1];
  strcpy(name, n);

  memset(&addr, 0, sizeof(addr));
  strcpy(addr.sun_path, name);

  addr.sun_family = AF_UNIX;

  remove(n);
  if (bind(sockfd, (struct sockaddr *) &addr, strlen(addr.sun_path) + sizeof (addr.sun_family)) < 0) {
    printf ( "InnerSocket::Bind - Unable to bind socket to %s :%s\n", name, strerror(errno) );
    throw ChannelError(errno);
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerSocket successfully bound to: %s", name));

  bound = 1;
}

inline void InnerSocket::Connect(char *n)
{
  struct sockaddr_un addr;

  if (!opened || connected)
    return;

  memset(&addr, 0, sizeof(addr));
  strcpy(addr.sun_path, n);
  // SETTARE LA sun_len?
  addr.sun_family = AF_UNIX;
  if (connect(sockfd, (struct sockaddr *)&addr, strlen(addr.sun_path) + sizeof (addr.sun_family)) < 0) {
    printf ( "InnerSocket::Connect - Unable to connect socket to %s: %s\n", name, strerror(errno) );
    throw ConnError(errno);
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerSocket successfully connected to: %s", n));

  connected = 1;
}

inline void InnerSocket::Send(const DCDT_Msg* msg)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  msghdr mh;
  iovec iov[1];
  int nsent;

  if (!opened || !connected)
    return;

  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  if (unblock_send_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (tsend_flag)
  {
    tv.tv_sec = tsend_value.tv_sec;
    tv.tv_usec = tsend_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr))
  {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
  }

  /* Send two messages so the receiver can get the header, know what's the length of the payload, allocate it and then receive the payload */
  memset(&mh, 0, sizeof(mh));
  iov[0].iov_base = msg->GetHeader();
  iov[0].iov_len = MSGHEADER_LEN;
  mh.msg_iov = iov;
  mh.msg_iovlen = 1;

  if ((nsent = sendmsg(sockfd, &mh, 0)) < 0)
    throw ConnError(errno);
  else if (nsent != MSGHEADER_LEN ) {

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner bytes sent: %i", nsent));

    throw ConnError();
  }

  iov[0].iov_base = msg->GetPayload();
  iov[0].iov_len = msg->ReadPayloadLen();
  if ((nsent = sendmsg(sockfd, &mh, 0)) < 0)
    throw ConnError(errno);
  else if (nsent != (int)iov[0].iov_len) {

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner bytes sent: %i", nsent));

    throw ConnError();
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner bytes sent: %i message successfully sent", nsent));

}

inline void InnerSocket::SendHS(char *to, HSMsgHeader *header, char* payload)
{
  fd_set wset;
  struct timeval tv, *tv_ptr = NULL;
  struct sockaddr_un addr;
  msghdr mh;
  iovec iov[2];
  int nsent;

  if (!opened)
    return;

  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  if (unblock_send_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (tsend_flag)
  {
    tv.tv_sec = tsend_value.tv_sec;
    tv.tv_usec = tsend_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, NULL, &wset, NULL, tv_ptr))
  {
    case 0: throw TimeOut();
    case -1: throw ConnError(errno);
  }

  memset(&addr, 0, sizeof(addr));
  strcpy(addr.sun_path, to);
  addr.sun_family = AF_UNIX;

  memset(&mh, 0, sizeof(mh));
  mh.msg_name = &addr;
  mh.msg_namelen = sizeof(addr);

  iov[0].iov_base = header;
  iov[0].iov_len = HSMSGHEADER_LEN;
  iov[1].iov_base = payload;
  header->payload_len = iov[1].iov_len = strlen(payload)+1;

  mh.msg_iov = iov;
  mh.msg_iovlen = 2;

  if ((nsent = sendmsg(sockfd, &mh, 0)) < 0)
    throw ConnError(errno);
  else if (nsent < (int)(iov[0].iov_len + iov[1].iov_len)) {
    
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner bytes sent: %i", nsent));
  
    throw ConnError();
  }
  
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner bytes sent: %i message successfully sent", nsent));
}


inline DCDT_Msg* InnerSocket::Receive()
{
  DCDT_MsgHeader *msg_header;
  int plen;
  DCDT_Msg *msg;
  void *buf;
  msghdr mh;
  iovec iov[1];

  int nread = 0;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerSocket::Receive"));

  if (!opened)
    return NULL;

  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);
  if (unblock_recv_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (trecv_flag)
  {
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr)) {
    case 0: return NULL;
    case -1: throw ConnError(errno);
  }

  // preparing data structure
  memset(&mh, 0, sizeof(mh));
  msg_header = new DCDT_MsgHeader;
  iov[0].iov_base = msg_header;
  iov[0].iov_len = MSGHEADER_LEN;
  mh.msg_iov = iov;
  mh.msg_iovlen = 1;
  nread = recvmsg( sockfd, &mh, 0 );
  // Now we can create a new mutex !
  msg_header->mtx= new DCDT_Mutex;
  if ( nread < 0) {
    delete msg_header;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("recvmsg error %d", nread));
    throw ConnError(errno);
  }
  else if (nread == 0 || nread != (int)MSGHEADER_LEN) {
    delete msg_header;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("recvmsg error %d", nread));
    throw ConnError();
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner byte read: %i", nread));

  msg = new DCDT_Msg(msg_header);
  plen = msg->ReadPayloadLen();
  buf = malloc ( plen );
  memset ( buf, 0, plen );
  iov[0].iov_base = buf;
  iov[0].iov_len = plen;
  mh.msg_iov = iov;
  mh.msg_iovlen = 1;
  if ((nread = recvmsg(sockfd, &mh, 0)) != plen ) {
    free(buf);
    delete msg_header;
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("recvmsg error %d", nread));
    throw ConnError(errno);
  }

  msg->SetPayload( buf, plen );
 
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner message successfully received (%d bytes)", plen ));

  return msg;
}

inline void InnerSocket::ReceiveHS(HSMsgHeader*& header, char*& payload)
{
  HSMsgHeader *msg_header;
  char *data;
  msghdr mh;
  iovec iov[2];

  int nread = 0;
  fd_set rset;
  struct timeval tv, *tv_ptr = NULL;

  if (!opened)
    return;

  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);
  if (unblock_recv_flag)
  {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    tv_ptr = &tv;
  }
  else if (trecv_flag)
  {
    tv.tv_sec = trecv_value.tv_sec;
    tv.tv_usec = trecv_value.tv_usec;
    tv_ptr = &tv;
  }
  else if (tlost_flag)
  {
    tv.tv_sec = tlost_value.tv_sec;
    tv.tv_usec = tlost_value.tv_usec;
    tv_ptr = &tv;
  }
  switch (select(sockfd+1, &rset, NULL, NULL, tv_ptr)) {
    case 0:
      throw TimeOut();
    case -1:
      throw ConnError(errno);
  }

  // preparing data structure
  memset(&mh, 0, sizeof(mh));
  msg_header = new HSMsgHeader();
  data = new char[UNIX_PATH_MAX];
  iov[0].iov_base = msg_header;
  iov[0].iov_len = HSMSGHEADER_LEN;
  iov[1].iov_base = data;
  iov[1].iov_len = UNIX_PATH_MAX;
  mh.msg_iov = iov;
  mh.msg_iovlen = 2;
  if ((nread = recvmsg(sockfd, &mh, 0)) < 0) {
    delete[] data;
    delete msg_header;
    throw ConnError(errno);
  }
  else if (nread == 0 || nread < (int)HSMSGHEADER_LEN) {
    delete[] data;
    delete msg_header;
    throw ConnError();
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner byte read: %i", nread));

  if (nread < (int)(HSMSGHEADER_LEN + (msg_header->payload_len))) {
    delete[] data;
    delete msg_header;
    throw ConnError();
  }

  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Inner message successfully received"));

  if ( header != NULL )
    printf ( "InnerSocket.h: header != NULL\n" );

  header = msg_header;
  payload = new char[msg_header->payload_len];
  memcpy(payload, data, msg_header->payload_len);
  delete[] data;
}

/*! \brief Close the inner channel and delete the associated file 
 *  \exception ConnError
 */
inline void InnerSocket::Close()
{
  if (opened) {
    if (close(sockfd) < 0)
      throw ConnError(errno);
    connected = opened = bound = sockfd = 0;
    remove(name);

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerSocket successfully closed"));
  }
}

inline void InnerSocket::ForcedClose()
{
  if (opened) {
    shutdown(sockfd, SHUT_RDWR);
    connected = opened = bound = sockfd = 0;
    remove(name);

    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("InnerSocket successfully shut"));
  }
}


inline void InnerSocket::SetLostTimer(int usec)
{
  tlost_value.tv_usec = usec % 1000000;
  tlost_value.tv_sec = usec / 1000000;
};

inline void InnerSocket::LostTimerOn()
{
  tlost_flag = 1;
};

inline void InnerSocket::LostTimerOff()
{
  tlost_flag = 0;
};

inline void InnerSocket::SetConnTimer(int usec)
{
  tconn_value.tv_usec = usec % 1000000;
  tconn_value.tv_sec = usec / 1000000;
};

inline void InnerSocket::ConnTimerOn()
{
  tconn_flag = 1;
};

inline void InnerSocket::ConnTimerOff()
{
  tconn_flag = 0;
};

inline void InnerSocket::SetSendTimer(int usec)
{
  tsend_value.tv_usec = usec % 1000000;
  tsend_value.tv_sec = usec / 1000000;
};

inline void InnerSocket::SendTimerOn()
{
  tsend_flag = 1;
  unblock_send_flag = 0;
};

inline void InnerSocket::SendTimerOff()
{
  tsend_flag = 0;
};

inline void InnerSocket::UnblockSend()
{
  unblock_send_flag = 1;
}

inline void InnerSocket::SetReceiveTimer(int usec)
{
  trecv_value.tv_usec = usec % 1000000;
  trecv_value.tv_sec = usec / 1000000;
};

inline void InnerSocket::ReceiveTimerOn()
{
  trecv_flag = 1;
  unblock_recv_flag = 0;
};

inline void InnerSocket::ReceiveTimerOff()
{
  trecv_flag = 0;
};

inline void InnerSocket::UnblockReceive()
{
  unblock_recv_flag = 1;
}

#endif
