/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_Agora.cpp

COPYRIGHT (C) 2002   Paolo Meriggi (meriggi@ing.unibs.it)
                     Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/


#ifndef DCDT_AGORA_CPP
#define DCDT_AGORA_CPP

using namespace std;

#include <DCDT_Agora.h>
#include <DCDT_Finder.h>
#include <DCDT_PostOffice.h>
#include <DCDT_Member.h>
#include <DCDT_MsgManager.h>
#include <DCDT_InnerLinkManager.h>
#include <DCDT_ctrace.h>

#ifdef USE_BOOST
static void ChangeProcessScheduler ( void )
{

}

#else
#include <unistd.h> // nice
/*! \brief Change the process scheduler and priority based on the user request
 * 
 */
static void ChangeProcessScheduler ( void )
{
  // Change the thread scheduler and priority if nedded
  int scheduler = SCHED_OTHER;
  struct sched_param sparam;
  
  if ( getenv ( "DCDT_SCHED_PRIO" ) != NULL )
    sparam.sched_priority = atoi(getenv( "DCDT_SCHED_PRIO" ));
  else
    sparam.sched_priority = 0;
  
  if ( getenv ( "DCDT_SCHED_RR" ) != NULL ) {
    scheduler = SCHED_RR;
    sparam.sched_priority = 1;
  }

  if ( ( scheduler != SCHED_OTHER ) && ( sparam.sched_priority != 0 ) ) {
    if ( sched_setscheduler( 0, scheduler,  &sparam ) != 0 ) {
      printf ( "Error: sched_setscheduler ( S=%d, P=%d ) = %s (%d)\n", scheduler,
	       sparam.sched_priority, strerror(errno), errno );
    }
  }

  if ( ( scheduler == SCHED_OTHER ) && ( sparam.sched_priority != 0 ) ) {
    if ( nice ( sparam.sched_priority ) != 0 )
      printf ( "Error: nice ( %d ) = %s (%d)\n", sparam.sched_priority, strerror(errno), errno );
  }
}
#endif

static void InstallSignalHandler ( void )
{
   struct sigaction sa;
   __sigset_t sigset;
                                                                                
   // Ignore SIGPIPE from sockets
   sa.sa_handler = SIG_IGN;
   memset(&sigset,0,sizeof(sigset));;
   sa.sa_mask = sigset;
   sa.sa_flags = 0;
   sa.sa_restorer = NULL;
   if ( sigaction(SIGPIPE,&sa,NULL) != 0 ) {
      printf ( "Error: Unable to ignore SIGPIPE\n" );
      exit ( 1 );
   }
}



/*! \fn  Class DCDT_Agora::DCDT_Agora
 *  \brief constructor for STANDALONE ( without network ) mode
 */
DCDT_Agora::DCDT_Agora( int PO_Type ) {

  // Enable tracing
  TRC_INIT ( "agora.log", TRC_ENABLED, TRC_ON, TRC_ALL, UNIT_MAX, 0 );
  TRC_ADD_THREAD ( "Agora", 0 );
  
  // TODO: Associare signalling in order to have a smooth closing
  //  signal( SIGINT, endme);


  InstallSignalHandler();

  // Initialization of System variables
  Status = DCDT_ZERO;
  //LastMemberID = 1;
  AgoraID = 0;
  MaxIDMember = 0;

  strcpy( IPStrAddr, "");

#ifdef SEQUENCER_VERSION
  Sequencer = new DCDT_Sequencer;
#endif

  // Initialize MemberTable
  for( int i = 0; i<MAX_MEMBERS; i++) {
    DeleteMemberTableEntry(i);
   }

  ChangeProcessScheduler();
  InstallSignalHandler();


  // Create a brand new postoffice
  PostOffice = new DCDT_PostOffice( this, PO_Type ) ;

  // Setup InnerLink
  InnerLinkManager = new DCDT_InnerLinkManager( this );
  if ( InnerLinkManager->LookForExistingAgora(AgoraID) ) {
    // There is an existing agora
    FirstAgora = false;
    
    // Let's create and run the MsgManager
    MsgManager = new DCDT_MsgManager( this ) ;
    AddSysMember( MsgManager, MPM_SYSTEM );
    ActivateSysMember( MsgManager, "MsgManager" );
    
    Delay( 20000 );
    
    cout << "\nRunning in standalone secondary mode\n";
  }
  else {
    FirstAgora = true;
    Finder = NULL;

    // Let's activate and run the InnerLinkManager
    AddSysMember (InnerLinkManager, MPM_SYSTEM );
    ActivateSysMember( InnerLinkManager, "InnerLinkManager" );
    
    // Let's create and run the MsgManager
    MsgManager = new DCDT_MsgManager( this ) ;
    AddSysMember( MsgManager, MPM_SYSTEM );
    ActivateSysMember( MsgManager, "MsgManager" );

    Delay( 20000 );
    
    cout << "\nRunning in standalone primary mode\n";
  }

};

/*! \fn DCDT_Agora::DCDT_Agora(char *filename)
 *  \brief Class constructor for NETWORK mode 
 */
DCDT_Agora::DCDT_Agora( char *filename, int PO_Type ) {
  // Enable tracing
  TRC_INIT ( "agora.log", TRC_ENABLED, TRC_ON, TRC_ALL, UNIT_MAX, 0 );
  TRC_ADD_THREAD ( "Agora", 0 );
  TRC_PRINT ( DCDT_TRC_AGORA, TRC1, ("Config: %s", filename));

  // Initialization of System variables
  Status = DCDT_ZERO;
  AgoraID = 0;
  MaxIDMember = 0;

  ChangeProcessScheduler();
  InstallSignalHandler();

#ifdef SEQUENCER_VERSION
  Sequencer = new DCDT_Sequencer;
#endif
  
  // Initialize MemberTable
  for( int i = 0; i<MAX_MEMBERS; i++) {
    DeleteMemberTableEntry(i);
  }

  PostOfficeType=POSTOFFICE_SLWB;

  commflag = CommConfig(filename);
  
  if ( PO_Type != 0 )
    PO_Type=PostOfficeType;
  
  PostOffice = new DCDT_PostOffice( this, PO_Type ) ;
  
  // Setup InnerLink
  InnerLinkManager = new DCDT_InnerLinkManager( this );
  if ( InnerLinkManager->LookForExistingAgora(AgoraID) ) {
    // There is an existing agora
    FirstAgora = false;
    
    // Let's create and run the MsgManager
    MsgManager = new DCDT_MsgManager( this ) ;
    AddSysMember( MsgManager, MPM_SYSTEM );
    ActivateSysMember( MsgManager, "MsgManager" );
    
    Delay( 20000 );
    
    cout << "\nRunning in normal secondary mode\n\nPress a key\n";
  }
  else {
    FirstAgora = true;
    
    // Let's activate and run the InnerLinkManager
    AddSysMember (InnerLinkManager, MPM_SYSTEM );
    ActivateSysMember( InnerLinkManager, "InnerLink" );
    
    // Let's create and run the MsgManager
    MsgManager = new DCDT_MsgManager( this) ;
    AddSysMember( MsgManager, MPM_SYSTEM );
    ActivateSysMember( MsgManager, "MsgManager" );
    
    Delay( 20000 );
    
    // Let's start a Finder
    if (commflag) {
      Finder = new DCDT_Finder(this, finder_data);

      AddSysMember(Finder, MPM_SYSTEM );
      ActivateSysMember(Finder, "Finder" );
      
      cout << "\nRunning in normal primary mode\n\nPress a key\n";
    }
    else
      cout << "\nRunning in standalone primary mode\n\nPress a key\n";
  }
};

/** Class destructor */
DCDT_Agora::~DCDT_Agora() {
  
  // TODO: Check that all Members was previously deleted
  delete PostOffice;
  delete MsgManager;
  delete InnerLinkManager; 
  
  // If we are the frist Agor, terminate the Finder thread,
  // it will free finder_data
  // otherwise we have to free it manually
  if( FirstAgora )  {
    if ( Finder ) {
      delete Finder;
    }
  }
  else
  if ( finder_data ) {
    if ( finder_data->fp )
      fclose ( finder_data->fp );
    delete finder_data;
  }

  TRC_REMOVE_THREAD ( pthread_self() );
  TRC_END();
};

/** AddMember */
void DCDT_Agora::AddMember(DCDT_Member * MemberPtr) {
  int newMemberID;

  // Create a new entry for the DCDT_Member in the DCDT_MemberList
  newMemberID = GetFirstFreeMemberTableSlot();
  
  // Set the top ID yet assigned
  if (newMemberID > MaxIDMember )
    MaxIDMember = newMemberID;
  
  MemberPtr->Profile = 0;
  MemberPtr->myID = newMemberID;
  MemberPtr->myAgoraID = AgoraID;
  MemberPtr->myPostOffice = PostOffice;
  MemberPtr->myFinder = NULL;
  strcpy(MemberPtr->myIPAddr, IPStrAddr);// serve per testing di rapida identificazione
  
#ifdef SEQUENCER_VERSION
  MemberPtr->mySequencer = Sequencer;
#endif
 
  MemberTable[newMemberID].active = false; 
  MemberTable[newMemberID].empty = false; 
  MemberTable[newMemberID].system = false; 
  MemberTable[newMemberID].MemberPtr = MemberPtr;
  MemberTable[newMemberID].Profile = MPM_USER;
  MemberTable[newMemberID].status = EMPTYSLOT;
  PostOffice->AddMember(newMemberID);

  TRC_PRINT ( DCDT_TRC_AGORA, TRC1, ("AddMember ID=%d", newMemberID));

};

/** AddSysMember */
void DCDT_Agora::AddSysMember(DCDT_Member * MemberPtr, unsigned int profile) {
  int newMemberID;
  
  // Create a new entry for the DCDT_Member in the DCDT_MemberList
  newMemberID = GetFirstFreeMemberTableSlot();
  
  // Set the top ID yet assigned
  if (newMemberID > MaxIDMember )
    MaxIDMember = newMemberID;
  
  MemberPtr->Profile = profile;
  MemberPtr->myID = newMemberID;
  MemberPtr->myAgoraID = AgoraID;
  MemberPtr->myPostOffice = PostOffice;
  MemberPtr->myFinder = Finder;
  
#ifdef SEQUENCER_VERSION
  MemberPtr->mySequencer = Sequencer;
#endif
  
  MemberTable[newMemberID].active = false; 
  MemberTable[newMemberID].empty = false; 
  MemberTable[newMemberID].system = true; 
  MemberTable[newMemberID].MemberPtr = MemberPtr;
  MemberTable[newMemberID].Profile = profile;
  MemberTable[newMemberID].status = EMPTYSLOT;
  PostOffice->AddSysMember(newMemberID);
  
  TRC_PRINT( DCDT_TRC_AGORA, TRC1, ("AddSysMember ID=%d Profile=%x", newMemberID, MemberTable[newMemberID].Profile));
  
};


/** RemoveMember */
void DCDT_Agora::RemoveMember(DCDT_Member * MemberPtr) {
  
  // Wait that the thread ends or is cancelled
  MemberPtr->Status = TERMINATING;

#ifdef USE_BOOST
  //MemberPtr->BoostThreadRef->join();
  ((boost::thread *)MemberPtr->myBoostThd)->join();
#else
  int ret;
  ret = pthread_join( (MemberPtr->myThread), NULL );
  if ( ret!= 0 ) {
    TRC_PRINT ( DCDT_TRC_AGORA, TRC1, ("pthread_join ID=%d Error=%d", MemberPtr->myID, ret));
    printf ( "pthread_join %p ( MemberID %d ) = %d\n", MemberPtr, MemberPtr->myID, ret );
  }
#endif
  
  TRC_PRINT ( DCDT_TRC_AGORA, TRC1, ("RemoveMember ID=%d", MemberPtr->myID));
  
  // TODO: check that the member wasn't working
  Delay(20000);
  
  // Delete all the references to the deleted member from PostOffice and Agora members table
  PostOffice->DeleteMemberTableEntry(MemberPtr->myID);
  DeleteMemberTableEntry(MemberPtr->myID);

  delete(MemberPtr);
};

/** ActivateMember */
int DCDT_Agora::ActivateMember(DCDT_Member * Memb, char *name) {
  
  return Memb->Activate( name );
};

/** ActivateSysMember */
int DCDT_Agora::ActivateSysMember(DCDT_Member * Memb, char *name) {

  return Memb->Activate( name );
};


/*! \fn void DCDT_Agora::LetsWork()
 *  \brief Set agora status to DCDT_LETSWORK then wait for a Shutdown or a democratic termination.
 */
void DCDT_Agora::LetsWork() {

  Status = DCDT_LETSWORK;

#ifdef SEQUENCER_VERSION
  while( Status == DCDT_LETSWORK){
    Sequencer->CheckNextDeadline();
    pthread_testcancel();
    //    Delay(100);
  };
#else
  
  bool vote_for_termination;
  int existing_members,terminating_votes;
  DCDT_MemberTableElem *member;

  while( Status == DCDT_LETSWORK){
    vote_for_termination = true;
    terminating_votes = 0;
    existing_members = 0;

    // if all user Members set TERMINATING, lets shutdown
    for(int i = 0; i< MAX_MEMBERS; i++) {
      member = &(MemberTable[i]);
      if (  member->empty == false && 
	    member->MemberPtr ) {

	existing_members++;

	if ( member->system == false) { 
	  if ( member->MemberPtr->Status == TERMINATING ) {
	    // vote_for_termination == true -> this member wants to close
	    vote_for_termination &= true;
	    terminating_votes ++;
	  }
	  else {
	    vote_for_termination = false;
	    break;
	  }
	}
      }
    }

    /* If all user member are in TERMINATING status the agora must terminate */
    if (vote_for_termination) { 
      Status = DCDT_TERMINATING;
      printf("AGORA: Vote for TERMINATION by %d user Members over %d existing\n"
	     , terminating_votes, existing_members );
      break;
    }

    Delay( 1000000 );
     
  } // while ( Status == DCDT_LETSWORK )
 

  if ( Status == DCDT_TERMINATING ) { 
    bool all_members_terminated = false;
    
    TRC_PRINT ( DCDT_TRC_AGORA, TRC1, ("Waiting for the termination of all members"));

    while ( !all_members_terminated ){ 

      all_members_terminated = true; 
      for(int i = MAX_MEMBERS-1; i>1; i--) {
	if ( MemberTable[i].empty == false && 
	     MemberTable[i].MemberPtr) {
	  if ( MemberTable[i].active == false )
	    all_members_terminated |= true;
	  else
	    all_members_terminated = false;
	}
      }

      Status = DCDT_SHUTDOWN;
      Delay (10000);
    }
   }
   
  TRC_PRINT ( DCDT_TRC_AGORA, TRC1, ("Returning from DoYourJob"));
#endif
};

/*! \brief Read agora configuration.
 *
 * The configuration file hold all the informations about network and static links, the first line must contains the ip address of the network interface, the multicast address and the multicast port where the Agor listens for external messages.\n
 * The format of the first line is:\n
 * \<NETWORK INTERFACE IP ADDRESS\> \<MULTICAST ADDRESS\> \<MULTICAST PORT\>\n
 * Ex: 192.168.26.250 225.0.0.2 2002\n
 * The following lines, if any, specify static links to other Agorae, the format of this lines is:\n
 * \<CHANNEL TYPE\> \<"link" or "bridge"\> \<LEVEL\> \<LOCAL IP\> \<LOCAL PORT\> \<REMOTE IP\> \<REMOTE PORT\>\n
 * \<CHANNEL TYPE\> can be be: IP, SER ...\n
 * "link" = ( ??? )\n
 * "bridge" = ( ??? )\n
 * Ex:TCP gateway 1 131.175.127.61 1234 198.177.56.9 1444\n
 * NOTE: If the configuration file contains invalid settings the system starts in STANDALONE mode\n
 */
bool DCDT_Agora::CommConfig(char *filename)
{
// TODO: Check if the parameters are valid, if not write a message to the user
  FILE *fp;
  char key[80],value[256];
  int retval=false, port;

  cout << "\nReading configuration file...\n";
  if ((fp = fopen(filename, "r"))) {
    finder_data = new FinderData();
    finder_data->fp = fp;

    while ( not feof(fp) ) {
      key[0]=0;
      value[0]=0;
      fscanf( fp, "%s %[^\n]", key, value );

      if ( strcasecmp( key, "Network" ) == 0 ) {
	cout << "Network configuration:\n";
	if ( sscanf ( value, "%s%s%i", finder_data->IfStrAddr, finder_data->McStrAddr, &port ) > 0 ) {
	  cout << finder_data->IfStrAddr << " " << finder_data->McStrAddr << " " << port << "\n";
	  if ( (inet_pton(AF_INET, finder_data->IfStrAddr, &finder_data->IfNumAddr) > 0 && (inet_pton(AF_INET, finder_data->McStrAddr, &finder_data->McNumAddr) > 0) ) ) {
	    finder_data->McPort = htons((short)port);
	    cout << "AgoraID: " << (unsigned int)(AgoraID = finder_data->IfNumAddr) << "\n";
	    strcpy(IPStrAddr, finder_data->IfStrAddr);
	    retval=true;
	  }
	  else
	    cout << "Error in 'Network' inet_pton\n";
	}
	else
	  cout << "Errore in 'Network' configuration\n";
      }
      else if ( strcasecmp( key, "Link" ) == 0 ) {
	cout << "Link\n";
      }
      else if ( strcasecmp( key, "PostOffice" ) == 0 ) {
	if ( strcasecmp( value, "SLWB" ) == 0 ) {
	  PostOfficeType = POSTOFFICE_SLWB;
	}
	
	if ( strcasecmp( value, "SLWDC" ) == 0 ) {
	  PostOfficeType = POSTOFFICE_SLWDC;
	}
	
	if ( strcasecmp( value, "SLWSM" ) == 0 ) {
	  PostOfficeType = POSTOFFICE_SLWSM;
	}
      }
    }

    return retval;

    retval = fscanf(fp, "%s%s%i", finder_data->IfStrAddr, finder_data->McStrAddr, &port);
    cout << "\n" << finder_data->IfStrAddr << " " << finder_data->McStrAddr << " " << port << "\n";
    if ((retval == 0) || (retval == EOF) || (inet_pton(AF_INET, finder_data->IfStrAddr, &finder_data->IfNumAddr) <= 0) || (inet_pton(AF_INET, finder_data->McStrAddr, &finder_data->McNumAddr) <= 0)) {
      cout << "\nError while reading configuration file, running in 'standalone mode'\n\nPress a key\n";
      //getchar();
      return false;
    }
    finder_data->McPort = htons((short)port);
    cout << "\nAgoraID: " << (unsigned int)(AgoraID = finder_data->IfNumAddr) << "\n";
    strcpy(IPStrAddr, finder_data->IfStrAddr);
    return true;
  }
  else {
    cout << "\nCan't find configuration file, running in 'standalone mode'\n\nPress a key\n";
    //getchar();
    strcpy ( IPStrAddr, "" );
    return false;
  }
};

/** SetStatus */
void DCDT_Agora::SetStatus(int NewStatus) {
  Status = NewStatus;
};


inline int DCDT_Agora::CheckIfAllMemberReady() {
      bool all_mebers_ready = true;

    // If all user Members are in READY status
    for(int i = 0; i<= MaxIDMember; i++) {
      if ( MemberTable[i].system == false &&
	   MemberTable[i].empty == false && 
	   MemberTable[i].MemberPtr) { 
	if ( MemberTable[i].MemberPtr->Status == READY )
	  all_mebers_ready |= true;
	else
	  all_mebers_ready = false;
      }
    }
    
    return all_mebers_ready;
};
#endif // define
