/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE__SLWDC_H
#define DCDT_POSTOFFICE_SLWDC_H

/*! \brief Informations about the single message stored.
 *
 *  When an elemet was deleted it checks if the message was read by all the members, if this is the last member the message is deleted, otherwise the readers counter is decremented.
 */
class DCDT_MsgListElem_SLWDC
{
 public:
  DCDT_MsgListElem_SLWDC ()
    {
      ToBeDeleted = false;
      MsgPtr = NULL;
      
      for(int i = 0; i < NUMINT_MEMBERS; i++)
	{
	  readers[i] = 0;
	};
    };
  
  ~DCDT_MsgListElem_SLWDC ()
    {
      next = NULL;
      prev = NULL;
      free (next);
      free (prev);
      
      // If this is the last reader we have to delete the message.
      // Oterwise we have to decrement the readers counter
      if (MsgPtr->Get_Readers_Counter() == 1)  
	delete MsgPtr;
      else
	MsgPtr->Decrement_Readers_Counter(); 
    };
  
  struct timeval insert_time;  
  
  bool ToBeDeleted;
  
  DCDT_Mutex mtx;

  /// Flag: true -> message requested by a local user member ( member profile = MPM_USER )  
  bool local;
  
  int LocalPublisherID;
  
  DCDT_MsgListElem_SLWDC *next;
  DCDT_MsgListElem_SLWDC *prev;
  
  /// Array of bit reporting members subscribed to this message type
  unsigned int readers[NUMINT_MEMBERS];
  
  /// Type of the message
  unsigned int type_mask[NUMINT_MSGTYPE];

  /// Message
  DCDT_Msg *MsgPtr;
  
  /** Mutex used when modifing the mask */
  DCDT_Mutex mtxmask;
  
};

// Structure for the registered members
typedef struct _MemberTableElem_SLWDC
{
  DCDT_Mutex mtx;
  
  // Flag: reporting whether or not this slot is hosting a member
  bool alive;
  
  // Message types that this member is subscribed to
  unsigned int LocalSubscribedMsg[NUMINT_MSGTYPE];
  unsigned int RemoteSubscribedMsg[NUMINT_MSGTYPE];

  // Flag: reports that the susbscription mask was updated
  bool update;
  
  DCDT_MsgListElem_SLWDC *HeadBuffer;
  DCDT_MsgListElem_SLWDC *TailBuffer;
  DCDT_Condition condition_var;

  // Flag: true if a member is waiting for a message
  bool waiting;
}
MemberTableElem_SLWDC;

/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice_SLWDC
{
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;
  friend class DCDT_Member;

 public:
    
  DCDT_PostOffice_SLWDC (DCDT_Agora * agora)
    {
      myAgora = agora;
      NumOfMsgPresent = 0;

      Head = new DCDT_MsgListElem_SLWDC;
      Tail = new DCDT_MsgListElem_SLWDC;
      
      // MsgList Initialization
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      Head->ToBeDeleted=false; 
      Head->next = Tail;
      gettimeofday(&(Head->insert_time), NULL);   
      
      // MsgList Initialization
      Tail->MsgPtr = NULL;
      Tail->prev = Head;
      Tail->ToBeDeleted=false;
      Tail->next = NULL;
      gettimeofday(&(Tail->insert_time), NULL);
      
#ifdef TEST_PART
      // Init list used during the test phase
      DCDT_TIME InitTime = GetTime();
      
      for (int j = 0; j < MAX_MEMBERS; j++)
	{
	  TestElem[j].prev = NULL;
	  TestElem[j].Data = InitTime;
	}
#endif

      // MemberTable initialization
      for (int i = 0; i < MAX_MEMBERS; i++)
	{
	  MemberTable[i].alive = false;
	  MemberTable[i].update = false;
	  MemberTable[i].TailBuffer = NULL;
	  MemberTable[i].HeadBuffer = NULL;
	  MemberTable[i].waiting = false;

	  for (int j = 0; j < NUMINT_MSGTYPE; j++) {
	    MemberTable[i].LocalSubscribedMsg[j] = 0;
	    MemberTable[i].RemoteSubscribedMsg[j] = 0;
	  }
	  
	}
      
      for (int i = 0; i < MAX_MSG_TYPE; i++)
	{
	  for (int j = 0; j < NUMINT_MEMBERS; j++) 
	    {
	      SubscribedTable[i].SubscribersIfLocal[j] = 0;
	      SubscribedTable[i].SubscribersIfRemote[j] = 0;
	    }
	}
      printf ("Postoffice created\n");
    };
  
  
  ~DCDT_PostOffice_SLWDC ()
    {
      // TODO: clean of MsgList
    };
  
  inline void LogMsg( DCDT_Msg *msg ) {
    char buffer[512];
    DCDT_MsgHeader *header;
    char *payload;
    int idmy=0,idx=0,payloadlen=0;

    if ( msg != NULL ) {
      header = msg->GetHeader();

      if ( (payload=(char *)msg->GetPayload()) != NULL ) {
	payloadlen=msg->ReadPayloadLen();
	
	while ( ( idx < payloadlen ) && ( idmy < (int)sizeof(buffer)-7 ) ){
	  if ( isprint( payload[idx] ) )
	    idmy += sprintf ( &buffer[idmy], "%c", payload[idx] );
	  else
	    idmy += sprintf ( &buffer[idmy], "<0x%.2hhX>", 
			      (unsigned char)payload[idx] );
	  idx++;
	}
      }
    else
      sprintf ( buffer, "NULL" );
    
      TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <%s>", msg->ReadType(), buffer) );
    }
    else
      TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <NULL>", msg->ReadType() ) );
  };

  inline void AddMember(int Mem_ID);
  inline void AddSysMember(int Mem_ID);
  
  inline void DeleteMemberTableEntry(int Mem_ID );
  
  inline void AddMsg (DCDT_Msg * MP, int Mem_ID);
  
  inline DCDT_Msg *ReadNextMsg (int MemID); // TODO: Check if return "const DCDT_MSG *" is better
  inline DCDT_Msg *ReadNextMsg (int MemID, int type );
  inline DCDT_Msg *ReadLastMsg (int MemID, int type );

  inline void CleanupList ();
  
  inline void SubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType=DCDT_ALL_MSG );
  inline void UnSubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType=DCDT_ALL_MSG );
  inline void UnSubscribeAll( int Member_ID );
  
  inline void UpdateSubscribersMask (int MemID);
  inline bool ReadSubscribedMsgMask(unsigned int *mask, int MemID);
  inline bool UpdateSubs(int MemID);
    
 private:
  
  inline void SetReaders (DCDT_MsgListElem_SLWDC * Elem, int num_type); 
  inline void RemoveBufferElement(int MemID, DCDT_MsgListElem_SLWDC *elem);
  DCDT_MsgListElem_SLWDC *Head, *Tail;

  // Pointer to the list's last element.
  // If we set this pointer at the beginning of a clear cycle the cleaner can be stopped before touching elements inserted during the cycle.
  DCDT_MsgListElem_SLWDC *LastInsert;

  
  
  inline int MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem_SLWDC *);
  inline int MatchReadersMask (int Member_ID, DCDT_MsgListElem_SLWDC *);
  
  inline void SetType (DCDT_MsgListElem_SLWDC * Elem, int type);

  // Functions used during the test phase
#ifdef TEST_PART
  inline void InsertListTime (int MemID, int TypeOp);
  inline void VisualList ();
  inline void SaveList (char* nomefile);
#else
#define InsertListTime(A,B)
#define VisualList()
#define SaveList(A)
#endif
    
  inline bool CheckTimeoutElem(DCDT_MsgListElem_SLWDC *); 
  inline int GetTypeFromMask (DCDT_MsgListElem_SLWDC * Elem);
  
  /*! \brief For each message type we have the corresponding member mask.
   *
   *  A member mask use a sequence of bit to repesent all the members, if a bit is set to 1 the corresponding member is subscribed to this type of messages.
   *  Ex: 00010001 -> Members 1 and 5 are subscribed to this message type.
   *  The member mask is composed of two parts local and remote, the local one was used when a message was locally originated or was received from a BRIDGE member; the remote one was used when the message was received from a STATIC or LINK member.
   */
  SubsMsgTypeElem SubscribedTable[MAX_MSG_TYPE];
  
  /// Mutex to access SubscribedMsgMask
  DCDT_Mutex subscribed_mtx;

  /// Message types requested by local members used by LinkTx members
  unsigned int SubscribedMsgMask[NUMINT_MSGTYPE];
  
  // Mutex to access UpdateMask_flag and MemberTable[i].update
  DCDT_Mutex UpdateMask_mtx;

  /// Flag: true when SubscribedMsgMask was updated
  bool UpdateMask_flag;

  MemberTableElem_SLWDC MemberTable[MAX_MEMBERS];
  
  int NumOfMsgPresent;
  DCDT_Agora *myAgora;  
};

inline void DCDT_PostOffice_SLWDC::AddMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice_SLWDC::AddSysMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice_SLWDC::DeleteMemberTableEntry(int Mem_ID){
  
  // TODO
  MemberTable[Mem_ID].alive = false;
  
};

/*! \brief Add a message to the PostOffice.
 *
 *  Add the new message to the queue owned by the dispatcher, in the next dispatch cycle the message will be sent to each member subscribed; each member will receive a copy of the message in his local queue.
 */ 
inline void DCDT_PostOffice_SLWDC::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;
  
  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ( "MemID: %d", Mem_ID));
  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( MP ); );

  InsertListTime(Mem_ID,WRITEINIT);
  DCDT_MsgListElem_SLWDC *newMsgListElem = new (DCDT_MsgListElem_SLWDC);
  MP->Increment_Readers_Counter();
  newMsgListElem->MsgPtr = MP;
  
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  
  
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);
  
  // Lock the dispatcher's queue and insert the new message
  InsertListTime(Mem_ID,WRITELOCKINIT);
  Tail->mtx.lock();
  InsertListTime(Mem_ID,WRITELOCKEND);

  // Add the message to the dispatcher's queue, just before the last element
  newMsgListElem->next = Tail; 
  newMsgListElem->prev = Tail->prev;
  Tail->prev->next = newMsgListElem;
  Tail->prev = newMsgListElem;

  LastInsert = newMsgListElem; 
 
  Tail->mtx.unlock(); 
  
  NumOfMsgPresent++;
  
  InsertListTime(Mem_ID,WRITEEND);

  TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Messages in queue: %d", NumOfMsgPresent) );
  TRC_VOID_RETURN ( DCDT_TRC_POSTOFFICE, TRC1 );
};

/*! \brief Delete read messages then read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If the queue is empty we have to wait until we receive a signal from our dispatcher; the signal means that we have a new message in our queue. Before returning the message we have to check if the current message available has to be deleted or his timeout was experied.
 *  \return NULL - The only message available was deleted so no message is available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice_SLWDC::ReadNextMsg (int MemID )
{
  DCDT_MsgListElem_SLWDC *Elem;
  
  TRC_ENTER( DCDT_TRC_POSTOFFICE, TRC1, ("MemID=%d", MemID));
  
  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock();
  InsertListTime(MemID,READLOCKEND);
  Elem = MemberTable[MemID].HeadBuffer;
  
  while ( 1 ) {
    // If there isn't a new message wait
    if ( Elem == NULL ) {
      MemberTable[MemID].mtx.unlock();
      
      // Wait a signal from our dispatcher
      InsertListTime(MemID,READLOCKINIT);
      MemberTable[MemID].condition_var.lock(); 
      InsertListTime(MemID,READLOCKEND); 
      InsertListTime(MemID,WAITINIT);
      MemberTable[MemID].waiting = true;
      MemberTable[MemID].condition_var.wait();
      InsertListTime(MemID,WAITEND);
      MemberTable[MemID].condition_var.unlock();
      
      MemberTable[MemID].mtx.lock();
      InsertListTime(MemID,READLOCKEND);
    }
    
    // If the queued message has to be deleted delete it ....
    if (Elem->ToBeDeleted || CheckTimeoutElem(Elem) ) {
      DCDT_MsgListElem_SLWDC *next=Elem->next;
      
      InsertListTime(MemID,DELETEINIT); 
      RemoveBufferElement( MemID, Elem );
      InsertListTime(MemID,DELETEEND);
      Elem=next;
    }
    else
      break;
  }

  Elem->ToBeDeleted = true;
  MemberTable[MemID].mtx.unlock();
  InsertListTime ( MemID, READDONEEND );

  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( Elem->MsgPtr ); );
  TRC_RETURN( DCDT_TRC_POSTOFFICE, TRC1, ("ReadNextMsg type"), Elem->MsgPtr );

  return (Elem->MsgPtr);
  
  /*  
  DCDT_MsgListElem_SLWDC *Elem;
  TRC_ENTER( DCDT_TRC_POSTOFFICE, TRC1, ("MemID=%d", MemID));

  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].condition_var.lock(); 
  InsertListTime(MemID,READLOCKEND); 

  // If we don't have a message wait a signal from our dispatcher
  while ( MemberTable[MemID].HeadBuffer == NULL ) {
    InsertListTime(MemID,WAITINIT);
    MemberTable[MemID].waiting = true;
    MemberTable[MemID].condition_var.wait();
    InsertListTime(MemID,WAITEND);
  }

  MemberTable[MemID].condition_var.unlock();
  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock();
  InsertListTime(MemID,READLOCKEND);
  Elem = MemberTable[MemID].HeadBuffer;
 
  if (Elem->ToBeDeleted || CheckTimeoutElem(Elem) ) {

    // If the first queued message has to be deleted delete it ....
    InsertListTime( MemID, DELETEINIT ); 
    RemoveBufferElement( MemID, Elem );
    InsertListTime( MemID, DELETEEND );

    // .. then read the next message
    // If we don't have a message wait a signal from our dispatcher
    while ( MemberTable[MemID].HeadBuffer == NULL ) {
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READDONEEND);
      InsertListTime(MemID,READLOCKINIT);
      MemberTable[MemID].condition_var.lock();
      InsertListTime(MemID,READLOCKEND);
      InsertListTime(MemID,WAITINIT);
      MemberTable[MemID].condition_var.wait();
      InsertListTime(MemID,WAITEND);
      MemberTable[MemID].mtx.lock();
      InsertListTime(MemID,READLOCKEND);
      MemberTable[MemID].condition_var.unlock();
    }

    Elem = MemberTable[MemID].HeadBuffer;
    if (Elem) {	      
      Elem->ToBeDeleted = true;
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READDONEEND);

      TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( Elem->MsgPtr ); );
      TRC_RETURN( DCDT_TRC_POSTOFFICE, TRC1, ("Delete & ReadNextMsg"), Elem->MsgPtr );
      return (Elem->MsgPtr);
    }
    
    // The only message available was deleted
    MemberTable[MemID].mtx.unlock();
    InsertListTime(MemID,READEND);

    TRC_RETURN( DCDT_TRC_POSTOFFICE, TRC1, ("NULL"), NULL );
    return ((DCDT_Msg *) NULL);
  }
  
  Elem->ToBeDeleted = true;
  MemberTable[MemID].mtx.unlock(); 
  InsertListTime(MemID,READDONEEND);

  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( Elem->MsgPtr ); );
  TRC_RETURN( DCDT_TRC_POSTOFFICE, TRC1, ("ReadNextMsg"), Elem->MsgPtr );
  return (Elem->MsgPtr);
  */
};

/*! \brief Delete read messages then read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If the queue is empty we have to wait until we receive a signal from our dispatcher; the signal means that we have a new message in our queue. Before returning the message we have to check if the current message available has to be deleted or his timeout was experied.
 *  \return NULL - The only message available was deleted so no message is available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice_SLWDC::ReadNextMsg (int MemID, int type)
{  
  DCDT_MsgListElem_SLWDC *Elem;

  TRC_ENTER( DCDT_TRC_POSTOFFICE, TRC1, ("MemID=%d", MemID));

  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock();
  InsertListTime(MemID,READLOCKEND);
  Elem = MemberTable[MemID].HeadBuffer;
 
  while ( 1 ) {
    // If there isn't a new message wait
    if ( Elem == NULL ) {
      MemberTable[MemID].mtx.unlock();
      
      // Wait a signal from our dispatcher
      InsertListTime(MemID,READLOCKINIT);
      MemberTable[MemID].condition_var.lock(); 
      InsertListTime(MemID,READLOCKEND); 
      InsertListTime(MemID,WAITINIT);
      MemberTable[MemID].waiting = true;
      MemberTable[MemID].condition_var.wait();
      InsertListTime(MemID,WAITEND);
      MemberTable[MemID].condition_var.unlock();
      
      MemberTable[MemID].mtx.lock();
      InsertListTime(MemID,READLOCKEND);
    }

    // If the queued message has to be deleted delete it ....
    if (Elem->ToBeDeleted || CheckTimeoutElem(Elem) ) {
      DCDT_MsgListElem_SLWDC *next=Elem->next;

      InsertListTime(MemID,DELETEINIT); 
      RemoveBufferElement( MemID, Elem );
      InsertListTime(MemID,DELETEEND);
      Elem=next;
    }
    else {
      if ( Elem->MsgPtr->ReadType() != type )
	Elem=Elem->next;
      else
	break;
    }
  }

  Elem->ToBeDeleted = true;
  MemberTable[MemID].mtx.unlock();
  InsertListTime ( MemID, READDONEEND );

  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( Elem->MsgPtr ); );
  TRC_RETURN( DCDT_TRC_POSTOFFICE, TRC1, ("ReadNextMsg type"), Elem->MsgPtr );

  return (Elem->MsgPtr);
  
}


inline DCDT_Msg *DCDT_PostOffice_SLWDC::ReadLastMsg (int MemID, int type)
{
  DCDT_MsgListElem_SLWDC *Elem, *found=NULL;

  TRC_ENTER( DCDT_TRC_POSTOFFICE, TRC1, ("MemID=%d", MemID));

  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock();
  InsertListTime(MemID,READLOCKEND);
  Elem = MemberTable[MemID].TailBuffer;
 
  while ( 1 ) {
    // If there isn't a new message wait
    if ( Elem == NULL ) {
      if ( found == NULL ) {
	MemberTable[MemID].mtx.unlock();
	
	// Wait a signal from our dispatcher
	InsertListTime(MemID,READLOCKINIT);
	MemberTable[MemID].condition_var.lock(); 
	InsertListTime(MemID,READLOCKEND); 
	InsertListTime(MemID,WAITINIT);
	MemberTable[MemID].waiting = true;
	MemberTable[MemID].condition_var.wait();
	InsertListTime(MemID,WAITEND);
	MemberTable[MemID].condition_var.unlock();
	
	MemberTable[MemID].mtx.lock();
	InsertListTime(MemID,READLOCKEND);
      }
      else
	break;
    }

    // If the queued message has to be deleted delete it ....
    if (Elem->ToBeDeleted || CheckTimeoutElem(Elem) ) {
      DCDT_MsgListElem_SLWDC *next=Elem->next;

      InsertListTime(MemID,DELETEINIT); 
      RemoveBufferElement( MemID, Elem );
      InsertListTime(MemID,DELETEEND);
      Elem=next;
    }
    else {
      if ( Elem->MsgPtr->ReadType() == type ) {
	if ( found == NULL )
	  found = Elem;
	else {
	  // Delete all the old messages
	  Elem->ToBeDeleted=true;
	  continue;
	}
      }
      Elem=Elem->next;
    }
  }

  found->ToBeDeleted = true;
  MemberTable[MemID].mtx.unlock();
  InsertListTime ( MemID, READDONEEND );

  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( found->MsgPtr ); );
  TRC_RETURN( DCDT_TRC_POSTOFFICE, TRC1, ("ReadNextMsg type"), found->MsgPtr );

  return (found->MsgPtr);
}

/*| \brief Delete the element pointed by \aelem from \aMemId's local queue 
 *
 *  NB The message is deleted when needed with the destructor of ElemDeletable
 *  \param elem - Element to be deleted
 */
inline void DCDT_PostOffice_SLWDC::RemoveBufferElement(int MemID, DCDT_MsgListElem_SLWDC *elem)    
{
  if ( elem != NULL ) {
    // Update the previous element
    if ( elem->prev != NULL )
      elem->prev->next = elem->next;
    
    // Update the next element
    if ( elem->next != NULL )
      elem->next->prev = elem->prev;

    if ( elem == MemberTable[MemID].HeadBuffer )
      MemberTable[MemID].HeadBuffer = elem->next;

    if ( elem == MemberTable[MemID].TailBuffer )
      MemberTable[MemID].TailBuffer = elem->prev;

    delete elem;

    NumOfMsgPresent--;
  }
}


/*! \brief Check if the message pointed by  \a elem was expired
 */ 
inline bool DCDT_PostOffice_SLWDC::CheckTimeoutElem( DCDT_MsgListElem_SLWDC * elem)
{
  struct timeval now;
  DCDT_MsgListElem_SLWDC * elemtmp;


  elemtmp = NULL;

  if ( elem != elemtmp  ) 
    {
      gettimeofday(&now, NULL) ;     
      return ( DeltaSeconds( elem->insert_time, now) > TIMEOUT_MSG);
    }
  else
    return false;
};

/*! \brief Slides the dispatcher's queue and manages the messages.
 *
 *  Slides the dispatcher's queue from the head and deliver the messages to the member according to the subscription informations.
 *  Then clean the dispatcher's queue; if there is only a single message to be processed nothing is done.
 */
inline void DCDT_PostOffice_SLWDC::CleanupList ()
{
  DCDT_MsgListElem_SLWDC *actElem;
  DCDT_MsgListElem_SLWDC *tmpLastInsert;
  DCDT_MsgListElem_SLWDC *tmpElem;
  int MemberID = 0;
  int msg_type;
  int temp,templ,tempr;

  actElem=Head->next;

  // We use a new temporary variable to prevent the deletion of a message inserted but not processed.
  // The new message will be processed in the next dispatcher cycle if the message isn't the last in the queue
  tmpLastInsert = LastInsert;
  
  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("") );

  if( NumOfMsgPresent>0 ) {
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("NumOfMsgPresent %d\n", NumOfMsgPresent ));

    while ( true ) {
      if ( actElem->MsgPtr == NULL ) {
	// Error end of list reached
	break;
      }
      InsertListTime(0,WRITEINIT);
      msg_type = actElem->MsgPtr->ReadType ();
      MemberID = 0;
      for (int i=0; i<NUMINT_MEMBERS;i++) {
	for (int t=1;t<=SIZEOF_INTEGER;t++) { 
		
	  MemberID++;
	  templ = SubscribedTable[msg_type].SubscribersIfLocal[i];
	  templ = templ << (SIZEOF_INTEGER - t);
	  templ = templ >> (SIZEOF_INTEGER-1);
	  
	  tempr = SubscribedTable[msg_type].SubscribersIfRemote[i];
	  tempr = tempr << (SIZEOF_INTEGER - t);
	  tempr = tempr >> (SIZEOF_INTEGER-1); 
	  
	  temp = (templ | tempr);
	  
	  if (temp && (MemberID != actElem->LocalPublisherID)) {

	    // Copy the message into the member's queue
	    DCDT_MsgListElem_SLWDC * newMsgListElem;
	    // newMsgListElem = new (DCDT_MsgListElem_SLWDC); ???
	    
	    actElem->MsgPtr->Increment_Readers_Counter();

	    newMsgListElem = new (DCDT_MsgListElem_SLWDC);

	    newMsgListElem->MsgPtr = actElem->MsgPtr; 
	    
	    newMsgListElem->next = NULL; 
	         
	    gettimeofday(&(newMsgListElem->insert_time), NULL);
	    SetType (newMsgListElem, msg_type);

	    InsertListTime(0,WRITELOCKINIT);
	    MemberTable[MemberID].mtx.lock();
	    MemberTable[MemberID].condition_var.lock();
	    InsertListTime(0,WRITELOCKEND);
	   
            if ( (MemberTable[MemberID].HeadBuffer) ) {
	      MemberTable[MemberID].TailBuffer->next = newMsgListElem;
            }
	    else {
	      MemberTable[MemberID].HeadBuffer = newMsgListElem; 
	    }

	    // If someone is waiting send a signal
	    if ( MemberTable[MemberID].waiting ) {
	      MemberTable[MemberID].waiting = false;
	      MemberTable[MemberID].condition_var.signal();
	    }

	    MemberTable[MemberID].TailBuffer = newMsgListElem;
	    MemberTable[MemberID].condition_var.unlock();
	    MemberTable[MemberID].mtx.unlock();
	  }
	}
      }

      InsertListTime(0,DELETEINIT);

      // Cut the actual element from the queue and delete it
      // Then we can go to the next message
      actElem->prev->next = actElem->next;
      actElem->next->prev = actElem->prev;
      
      tmpElem = actElem;
      
      actElem = actElem->next;
      delete tmpElem;
      NumOfMsgPresent--;

      if ( tmpLastInsert == tmpElem )
	break;

      InsertListTime(0,DELETEEND);
      InsertListTime(0,WRITEEND); 
    }
  }
  
  TRC_VOID_RETURN ( DCDT_TRC_POSTOFFICE, TRC1 );
};



/*! \brief Update the readers mask for the element \a Elem and message type \a num_type.
 */
inline void DCDT_PostOffice_SLWDC::SetReaders (DCDT_MsgListElem_SLWDC * Elem, int num_type)
{  
  int i;
  int PublisherID, offset, pos;
  unsigned int mask;


  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetReaders"));
  
  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    PublisherID = Elem->LocalPublisherID;
    
    offset = (PublisherID - 1) / SIZEOF_INTEGER;
    pos = PublisherID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    // Fill the readers mask 
    for (i = 0; i < NUMINT_MEMBERS; i++) {
      if (Elem->local || myAgora->CheckMemberProfile(PublisherID, MPM_BRIDGE))
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfLocal[i];
      else
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfRemote[i];
      
      
    };
    
    // Avoid self-reading
    Elem->readers[offset] &= mask;
  }
};

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice_SLWDC::SubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType )
{
  int offset, pos;

  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Subscription request from Member %d to Msg Type %d", Member_ID, ReqType ));
    
    // Update the member subscription table
    // Don't need a lock because only the same member can update his subscription mask
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);

    if ( ReqType & DCDT_LOCAL_MSG )
      MemberTable[Member_ID].LocalSubscribedMsg[offset] |= (1 << (pos - 1));
    
    if ( ReqType & DCDT_REMOTE_MSG )
      MemberTable[Member_ID].RemoteSubscribedMsg[offset] |= (1 << (pos - 1));
    
#ifdef DCDT_USE_DEBUG
    printf ("MemberTable[%d].LocalSubscribedMsg=",Member_ID);
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].LocalSubscribedMsg[i]);
    printf ("\n");
    
    printf ("MemberTable[%d].RemoteSubscribedMsg=",Member_ID);
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].RemoteSubscribedMsg[i]);
    printf ("\n");
#endif

    // Update the message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    
    SubscribedTable[MsgType].mtx.lock ();
    
    // Reading politics: 
    // - local members can read remote or locale msgs
    // - bridges read all msgs;
    // - remote members read local msgs

    // Local messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_USER ) && ( ReqType & DCDT_LOCAL_MSG ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_LINK ) ) )
      SubscribedTable[MsgType].SubscribersIfLocal[offset] |= (1 << (pos - 1));
    
    // Remote messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( ( ReqType & DCDT_REMOTE_MSG ) && (myAgora->CheckMemberProfile(Member_ID, MPM_USER)) ) )
      SubscribedTable[MsgType].SubscribersIfRemote[offset] |= (1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.unlock ();
    
#ifdef DCDT_USE_DEBUG
    printf ("SubscribedTable[%d].SubscribersIfLocal=",MsgType);
    for (int i = (NUMINT_MEMBERS - 1); i >= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfLocal[i]);
    printf ( "\n" );
    
    printf("SubscribedTable[%d].SubscribersIfRemote=",MsgType);
    for (int i = (NUMINT_MEMBERS - 1); i>= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfRemote[i]);
    printf ("\n");
#endif
  
    // Update the update flags
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  } //end if valid msg type
};

/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice_SLWDC::UnSubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType)
{
  int  offset, pos;
  unsigned int mask;
  
  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    // Update member subscription table
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    if ( ReqType & DCDT_LOCAL_MSG )
      MemberTable[Member_ID].LocalSubscribedMsg[offset] &= mask;

    if ( ReqType & DCDT_REMOTE_MSG )
      MemberTable[Member_ID].RemoteSubscribedMsg[offset] &= mask;
    
#ifdef DCDT_USE_DEBUG
    printf ("MemberTable[%d].LocalSubscribedMsg=",Member_ID);
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].LocalSubscribedMsg[i]);
    printf ("\n");
    
    printf ("MemberTable[%d].RemoteSubscribedMsg=",Member_ID);
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].RemoteSubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.lock ();

    // Local messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_USER ) && ( ReqType & DCDT_LOCAL_MSG ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_LINK ) ) )
      SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
    
    // Remote messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( ( ReqType & DCDT_REMOTE_MSG ) && (myAgora->CheckMemberProfile(Member_ID, MPM_USER) ) ) )
      SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
    
    SubscribedTable[MsgType].mtx.unlock ();
    
    // Update the update flags    
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  }

  // TODO Check if there isn't messages with the same Member_ID to be deleted
};

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 */
inline void DCDT_PostOffice_SLWDC::UnSubscribeAll ( int Member_ID )
{
  int i, j, offset, pos, MsgType;
  unsigned int msg_mask, mask;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  // For all the messages:
  // - If we are susbscribed unsubscribe and update the message subscription mask
  for (i = 0; i < NUMINT_MSGTYPE; i++) {
    for (j = 0; j < SIZEOF_INTEGER; j++) {
      msg_mask = 1;
      msg_mask <<= j;

      if ( ( msg_mask & MemberTable[Member_ID].LocalSubscribedMsg[i] ) ||
	   ( msg_mask & MemberTable[Member_ID].RemoteSubscribedMsg[i] ) )
	{
	  MsgType = SIZEOF_INTEGER * i + j + 1;
	  SubscribedTable[MsgType].mtx.lock ();
	  
	  // Unsubscribe Member_ID from local and remote msg types
	  SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
	  SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
	  
	if ( myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	     myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ))
	  SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
	
	SubscribedTable[MsgType].mtx.unlock ();
      }
    }
    
    MemberTable[Member_ID].LocalSubscribedMsg[i] = 0;
    MemberTable[Member_ID].RemoteSubscribedMsg[i] = 0;
  }
  
  UpdateMask_mtx.lock();
  for (int i=0; i<Member_ID; i++)
    MemberTable[i].update = true;
  for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
    MemberTable[i].update = true;
  UpdateMask_flag = true;
  UpdateMask_mtx.unlock();
  
}

/*! \brief Check if member \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change the values so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The meber isn't subscribed to this message. 
 */
/*
inline int DCDT_PostOffice_SLWDC::MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem_SLWDC * Elem)
{
  int i;
  unsigned int match = 0;

  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--) {
    match |= (Elem->type_mask[i] & MemberTable[Member_ID].SubscribedMsg[i]);
  }
  
  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("MatchMsgTypeMask Member_ID %d - Return %d", Member_ID, match ));
  return match;
};
*/

/*! \brief Check if \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The member isn't subscribed to this message.
 */
/*
inline int DCDT_PostOffice_SLWDC::MatchReadersMask (int Member_ID, DCDT_MsgListElem_SLWDC * Elem)
{
  int offset, pos;
  unsigned int mask;
  int match = 0;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = 1 << (pos - 1);
  match = (Elem->readers[offset] & mask);

  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("readers[%d]=%d, mask = %d, match=%d", offset, Elem->readers[offset], mask, match));
  
  return match;
};
*/

/*! \brief Set the type of the message point by \a Elem to \a num_type.
 *
 *  This method is used only during the insert phase so no lock is required because there aren't concurrency problems.
 */
inline void DCDT_PostOffice_SLWDC::SetType (DCDT_MsgListElem_SLWDC * Elem, int num_type)
{
  int offset, pos;

  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    offset = (num_type - 1) / SIZEOF_INTEGER;
    pos = num_type - (offset * SIZEOF_INTEGER);
    
    Elem->type_mask[offset] |= 1 << (pos -1);
    
  }
  else {
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetType ERROR"));
  }
}



/*! \brief Update the SubscribedMsgMask used by LinkTx when talking with another Agora.
 *
 *  Rebuild the global subscribers mask; the global mask is build upon the single message mask of each active member, don't include the mask of the member \a MemID; if \a MemID is a bridge all the members masks are used otherwise only the masks of MPM_USER and MPM_BRIDGE type members are used; so in a LinkTx member we can build the subscriber mask that needs to be passed to the other side of the link.
 */
inline void DCDT_PostOffice_SLWDC::UpdateSubscribersMask (int MemID)
{
  int i, j, isBridge;
  unsigned int temp[NUMINT_MSGTYPE];
  
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    temp[j] = 0;
  
  isBridge=myAgora->CheckMemberProfile( MemID, MPM_BRIDGE );

  // Build a new temporary mask
  for (i = 0; i < MAX_MEMBERS; i++)
    {
      if (MemberTable[i].alive && (i != MemID)) {
	// MemID == MPM_BRIDGE -> request any message needed by local members
        // MemID != MPM_BRIDGE -> request messages only for MPM_USER and MPM_BRIDGE members ( not dynamic links, MPM_LINK )

	if( isBridge ) {
	  /* MPM_BRIDGE -> relay for local msg use Local Mask */
	  if ( myAgora->CheckMemberProfile(i, MPM_USER ) ||
	       myAgora->CheckMemberProfile(i, MPM_BRIDGE ) ) {
	    for (j = 0; j < NUMINT_MSGTYPE; j++)
	      temp[j] |= MemberTable[i].LocalSubscribedMsg[j];
	  }
	}
	else {
	  /* !MPM_BRIDGE -> use Remote Mask */
	  if ( myAgora->CheckMemberProfile(i, MPM_USER ) ||
               myAgora->CheckMemberProfile(i, MPM_BRIDGE ) ) {
            for (j = 0; j < NUMINT_MSGTYPE; j++)
              temp[j] |= MemberTable[i].RemoteSubscribedMsg[j];
	  }
	}
      }
    }
  
  subscribed_mtx.lock ();
  
  // Now update the mask
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    SubscribedMsgMask[j] = temp[j];
  
  subscribed_mtx.unlock ();  
};

/*! \brief Returns the decimal value corresponding to the mask signature of the element pointed by \a Elem.
 */
inline int DCDT_PostOffice_SLWDC::GetTypeFromMask (DCDT_MsgListElem_SLWDC * Elem)
{
  int i, j, temp;
  
  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
    {
      if (Elem->type_mask[i])
    	{
  	  temp = Elem->type_mask[i];
  	  j = 0;
  	  while (temp)
	    {
	      temp = temp >> 1;
	      j++;
	    };
	  return (i * SIZEOF_INTEGER + j);
	};
    };
};


/*! \brief Check if SubscribedMsgMask is changed.
 *
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed  
 */
inline bool DCDT_PostOffice_SLWDC::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  bool match=true;
  
  UpdateMask_mtx.lock();
  if(MemberTable[MemID].update) 
    {
      UpdateSubscribersMask(MemID);
      MemberTable[MemID].update = false;
    };
  UpdateMask_mtx.unlock();
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    match = match && (mask[i] == SubscribedMsgMask[i]);
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mask[i] = SubscribedMsgMask[i];

  // The old mask matches ?
  if (!match)
    return true;
  else
    return false;
}

inline bool DCDT_PostOffice_SLWDC::UpdateSubs(int MemID) 
{
  return MemberTable[MemID].update;
}

#ifdef TEST_PART

inline void DCDT_PostOffice_SLWDC::InsertListTime (int MemID, int Type_Op)
{
  DCDT_TIME x = GetTime();
  TimeTesterElem *newElem;
  newElem = new TimeTesterElem;
  newElem->Data = x;
  newElem->Type_Op = Type_Op;
  newElem->prev = TestElem[MemID].prev;
  TestElem[MemID].prev = newElem; 
} 


inline void DCDT_PostOffice_SLWDC::VisualList()
{
  TimeTesterElem *newElem;
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    newElem = TestElem[i].prev;
    while (newElem) {
      cout<<"["<<i<<","<<newElem->Type_Op<<","<<newElem->Data-TestElem[i].Data<<"]";
      newElem=newElem->prev;
    }
  } 
}

inline void DCDT_PostOffice_SLWDC::SaveList(char* nomefile)  
{
  TimeTesterElem *newElem; 
  FILE *fout;
  time_t longtime;
  struct tm* time_experiment;
  
  char string[100] = "DATI_";

  strcat(string,nomefile); 
  
  fout = fopen (string,"wb");

  // Insert time and date of the operation
  time(&longtime);
  time_experiment = localtime (&longtime);

  putw(time_experiment->tm_hour,fout);
  putw(time_experiment->tm_min,fout); 
  putw(time_experiment->tm_mday,fout); 
  putw(time_experiment->tm_mon,fout);
  putw(1900 + time_experiment->tm_year,fout);

  // Software version and system parameters
  putw(VERSION,fout);
  putw(NUMMEMBERFORMAIN,fout);

  // Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
  putw(NUMREADER,fout);      
  putw(NUMWRITER,fout);      
  putw(NUMREADERWRITER,fout); 
  
  // Number of cycles  
  putw(READCYCLE,fout); 
  putw(WRITECYCLE,fout);        
  putw(READWRITECYCLE,fout);   
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    
    newElem = TestElem[i].prev;
    
    while (newElem) {
      putw(newElem->Type_Op,fout);
      putw(newElem->Data-TestElem[i].Data,fout);
      
      newElem=newElem->prev;
    }
    putw(666,fout);
  }

  fclose(fout);
}

#endif /* TEST_PART */

#endif /* DCDT_POSTOFFICE__SLWDC_H */
