/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE_H
#define DCDT_POSTOFFICE_H



#ifdef TEST_PART

typedef struct _TimeTesterElem
{
  // Time of the measure
  DCDT_TIME Data;
  
  // Type of the measure
  int Type_Op;
    
  _TimeTesterElem *prev;
}
TimeTesterElem;

#endif

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <iostream>
#include <DCDT_Agora.h>
#include <DCDT_Defs.h>
#include <sys/time.h>
#include <DCDT_Time.h>
#include <DCDT_Msg.h>
#include <DCDT_ctrace.h>

class DCDT_Member;

/// Structure needed to keep track of various member's subscriptions
typedef struct _SubsMsgElem
{
  DCDT_Mutex mtx;
  // Local and Remote are referred to the msg, i.e. to the its publisher

  /// Suscribers for local generated messages
  unsigned int SubscribersIfLocal[NUMINT_MEMBERS];

  /// Subscribers for remote generated messages
  unsigned int SubscribersIfRemote[NUMINT_MEMBERS];
}
SubsMsgTypeElem;

/*! \brief Informations about the single message stored.
 *
 *  When an elemet was deleted it checks if the message was read by all the members, if this is the last member the message is deleted, otherwise the readers counter is decremented.
 */
class DCDT_MsgListElem
{
 public:
  DCDT_MsgListElem ()
    {
      ToBeDeleted = false;
      MsgPtr = NULL;
      
      for(int i = 0; i < NUMINT_MEMBERS; i++)
	{
	  readers[i] = 0;
	};
      
      
      numMemberPointing=0;
      
    };
  
  ~DCDT_MsgListElem ()
    {
      next = NULL;
      prev = NULL;
      free (next);
      free (prev);
      
      // If this is the last reader we have to delete the message.
      // Oterwise we have to decrement the readers counter
      if (MsgPtr->Get_Readers_Counter() == 1)  
	delete MsgPtr;
      else
	MsgPtr->Decrement_Readers_Counter(); 
    };
  
  struct timeval insert_time;  
  
  bool ToBeDeleted;
  
  


  /// Flag: true -> message requested by a local user member ( member profile = MPM_USER )  
  bool local;
  
  int LocalPublisherID;
  
  DCDT_MsgListElem *next;
  DCDT_MsgListElem *prev;
  
  
  
  /// Array of bit reporting members subscribed to this message type
  unsigned int readers[NUMINT_MEMBERS];
  
  /// Type of the message
  unsigned int type_mask[NUMINT_MSGTYPE];

  DCDT_Msg *MsgPtr;
  
  // Counter for the members pointing at this element.
  // In the BRIDGE version the couter must be syncronized with a mutex
  unsigned int numMemberPointing;
  
  
  
  /** Mutex used when modifing the mask */
  DCDT_Mutex mtxmask;
  
};

// Structure for the registered members
typedef struct _MemberTableElem
{
  DCDT_Mutex mtx;
  
  // Flag: reporting whether or not this slot is hosting a member
  bool alive;
  
  // subcribed msg list 
  unsigned int SubscribedMsg[NUMINT_MSGTYPE];

  // Flag: reports that the susbscription mask was updated
  bool update;
  
  /// Pointer to the actual visited element
  DCDT_MsgListElem *ActReadMsgElem;

  /// Pointer to the last visited element
  DCDT_MsgListElem *LastVisitedMsgElem;
  
  

  // Pointers to the MemberTable when every single member has a buffer
  // Allow the buffer slide
  

  
}
MemberTableElem;

/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice
{
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;
  friend class DCDT_Member;

 public:
    
  DCDT_PostOffice (DCDT_Agora * agora)
    {
      myAgora = agora;
      NumOfMsgPresent = 0;

      Head = new DCDT_MsgListElem;
      
      
      
      
      
      
      Tail = Head;
      Head->MsgPtr = NULL;
      Head->prev = NULL;
      Head->next = NULL;
      
      

      
      
      
      
#ifdef TEST_PART
      
      // Init list used during the test phase
      DCDT_TIME InitTime = GetTime();
      
      
      for (int j = 0; j < MAX_MEMBERS; j++)
	{
	  TestElem[j].prev = NULL;
	  TestElem[j].Data = InitTime;
	}
      
#endif


      // MemberTable initialization
      for (int i = 0; i < MAX_MEMBERS; i++)
	{
	  MemberTable[i].alive = false;
	  MemberTable[i].update = false;
	  
	  




	  
	  for (int j = 0; j < NUMINT_MSGTYPE; j++)
	    MemberTable[i].SubscribedMsg[j] = 0;
	}
      
      for (int i = 0; i < MAX_MSG_TYPE; i++)
	{
	  for (int j = 0; j < NUMINT_MEMBERS; j++) 
	    {
	      SubscribedTable[i].SubscribersIfLocal[j] = 0;
	      SubscribedTable[i].SubscribersIfRemote[j] = 0;
	    }
	}
      printf ("Postoffice created\n");
    };
  

  ~DCDT_PostOffice ()
    {
      // TODO: clean of MsgList
    };
  
  inline void LogMsg( DCDT_Msg *msg ) {
    char buffer[512];
    DCDT_MsgHeader *header;
    char *payload;
    int idmy=0,idx=0,payloadlen=0;

    if ( msg != NULL ) {
      header = msg->GetHeader();

      if ( (payload=(char *)msg->GetPayload()) != NULL ) {
	payloadlen=msg->ReadPayloadLen();
	
	while ( ( idx < payloadlen ) && ( idmy < (int)sizeof(buffer)-7 ) ){
	  if ( isprint( payload[idx] ) )
	    idmy += sprintf ( &buffer[idmy], "%c", payload[idx] );
	  else
	    idmy += sprintf ( &buffer[idmy], "<0x%.2hhX>", 
			      (unsigned char)payload[idx] );
	  idx++;
	}
      }
    else
      sprintf ( buffer, "NULL" );
    
      TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <%s>", msg->ReadType(), buffer) );
    }
    else
      TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <NULL>", msg->ReadType() ) );
  };

  inline void AddMember(int Mem_ID);
  inline void AddSysMember(int Mem_ID);
  
  inline void DeleteMemberTableEntry(int Mem_ID );
  
  inline void AddMsg (DCDT_Msg * MP, int Mem_ID);
  
  inline DCDT_Msg * ReadNextMsg (int MemID); // TODO: Check if return "const DCDT_MSG *" is better

  inline void CleanupList ();
  
  inline void SubscribeMsgTypeID (int Member_ID, int MsgType);
  inline void UnSubscribeMsgTypeID (int Member_ID, int MsgType);
  inline void UnSubscribeAll( int Member_ID );
  
  inline DCDT_MsgListElem *SearchMsg (DCDT_Msg *);
  
  inline void UpdateSubscribersMask (int MemID);
  inline bool ReadSubscribedMsgMask(unsigned int *mask, int MemID);
  inline bool UpdateSubs(int MemID);
    
 private:
  
  inline void SetReaders (DCDT_MsgListElem * Elem, int num_type); 
  

  
  
  
  

  
  
  inline int MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem *);
  inline int MatchReadersMask (int Member_ID, DCDT_MsgListElem *);
  
  inline void SetType (DCDT_MsgListElem * Elem, int type);

  // Functions used during the test phase
#ifdef TEST_PART
  inline void InsertListTime (int MemID, int TypeOp);
  inline void VisualList ();
  inline void SaveList (char* nomefile);
#else
#endif
    
  inline bool CheckTimeoutElem(DCDT_MsgListElem *); 
  
  inline int GetTypeFromMask (DCDT_MsgListElem * Elem);
  
  /*! \brief For each message type we have the corresponding member mask.
   *
   *  A member mask use a sequence of bit to repesent all the members, if a bit is set to 1 the corresponding member is subscribed to this type of messages.
   *  Ex: 00010001 -> Members 1 and 5 are subscribed to this message type.
   *  The member mask is composed of two parts local and remote, the local one was used when a message was locally originated or was received from a BRIDGE member; the remote one was used when the message was received from a STATIC or LINK member.
   */
  SubsMsgTypeElem SubscribedTable[MAX_MSG_TYPE];
  
  /// Mutex to access SubscribedMsgMask
  DCDT_Mutex subscribed_mtx;

  /// Message types requested by local members used by LinkTx members
  unsigned int SubscribedMsgMask[NUMINT_MSGTYPE];
  
  // Mutex to access UpdateMask_flag and MemberTable[i].update
  DCDT_Mutex UpdateMask_mtx;

  /// Flag: true when SubscribedMsgMask was updated
  bool UpdateMask_flag;

  
  MemberTableElem MemberTable[MAX_MEMBERS];
  
  int NumOfMsgPresent;

  DCDT_Agora *myAgora;  
};




inline void DCDT_PostOffice::AddMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice::AddSysMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice::DeleteMemberTableEntry(int Mem_ID){
  
  // TODO
  MemberTable[Mem_ID].alive = false;
  
};







/*! \brief Add a new message to the PostOffice queue.
 *
 * To add a new message to the PostOffice queue a member must have to acquire the write lock to the queue then he can add the message and release the lock.
 */
inline void DCDT_PostOffice::AddMsg (DCDT_Msg * MP, int Mem_ID)
{
  int msg_type;

  InsertListTime(Mem_ID,WRITEINIT); 
  InsertListTime(Mem_ID,WRITELOCKINIT);
  insert_mtx.lock ();	
  InsertListTime(Mem_ID,WRITELOCKEND);

  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ( "MemID: %d", Mem_ID));
  TRC_EXEC ( DCDT_TRC_POSTOFFICE, TRC0, LogMsg( MP ); );

  DCDT_MsgListElem *newMsgListElem = new (DCDT_MsgListElem);
  newMsgListElem->MsgPtr = MP;
  newMsgListElem->LocalPublisherID = Mem_ID;
  newMsgListElem->prev = NULL;
  newMsgListElem->next = NULL;
  newMsgListElem->local = myAgora->CheckMemberProfile(Mem_ID, MPM_USER);
  newMsgListElem->numMemberPointing = 0; 
  gettimeofday(&(newMsgListElem->insert_time), NULL);
  
  msg_type = MP->ReadType ();
  
  SetType (newMsgListElem, msg_type);

  SetReaders (newMsgListElem, msg_type);

  newMsgListElem->prev = Tail;
  Tail->next = newMsgListElem;
  Tail = newMsgListElem;
  
  NumOfMsgPresent++;

  insert_mtx.unlock ();
  InsertListTime(Mem_ID,WRITEEND);

  TRC_VOID_RETURN ( DCDT_TRC_POSTOFFICE, TRC1 );
};














/*! \brief Check if all the members subscribed to a message have already read it.
    \return TRUE  - All the members have read the message ( ??? )
            FALSE - Not all the members have read the message 
 */
inline int DCDT_PostOffice::CheckIfMsgRead (DCDT_MsgListElem * elem) {
  int temp;
  
  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ( "CheckIfMsgRead %p", elem ) );

  temp = 0;
  
  for (int i = 0; i < NUMINT_MEMBERS; i++) {

    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("[%d]", elem->readers[i]));
    
    temp |= elem->readers[i];
  }

  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("temp =%d, !temp=%d\n", temp, !temp));
  return !temp;
};








/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the message queue; if we found one we have to return a pointer to it.
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *
DCDT_PostOffice::ReadNextMsg (int MemID)
{
  DCDT_MsgListElem *nextElem;

  InsertListTime(MemID,READINIT); 
  InsertListTime(MemID,READLOCKINIT);
  insert_mtx.lock ();
  InsertListTime(MemID,READLOCKEND);

  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("MemID: %d", MemID) );
 
  // Slide the message list until we find an interesting message
  nextElem = MemberTable[MemID].LastVisitedMsgElem->next;
 
  while ((nextElem != NULL)) {

    MemberTable[MemID].LastVisitedMsgElem->numMemberPointing--;
    MemberTable[MemID].LastVisitedMsgElem = nextElem;
    nextElem->numMemberPointing++;
    
    if (MatchMsgTypeMask (MemID, nextElem) && MatchReadersMask (MemID, nextElem)) {
      // We are subscribed to this type of messages, we have to return it
      MemberTable[MemID].ActReadMsgElem = nextElem;
      insert_mtx.unlock ();
      InsertListTime(MemID,READDONEEND);
      
      TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, (""), nextElem->MsgPtr );
      return (nextElem->MsgPtr);
    }
     
    nextElem = MemberTable[MemID].LastVisitedMsgElem->next;
  }

  // No new messages found
  insert_mtx.unlock ();

  InsertListTime(MemID,READEND);
  TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("NULL"), NULL );
  return ((DCDT_Msg *) NULL);
};















/*! \brief Check if a message was read by all the member, if it is mark the message as deletable.
 */
inline void DCDT_PostOffice::SetActMsgDeletable (int Member_ID)
{
  int offset, pos;
  unsigned int mask;
  
  // Lock the message queue
  insert_mtx.lock();

  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  MemberTable[Member_ID].ActReadMsgElem->readers[offset] &= mask;
  
  
  // TODO: If we have changed the readers mask change the pointer to the
  //       read message, so we can guarantee the consistency
  
  if (CheckIfMsgRead (MemberTable[Member_ID].ActReadMsgElem))
    {
      TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Deleting an element\n"));    
      MemberTable[Member_ID].ActReadMsgElem->ToBeDeleted = true;
    }

  // Unlock the message queue
  insert_mtx.unlock();

};







/*! \brief Delete the element \a Elem locking the global message queue.
 */
inline void DCDT_PostOffice::RemoveListElement (DCDT_MsgListElem * Elem)
{
  
  int PointingMemID;
  int counter=0;

  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg %p", Elem ));

  if ( Elem ) {
    if (Elem->next) {
      if (Elem->prev) {
	TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("The element is in the middle. Deleting it."));
	
	Elem->prev->next = Elem->next;
	Elem->next->prev = Elem->prev;
	
	delete Elem;
      }
      else {
	TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Deleting the first element"));
	Head->next = Elem->next;
	Elem->next->prev = NULL;
	
	delete Elem;
      }
    }
    else {
      TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Deleting the last element"));
      Tail = Elem->prev;
      Elem->prev->next = NULL;
      
      delete Elem;
    }
    
    NumOfMsgPresent--;
  }
  
  TRC_VOID_RETURN( DCDT_TRC_POSTOFFICE, TRC1 );
};









/*! \brief Check if the message pointed by  \a elem was expired
 */ 
inline bool DCDT_PostOffice::CheckTimeoutElem( DCDT_MsgListElem * elem)
{
  struct timeval now;
  DCDT_MsgListElem * elemtmp;


  elemtmp = Head;

  if ( elem != elemtmp  ) 
    {
      gettimeofday(&now, NULL) ;     
      return ( DeltaSeconds( elem->insert_time, now) > TIMEOUT_MSG);
    }
  else
    return false;
};





/** \brief Cleanup the message list.
 *
 *  Lock and slide the message list from head to tail.
 *  If we find a message that was marked for deletion ( means that all subscribers have already read the message ) delete it.
 */

inline void DCDT_PostOffice::CleanupList ()
{
  DCDT_MsgListElem *actElem, *tempElem;
  InsertListTime(0,DELETELOCKINIT);
  insert_mtx.lock ();
  InsertListTime(0,DELETELOCKEND);

  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("") );

  if( NumOfMsgPresent ) {
    actElem = Head->next;
    
    while (actElem) {
      tempElem = actElem;	
      
      actElem = actElem->next;
      
      if ( tempElem->ToBeDeleted || CheckTimeoutElem( tempElem)) {
	
	if (tempElem->numMemberPointing == 0) {
	  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Deleting an element (ToBeDeleted = %d, Timeout = %d)",tempElem->ToBeDeleted ,CheckTimeoutElem( tempElem)));    
	  InsertListTime(0,DELETEINIT);
	  RemoveListElement (tempElem);
	  InsertListTime(0,DELETEEND);
	}
      } 
    }
  }
  
  insert_mtx.unlock ();

  TRC_VOID_RETURN ( DCDT_TRC_POSTOFFICE, TRC1 );
};














/*! \brief Update the readers mask for the element \a Elem and message type \a num_type.
 */
inline void DCDT_PostOffice::SetReaders (DCDT_MsgListElem * Elem, int num_type)
{  
  int i;
  int PublisherID, offset, pos;
  unsigned int mask;


  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetReaders"));
  
  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    PublisherID = Elem->LocalPublisherID;
    
    offset = (PublisherID - 1) / SIZEOF_INTEGER;
    pos = PublisherID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    // Fill the readers mask 
    for (i = 0; i < NUMINT_MEMBERS; i++) {
      if (Elem->local || myAgora->CheckMemberProfile(PublisherID, MPM_BRIDGE))
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfLocal[i];
      else
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfRemote[i];
      
      
    };
    
    // Avoid self-reading
    Elem->readers[offset] &= mask;
  }
};

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::SubscribeMsgTypeID (int Member_ID, int MsgType)
{
  int offset, pos;

  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Subscription request from Member %d to Msg Type %d", Member_ID, MsgType));
    
    // Update the member subscription table
    // Don't need a lock because only the same member can update his subscription mask
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    MemberTable[Member_ID].SubscribedMsg[offset] |= (1 << (pos - 1));
    
#ifdef _DEBUG_DCDT
    printf ("MemberTable=");
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].SubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update the message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    
    SubscribedTable[MsgType].mtx.lock ();
    
    // Reading politics: local members and bridges read all msgs;
    // remote members read local msgs
    
    SubscribedTable[MsgType].SubscribersIfLocal[offset] |= (1 << (pos - 1));
    if (myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE))
      SubscribedTable[MsgType].SubscribersIfRemote[offset] |= (1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.unlock ();
    
    
#ifdef _DEBUG_DCDT
    printf ("SubscribedTable: Local=");
    for (int i = (NUMINT_MEMBERS - 1); i >= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfLocal[i]);
    printf(" Remote=");
    for (int i = (NUMINT_MEMBERS - 1); i>= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfRemote[i]);
    printf ("\n\n");
#endif
    
    // Update the update flags
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  } //end if valid msg type
};

/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::UnSubscribeMsgTypeID (int Member_ID, int MsgType)
{
  int  offset, pos;
  unsigned int mask;
  
  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    // Update member subscription table
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    MemberTable[Member_ID].SubscribedMsg[offset] &= mask;
    
    // DEBUG
#ifdef _DEBUG_DCDT
    printf ("MemberTable=");
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].SubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.lock ();
    
    SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
    if (myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE))
      SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
    
    SubscribedTable[MsgType].mtx.unlock ();

    // Update the update flags    
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  }

  // TODO Check if there isn't messages with the same Member_ID to be deleted
};

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 */
inline void DCDT_PostOffice::UnSubscribeAll ( int Member_ID )
{
  int i, j, offset, pos, MsgType;
  unsigned int msg_mask, mask;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  // For all the messages:
  // - If we are susbscribed unsubscribe and update the message subscription mask
  for (i = 0; i < NUMINT_MSGTYPE; i++) {
    for (j = 0; j < SIZEOF_INTEGER; j++) {
      msg_mask = 1;
      msg_mask <<= j;
      if (msg_mask & MemberTable[Member_ID].SubscribedMsg[i]) {
	MsgType = SIZEOF_INTEGER * i + j + 1;
	SubscribedTable[MsgType].mtx.lock ();
	
	SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
	if ( myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	     myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ))
	  SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
	
	SubscribedTable[MsgType].mtx.unlock ();
      }
    }
    
    MemberTable[Member_ID].SubscribedMsg[i] = 0;
  }
  
  UpdateMask_mtx.lock();
  for (int i=0; i<Member_ID; i++)
    MemberTable[i].update = true;
  for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
    MemberTable[i].update = true;
  UpdateMask_flag = true;
  UpdateMask_mtx.unlock();
  
}

/*! \brief Check if member \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change the values so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The meber isn't subscribed to this message. 
 */
inline int DCDT_PostOffice::MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem * Elem)
{
  int i;
  unsigned int match = 0;

  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--) {
    match |= (Elem->type_mask[i] & MemberTable[Member_ID].SubscribedMsg[i]);
  }
  
  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("MatchMsgTypeMask Member_ID %d - Return %d", Member_ID, match ));
  return match;
};

/*! \brief Check if \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The member isn't subscribed to this message.
 */
inline int DCDT_PostOffice::MatchReadersMask (int Member_ID, DCDT_MsgListElem * Elem)
{
  int offset, pos;
  unsigned int mask;
  int match = 0;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = 1 << (pos - 1);
  match = (Elem->readers[offset] & mask);

  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("readers[%d]=%d, mask = %d, match=%d", offset, Elem->readers[offset], mask, match));
  
  return match;
};

/*! \brief Set the type of the message point by \a Elem to \a num_type.
 *
 *  This method is used only during the insert phase so no lock is required because there aren't concurrency problems.
 */
inline void DCDT_PostOffice::SetType (DCDT_MsgListElem * Elem, int num_type)
{
  int offset, pos;

  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    offset = (num_type - 1) / SIZEOF_INTEGER;
    pos = num_type - (offset * SIZEOF_INTEGER);
    
    Elem->type_mask[offset] |= 1 << (pos -1);
    
  }
  else {
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetType ERROR"));
  }
}



/*! \brief Update the SubscribedMsgMask used by LinkTx when talking with another Agora.
 *
 *  Rebuild the global subscribers mask; the global mask is build upon the single message mask of each active member, don't include the mask of the member \a MemID; if \a MemID is a bridge all the members masks are used otherwise only the masks of MPM_USER and MPM_BRIDGE type members are used; so in a LinkTx member we can build the subscriber mask that needs to be passed to the other side of the link.
 */
inline void DCDT_PostOffice::UpdateSubscribersMask (int MemID)
{
  int i, j;
  unsigned int temp[NUMINT_MSGTYPE];
  
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    temp[j] = 0;
  
  // Build a new temporary mask
  for (i = 0; i < MAX_MEMBERS; i++)
    {
      if (MemberTable[i].alive && (i != MemID)) {
	// MemID == MPM_BRIDGE -> request any message needed by local members
        // MemID != MPM_BRIDGE -> request messages only for MPM_USER and MPM_BRIDGE members ( not dynamic links, MPM_LINK )
	if(myAgora->CheckMemberProfile(i, MPM_USER) ||
	   myAgora->CheckMemberProfile(i, MPM_BRIDGE) || 
	   myAgora->CheckMemberProfile(MemID, MPM_BRIDGE)) {
	  for (j = 0; j < NUMINT_MSGTYPE; j++)
	    temp[j] |= MemberTable[i].SubscribedMsg[j];
	}
      }
    }
  
  subscribed_mtx.lock ();
  
  // Now update the mask
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    SubscribedMsgMask[j] = temp[j];
  
  subscribed_mtx.unlock ();  
};

/*! \brief Returns the decimal value corresponding to the mask signature of the element pointed by \a Elem.
 */
inline int DCDT_PostOffice::GetTypeFromMask (DCDT_MsgListElem * Elem)
{
  int i, j, temp;
  
  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
    {
      if (Elem->type_mask[i])
    	{
  	  temp = Elem->type_mask[i];
  	  j = 0;
  	  while (temp)
	    {
	      temp = temp >> 1;
	      j++;
	    };
	  return (i * SIZEOF_INTEGER + j);
	};
    };
};


/*! \brief Check if SubscribedMsgMask is changed.
 *
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed  
 */
inline bool DCDT_PostOffice::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  bool match=true;
  
  UpdateMask_mtx.lock();
  if(MemberTable[MemID].update) 
    {
      UpdateSubscribersMask(MemID);
      MemberTable[MemID].update = false;
    };
  UpdateMask_mtx.unlock();
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    match = match && (mask[i] == SubscribedMsgMask[i]);
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mask[i] = SubscribedMsgMask[i];

  // The old mask matches ?
  if (!match)
    return true;
  else
    return false;
}

inline bool DCDT_PostOffice::UpdateSubs(int MemID) 
{
  return MemberTable[MemID].update;
}

#ifdef TEST_PART

inline void DCDT_PostOffice::InsertListTime (int MemID, int Type_Op)
{
  DCDT_TIME x = GetTime();
  TimeTesterElem *newElem;
  newElem = new TimeTesterElem;
  newElem->Data = x;
  newElem->Type_Op = Type_Op;
  newElem->prev = TestElem[MemID].prev;
  TestElem[MemID].prev = newElem; 
} 


inline void DCDT_PostOffice::VisualList()
{
  TimeTesterElem *newElem;
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    newElem = TestElem[i].prev;
    while (newElem) {
      cout<<"["<<i<<","<<newElem->Type_Op<<","<<newElem->Data-TestElem[i].Data<<"]";
      newElem=newElem->prev;
    }
  } 
}

inline void DCDT_PostOffice::SaveList(char* nomefile)  
{
  TimeTesterElem *newElem; 
  FILE *fout;
  time_t longtime;
  struct tm* time_experiment;
  
  char string[100] = "DATI_";

  strcat(string,nomefile); 
  
  fout = fopen (string,"wb");

  // Insert time and date of the operation
  time(&longtime);
  time_experiment = localtime (&longtime);

  putw(time_experiment->tm_hour,fout);
  putw(time_experiment->tm_min,fout); 
  putw(time_experiment->tm_mday,fout); 
  putw(time_experiment->tm_mon,fout);
  putw(1900 + time_experiment->tm_year,fout);

  // Software version and system parameters
  putw(VERSION,fout);
  putw(NUMMEMBERFORMAIN,fout);

  // Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
  putw(NUMREADER,fout);      
  putw(NUMWRITER,fout);      
  putw(NUMREADERWRITER,fout); 
  
  // Number of cycles  
  putw(READCYCLE,fout); 
  putw(WRITECYCLE,fout);        
  putw(READWRITECYCLE,fout);   
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    
    newElem = TestElem[i].prev;
    
    while (newElem) {
      putw(newElem->Type_Op,fout);
      putw(newElem->Data-TestElem[i].Data,fout);
      
      newElem=newElem->prev;
    }
    putw(666,fout);
  }

  fclose(fout);
}

#endif

