/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE_VIRTUAL_H
#define DCDT_POSTOFFICE_VIRTUAL_H

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <iostream>
#include <DCDT_Agora.h>
#include <DCDT_Defs.h>
#include <sys/time.h>
#include <DCDT_Time.h>
#include <DCDT_Msg.h>
#include <DCDT_ctrace.h>

#ifdef TEST_PART

typedef struct _TimeTesterElem
{
  // Time of the measure
  DCDT_TIME Data;
  
  // Type of the measure
  int Type_Op;
    
  _TimeTesterElem *prev;
}
TimeTesterElem;

#endif

class DCDT_Member;

/// Structure needed to keep track of various member's subscriptions
typedef struct _SubsMsgElem
{
  DCDT_Mutex mtx;
  // Local and Remote are referred to the msg, i.e. to the its publisher

  /// Suscribers for local generated messages
  unsigned int SubscribersIfLocal[NUMINT_MEMBERS];

  /// Subscribers for remote generated messages
  unsigned int SubscribersIfRemote[NUMINT_MEMBERS];
}
SubsMsgTypeElem;


/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice
{
  
 public:
    
  DCDT_PostOffice (DCDT_Agora * agora)
    {
      myAgora = agora;
      NumOfMsgPresent = 0;

#ifdef TEST_PART
      
      // Init list used during the test phase
      DCDT_TIME InitTime = GetTime();
      for (int j = 0; j < MAX_MEMBERS; j++)
	{
	  TestElem[j].prev = NULL;
	  TestElem[j].Data = InitTime;
	}
      
#endif


      // MemberTable initialization
      for (int i = 0; i < MAX_MEMBERS; i++)
	{
	  for (int j = 0; j < NUMINT_MSGTYPE; j++)
	    MemberTable[i].SubscribedMsg[j] = 0;
	}
      
      for (int i = 0; i < MAX_MSG_TYPE; i++)
	{
	  for (int j = 0; j < NUMINT_MEMBERS; j++) 
	    {
	      SubscribedTable[i].SubscribersIfLocal[j] = 0;
	      SubscribedTable[i].SubscribersIfRemote[j] = 0;
	    }
	}
      
      printf ("Postoffice created\n");
    };
  

  ~DCDT_PostOffice ()
    {
      // TODO: clean of MsgList
    };
  
  inline void LogMsg( DCDT_Msg *msg ) {
    char buffer[512];
    DCDT_MsgHeader *header;
    char *payload;
    int idmy=0,idx=0,payloadlen=0;

    if ( msg != NULL ) {
      header = msg->GetHeader();

      if ( (payload=(char *)msg->GetPayload()) != NULL ) {
	payloadlen=msg->ReadPayloadLen();
	
	while ( ( idx < payloadlen ) && ( idmy < (int)sizeof(buffer)-7 ) ){
	  if ( isprint( payload[idx] ) )
	    idmy += sprintf ( &buffer[idmy], "%c", payload[idx] );
	  else
	    idmy += sprintf ( &buffer[idmy], "<0x%.2hhX>", 
			      (unsigned char)payload[idx] );
	  idx++;
	}
      }
    else
      sprintf ( buffer, "NULL" );
    
      TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <%s>", msg->ReadType(), buffer) );
    }
    else
      TRC_PRINT( DCDT_TRC_POSTOFFICE, TRC0, ("Type: %d Payload <NULL>", msg->ReadType() ) );
  };

  virtual void AddMember(int Mem_ID) = 0;
  virtual void AddSysMember(int Mem_ID) = 0;
  virtual void DeleteMemberTableEntry(int Mem_ID ) = 0;
  virtual void AddMsg (DCDT_Msg * MP, int Mem_ID) = 0;
  virtual DCDT_Msg * ReadNextMsg (int MemID) = 0; // TODO: Check if return "const DCDT_MSG *" is better
  virtual void CleanupList () = 0;
  
  virtual void SubscribeMsgTypeID (int Member_ID, int MsgType) = 0;
  virtual void UnSubscribeMsgTypeID (int Member_ID, int MsgType) = 0;
  virtual void UnSubscribeAll( int Member_ID ) = 0;
  
  virtual void UpdateSubscribersMask (int MemID) = 0;
  virtual bool ReadSubscribedMsgMask(unsigned int *mask, int MemID) = 0;
  virtual bool UpdateSubs(int MemID) = 0;
    
 private:

#ifdef TEST_PART
  // Functions used during the tests
  inline void InsertListTime (int MemID, int TypeOp);
  inline void VisualList ();
  inline void SaveList (char* nomefile);
#else
  inline void InsertListTime (int MemID, int TypeOp){}; 
  inline void VisualList (){};
  inline void SaveList (char *nomefile){};  
#endif
  
  /*! \brief For each message type we have the corresponding member mask.
   *
   *  A member mask use a sequence of bit to repesent all the members, if a bit is set to 1 the corresponding member is subscribed to this type of messages.
   *  Ex: 00010001 -> Members 1 and 5 are subscribed to this message type.
   *  The member mask is composed of two parts local and remote, the local one was used when a message was locally originated or was received from a BRIDGE member; the remote one was used when the message was received from a STATIC or LINK member.
   */
  SubsMsgTypeElem SubscribedTable[MAX_MSG_TYPE];
  
  /// Mutex to access SubscribedMsgMask
  DCDT_Mutex subscribed_mtx;

  /// Message types requested by local members used by LinkTx members
  unsigned int SubscribedMsgMask[NUMINT_MSGTYPE];
  
  /// Flag: true when SubscribedMsgMask was updated
  bool UpdateMask_flag;

#ifdef TEST_PART 

  // Lists used during the test phase, each member has a private list to avoid
  // concurrent access
  TimeTesterElem TestElem[MAX_MEMBERS];
  
#endif
    
  int NumOfMsgPresent;

  DCDT_Agora *myAgora;  
};

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::SubscribeMsgTypeID (int Member_ID, int MsgType)
{
  int offset, pos;

  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Subscription request from Member %d to Msg Type %d", Member_ID, MsgType));
    
    // Update the member subscription table
    // Don't need a lock because only the same member can update his subscription mask
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    MemberTable[Member_ID].SubscribedMsg[offset] |= (1 << (pos - 1));
    
#ifdef _DEBUG_DCDT
    printf ("MemberTable=");
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].SubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update the message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    
    SubscribedTable[MsgType].mtx.lock ();
    
    // Reading politics: local members and bridges read all msgs;
    // remote members read local msgs
    
    SubscribedTable[MsgType].SubscribersIfLocal[offset] |= (1 << (pos - 1));
    if (myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE))
      SubscribedTable[MsgType].SubscribersIfRemote[offset] |= (1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.unlock ();
    
    
#ifdef _DEBUG_DCDT
    printf ("SubscribedTable: Local=");
    for (int i = (NUMINT_MEMBERS - 1); i >= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfLocal[i]);
    printf(" Remote=");
    for (int i = (NUMINT_MEMBERS - 1); i>= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfRemote[i]);
    printf ("\n\n");
#endif
    
    // Update the update flags
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  } //end if valid msg type
};

/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice::UnSubscribeMsgTypeID (int Member_ID, int MsgType)
{
  int  offset, pos;
  unsigned int mask;
  
  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    // Update member subscription table
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    MemberTable[Member_ID].SubscribedMsg[offset] &= mask;
    
    // DEBUG
#ifdef _DEBUG_DCDT
    printf ("MemberTable=");
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].SubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.lock ();
    
    SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
    if (myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE))
      SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
    
    SubscribedTable[MsgType].mtx.unlock ();

    // Update the update flags    
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  }

  // TODO Check if there isn't messages with the same Member_ID to be deleted
};

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 */
inline void DCDT_PostOffice::UnSubscribeAll ( int Member_ID )
{
  int i, j, offset, pos, MsgType;
  unsigned int msg_mask, mask;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  // For all the messages:
  // - If we are susbscribed unsubscribe and update the message subscription mask
  for (i = 0; i < NUMINT_MSGTYPE; i++) {
    for (j = 0; j < SIZEOF_INTEGER; j++) {
      msg_mask = 1;
      msg_mask <<= j;
      if (msg_mask & MemberTable[Member_ID].SubscribedMsg[i]) {
	MsgType = SIZEOF_INTEGER * i + j + 1;
	SubscribedTable[MsgType].mtx.lock ();
	
	SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
	if ( myAgora->CheckMemberProfile(Member_ID, MPM_USER) || 
	     myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ))
	  SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
	
	SubscribedTable[MsgType].mtx.unlock ();
      }
    }
    
    MemberTable[Member_ID].SubscribedMsg[i] = 0;
  }
  
  UpdateMask_mtx.lock();
  for (int i=0; i<Member_ID; i++)
    MemberTable[i].update = true;
  for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
    MemberTable[i].update = true;
  UpdateMask_flag = true;
  UpdateMask_mtx.unlock();
  
}

/*! \brief Check if member \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change the values so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The meber isn't subscribed to this message. 
 */
inline int DCDT_PostOffice::MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem * Elem)
{
  int i;
  unsigned int match = 0;

  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--) {
    match |= (Elem->type_mask[i] & MemberTable[Member_ID].SubscribedMsg[i]);
  }
  
  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("MatchMsgTypeMask Member_ID %d - Return %d", Member_ID, match ));
  return match;
};

/*! \brief Check if \a Member_ID is interested in the message pointed by \a Elem.
 *
 *  We don't change so no lock is needed.
 *  \return TRUE - The member is subscribed to this message.
 *  \return FALSE - The member isn't subscribed to this message.
 */
inline int DCDT_PostOffice::MatchReadersMask (int Member_ID, DCDT_MsgListElem * Elem)
{
  int offset, pos;
  unsigned int mask;
  int match = 0;
  
  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = 1 << (pos - 1);
  match = (Elem->readers[offset] & mask);

  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("readers[%d]=%d, mask = %d, match=%d", offset, Elem->readers[offset], mask, match));
  
  return match;
};

/*! \brief Set the type of the message point by \a Elem to \a num_type.
 *
 *  This method is used only during the insert phase so no lock is required because there aren't concurrency problems.
 */
inline void DCDT_PostOffice::SetType (DCDT_MsgListElem * Elem, int num_type)
{
  int offset, pos;

  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    offset = (num_type - 1) / SIZEOF_INTEGER;
    pos = num_type - (offset * SIZEOF_INTEGER);
    
    Elem->type_mask[offset] |= 1 << (pos -1);
    
  }
  else {
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetType ERROR"));
  }
}


#ifdef SINGLE_LOCK_WITH_BRIDGE_WITH_CLEANER_FROM_HEAD_TO_TAIL

/*! \brief Safe method to go to the next message.
 */
inline DCDT_MsgListElem * DCDT_PostOffice::GoToNextListMsg (DCDT_MsgListElem * Elem,int MemID)
{
  DCDT_MsgListElem* Elemtmp ;

  // Lock those mutex in sequence:
  // 1 - Counter of members pointing to this element
  // 2 - Pointer to the next element
  // 3 - Pointer to the previous element of the next element
  // 4 - Counter of members pointing to the next element
  InsertListTime(MemID,READLOCKINIT);

  Elem->mtxnumMemberPointing.lock(); 
  Elem->mtxNextPointer.lock();
  Elem->next->mtxPrevPointer.lock();
  Elem->next->mtxnumMemberPointing.lock();
  
  InsertListTime(MemID,READLOCKEND);

  // Go to the next element
  Elemtmp=Elem;
  Elem=Elem->next;

  Elem->numMemberPointing++;
  Elemtmp->numMemberPointing--;
     
  MemberTable[MemID].LastVisitedMsgElem=Elem;

  // Unlock all the mutex
  Elemtmp->mtxnumMemberPointing.unlock();
  Elemtmp->mtxNextPointer.unlock();
  Elem->mtxPrevPointer.unlock();
  Elem->mtxnumMemberPointing.unlock();
  Elemtmp=NULL;
  
  // We are on the next element
  return (Elem);
  
  
};

#endif

/*! \brief Update the SubscribedMsgMask used by LinkTx when talking with another Agora.
 *
 *  Rebuild the global subscribers mask; the global mask is build upon the single message mask of each active member, don't include the mask of the member \a MemID; if \a MemID is a bridge all the members masks are used otherwise only the masks of MPM_USER and MPM_BRIDGE type members are used; so in a LinkTx member we can build the subscriber mask that needs to be passed to the other side of the link.
 */
inline void DCDT_PostOffice::UpdateSubscribersMask (int MemID)
{
  int i, j;
  unsigned int temp[NUMINT_MSGTYPE];
  
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    temp[j] = 0;
  
  // Build a new temporary mask
  for (i = 0; i < MAX_MEMBERS; i++)
    {
      if (MemberTable[i].alive && (i != MemID)) {
	// MemID == MPM_BRIDGE -> request any message needed by local members
        // MemID != MPM_BRIDGE -> request messages only for MPM_USER and MPM_BRIDGE members ( not dynamic links, MPM_LINK )
	if(myAgora->CheckMemberProfile(i, MPM_USER) ||
	   myAgora->CheckMemberProfile(i, MPM_BRIDGE) || 
	   myAgora->CheckMemberProfile(MemID, MPM_BRIDGE)) {
	  for (j = 0; j < NUMINT_MSGTYPE; j++)
	    temp[j] |= MemberTable[i].SubscribedMsg[j];
	}
      }
    }
  
  subscribed_mtx.lock ();
  
  // Now update the mask
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    SubscribedMsgMask[j] = temp[j];
  
  subscribed_mtx.unlock ();  
};

/*! \brief Returns the decimal value corresponding to the mask signature of the element pointed by \a Elem.
 */
inline int DCDT_PostOffice::GetTypeFromMask (DCDT_MsgListElem * Elem)
{
  int i, j, temp;
  
  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
    {
      if (Elem->type_mask[i])
    	{
  	  temp = Elem->type_mask[i];
  	  j = 0;
  	  while (temp)
	    {
	      temp = temp >> 1;
	      j++;
	    };
	  return (i * SIZEOF_INTEGER + j);
	};
    };
};


/*! \brief Check if SubscribedMsgMask is changed.
 *
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed  
 */
inline bool DCDT_PostOffice::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  bool match=true;
  
  UpdateMask_mtx.lock();
  if(MemberTable[MemID].update) 
    {
      UpdateSubscribersMask(MemID);
      MemberTable[MemID].update = false;
    };
  UpdateMask_mtx.unlock();
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    match = match && (mask[i] == SubscribedMsgMask[i]);
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mask[i] = SubscribedMsgMask[i];

  // The old mask matches ?
  if (!match)
    return true;
  else
    return false;
}

inline bool DCDT_PostOffice::UpdateSubs(int MemID) 
{
  return MemberTable[MemID].update;
}

#ifdef TEST_PART

inline void DCDT_PostOffice::InsertListTime (int MemID, int Type_Op)
{
  DCDT_TIME x = GetTime();
  TimeTesterElem *newElem;
  newElem = new TimeTesterElem;
  newElem->Data = x;
  newElem->Type_Op = Type_Op;
  newElem->prev = TestElem[MemID].prev;
  TestElem[MemID].prev = newElem; 
} 


inline void DCDT_PostOffice::VisualList()
{
  TimeTesterElem *newElem;
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    newElem = TestElem[i].prev;
    while (newElem) {
      cout<<"["<<i<<","<<newElem->Type_Op<<","<<newElem->Data-TestElem[i].Data<<"]";
      newElem=newElem->prev;
    }
  } 
}

inline void DCDT_PostOffice::SaveList(char* nomefile)  
{
  TimeTesterElem *newElem; 
  FILE *fout;
  time_t longtime;
  struct tm* time_experiment;
  
  char string[100] = "DATI_";

  strcat(string,nomefile); 
  
  fout = fopen (string,"wb");

  // Insert time and date of the operation
  time(&longtime);
  time_experiment = localtime (&longtime);

  putw(time_experiment->tm_hour,fout);
  putw(time_experiment->tm_min,fout); 
  putw(time_experiment->tm_mday,fout); 
  putw(time_experiment->tm_mon,fout);
  putw(1900 + time_experiment->tm_year,fout);

  // Software version and system parameters
  putw(VERSION,fout);
  putw(NUMMEMBERFORMAIN,fout);

  // Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
  putw(NUMREADER,fout);      
  putw(NUMWRITER,fout);      
  putw(NUMREADERWRITER,fout); 
  
  // Number of cycles  
  putw(READCYCLE,fout); 
  putw(WRITECYCLE,fout);        
  putw(READWRITECYCLE,fout);   
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    
    newElem = TestElem[i].prev;
    
    while (newElem) {
      putw(newElem->Type_Op,fout);
      putw(newElem->Data-TestElem[i].Data,fout);
      
      newElem=newElem->prev;
    }
    putw(666,fout);
  }

  fclose(fout);
}

#endif

#endif
