#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>

#include <pthread.h>
#include "latency-graph.h"

struct params {
   int num;
   unsigned long period;
};

void *ThreadStartRoutine( void *arg )
{
  struct params *par=(struct params *)arg;
  double mytime1=0.0, mytime2, delta;
  char name[80];

  sprintf ( name, "sample%d.txt", par->num );

  printf ( "Starting thread %d - period %ld\n", par->num, par->period );
  lat_record_start( par->num );

  while ( !lat_check_samplerange(par->num) ) {
     if ( mytime1 != 0.0 ) {
        mytime2 = lat_gettime();
        lat_record_latency( par->num, mytime2-mytime1 );
        lat_next_sample( par->num );
     }
     else {
        mytime2 = lat_gettime();
     }

     mytime1 = mytime2;
     lat_record_actual_time( par->num, mytime1 );
     usleep ( par->period );
  }

  printf ( "Ending thread %d\n", par->num );
  dump_sample( par->num, name );
}

int main(void)
{
  int i;
  double mytime1;
  double mytime2;
  double delta;
  pthread_t thd1, thd2, thd3;
  struct params par1, par2, par3;

  // sets realtime priority (SCHED_FIFO, max priority)
  // see latency-graph.c
  set_realtime_priority(SCHED_FIFO,1);

  // keep all memory locked into physical mem, to guarantee realtime-behaviour
  mlockall(MCL_CURRENT|MCL_FUTURE);

  // initializes the API , 2 latency-curves , 200 sampling points
  lat_init(2,200); 

  // sets the deadline of the first latency-curve (0) to 80ms
  lat_set_deadline(0, 0.080);

  //  sets the range1 of the first latency-curve to 28ms - 32ms
  lat_set_range1(0, 0.028,0.032);

  // sets range2 of the first-latency-curve to 20ms - 40ms
  lat_set_range2(0, 0.020,0.040);

  //  sets the flags for the first latency-curve
  //  LAT_SHOW_DEADLINE shows the deadline and the statistics about
  //  deadline misses
  //  LAT_SHOW_RANGE1 shows the range1 bounds and the statistics
  //  about range crossings
  //  same for LAT_SHOW_RANGE2
  //  note that no flags were set for the second latency-curve (1) ,
  //  therefore no deadline and ranges and associated statistics are shown
//  lat_set_flags(0,LAT_SHOW_DEADLINE | LAT_SHOW_RANGE1 | LAT_SHOW_RANGE2);

  //  sets the color of the first latency curve
  lat_set_color_main(0,0,128,255);

  //  sets the color of the second latency curve
  lat_set_color_main(1,0,255,255);

  par1.num=0;
  par1.period=80000;
  if ( pthread_create ( &thd1, NULL, ThreadStartRoutine, &par1 ) )
     printf ( "Error creating thread 1\n" );

  par2.num=1;
  par2.period=20000;
  if ( pthread_create ( &thd2, NULL, ThreadStartRoutine, &par2 ) )
     printf ( "Error creating thread 2\n" );

  pthread_join ( thd1, NULL );
  pthread_join ( thd2, NULL );

  // draws the graph into the file named "test.gif"

  draw_latency_graph("test.gif");
 
  dump_samples ( "samples.txt" ); 
  // that's all folks :-)
  return(0);

}

