#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>

#include "latency-graph.h"

int main(void)
{
  int i;
  double mytime1;
  double mytime2;
  double delta;

  // sets realtime priority (SCHED_FIFO, max priority)
  // see latency-graph.c
  set_realtime_priority(SCHED_FIFO,1);

  // keep all memory locked into physical mem, to guarantee realtime-behaviour
  mlockall(MCL_CURRENT|MCL_FUTURE);

  // initializes the API , 2 latency-curves , 200 sampling points
  lat_init(2,200); 

  // sets the deadline of the first latency-curve (0) to 80ms
  lat_set_deadline(0, 0.080);

  //  sets the range1 of the first latency-curve to 28ms - 32ms
  lat_set_range1(0, 0.028,0.032);

  // sets range2 of the first-latency-curve to 20ms - 40ms
  lat_set_range2(0, 0.020,0.040);

  //  sets the flags for the first latency-curve
  //  LAT_SHOW_DEADLINE shows the deadline and the statistics about
  //  deadline misses
  //  LAT_SHOW_RANGE1 shows the range1 bounds and the statistics
  //  about range crossings
  //  same for LAT_SHOW_RANGE2
  //  note that no flags were set for the second latency-curve (1) ,
  //  therefore no deadline and ranges and associated statistics are shown
//  lat_set_flags(0,LAT_SHOW_DEADLINE | LAT_SHOW_RANGE1 | LAT_SHOW_RANGE2);

  //  sets the color of the first latency curve
  lat_set_color_main(0,0,128,255);

  //  sets the color of the second latency curve
  lat_set_color_main(1,0,255,255);

  //  starts recording 
  lat_record_start();
  
  for(i=0;i<200;i++)
  {
    // check if there is still room left in the array, otherwise exit
    if(lat_check_samplerange) break;

//  gets actual time in seconds 64bit floating point, should provide at least
//  usec accuracy

    mytime1=lat_gettime();

    //  records the actual time (x-axis)
    lat_record_actual_time(mytime1);

    //  this is the function we want to profile, 
    //  as we see from the graphs usleep() sleeps 30ms instead of 20ms , and
    //  the time varies as the system load increases
    usleep(20000);

    //  get the actual time after usleep() returns
    mytime2=lat_gettime();

    //  compute the difference
    delta=mytime2-mytime1;

    //  record the point into the first latency-curve (0) array
    lat_record_latency(0,delta);


    // just do the same as above but this time with usleep(50ms)
    // to show that more than one latency-curve is allowed
    mytime1=lat_gettime();
    usleep(50000);
    mytime2=lat_gettime();
    delta=mytime2-mytime1;
    lat_record_latency(1,delta);

    //  finally advance the index to record the next sample
    lat_next_sample();
  }

  // draws the graph into the file named "test.gif"

  draw_latency_graph("test.gif");

  // that's all folks :-)
  return(0);

}

