/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!!                   INFO SOLUTION SPA                     
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                    
**!!! PROJECT       : ROBY                                 
**!!!                                                                    
**!!! CLASS         : ModuleExpert
**!!!
**!!! PACKAGE       : DCDT
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : ModuleExpert.cpp
**!!!                                                                   
**!!! CREATION DATE : 25.10.2001         
**!!!                                                                   
**!!! AUTHOR        : P.M.  Roby  Team                             
**!!!                 Cristian Giussani
**!!!                                                                   
**!!!                                                                   
**!!!___________________________________________________________________
**!!!                                                                   
**!!! SPECS ON DOCS : <Spec. Doc Name A>  Rev.: AA.aa                   
**!!!                                                                     
**!!!                 <Spec. Doc Name B>  Rev.: BB.bb                   
**!!!                ............                                       
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*/


/*----------------------------------------------------------------------
 *--------------------- SYSTEM  INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include <string.h>
#include <iostream>
/*----------------------------------------------------------------------
 *--------------------- PROJECT INCLUDE FILES --------------------------
 *---------------------------------------------------------------------*/
#include "ModuleExpert.h"
#include "const.h"

unsigned int LocalGetTime()
{
  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv,&tz);
  return((unsigned int)(tv.tv_sec * 1000000 + tv.tv_usec));
}

/*----------------------------------------------------------------------
 *-----------------------  DEFINE MACRO  -------------------------------
 *---------------------------------------------------------------------*/
//const int ONE_READ_ACCESS=1;
/*----------------------------------------------------------------------
 *-----------------------  TYPES DECLARATION  --------------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *--------------------  STATIC VARIABLE DECLARATION  -------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *--------------------  STATIC FUNCTION DECLARATION  -------------------
 *---------------------------------------------------------------------*/
#define MODULE_EXPERT_DEBUG 1
/*----------------------------------------------------------------------
 *--------------------  CONSTRUCTOR DESTRUCTOR  ------------------------
 *---------------------------------------------------------------------*/
ModuleExpert :: ModuleExpert (ModuleDispatch *Dispatch,
			      unsigned int period ):  
  DCDT_Member( Dispatch, (DCDT_TIME)period )
{
  Dispatcher = Dispatch;
  IncomingMessage = NULL;
#ifdef EXPERTS_TIME
  mExpertTime = LocalGetTime();
#endif
}

   
ModuleExpert ::  ~ModuleExpert ()
{
}

/*----------------------------------------------------------------------
 *-------------------- PUBLIC METHODS DEFINITIONS ----------------------
 *---------------------------------------------------------------------*/
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  AddMessageRequest
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
int ModuleExpert::AddMessageRequest(int  Type,
				    RequestType Flag)
{
  switch ( Flag ) {
   
  case LOCAL:
    SubscribeMsgType ( Type, DCDT_LOCAL_MSG );
    ReqMessages.push_back(Type);
    break;

  case REMOTE:
    SubscribeMsgType( Type, DCDT_REMOTE_MSG );
    break;

  default:
    return RC_ERROR;

  }

  return RC_OK;  
}
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  RemMessageRequest
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
int ModuleExpert::RemMessageRequest(int  Type, RequestType Flag)
{
  ReqMessagesList::iterator TypeInList;

  switch ( Flag ) {
  case LOCAL:
    UnSubscribeMsgType ( Type, DCDT_LOCAL_MSG );

    // Erase Requested Messages from List
    TypeInList= ReqMessages.begin();

    while (TypeInList!=ReqMessages.end()) {
      if (*TypeInList==Type) {
        ReqMessages.erase(TypeInList);
        return ( RC_OK );
      }
      TypeInList++;
    }
    return ( RC_ERROR );
    break;

  case REMOTE:
    UnSubscribeMsgType ( Type, DCDT_REMOTE_MSG );
    break;

  default:
    return RC_ERROR;
  }

  return RC_OK;  
}

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  SendMessage
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/ 
int ModuleExpert::SendMessage(unsigned char * Buffer, 
			      unsigned short Length,
	 		      unsigned short Type,
			      char Priority)
{

  DCDT_Msg* Message=CreateMsg( Type, Priority );
  Message->AllocPayload ( Length, Buffer );
  ShareMsg(Message);

  // If we have requested the same type of message make a local copy
  if (IsPrivateReqMessages(Type)){
    DCDT_Msg* MyMessage=DupMsg( Message );

#ifdef MODULE_EXPERT_DEBUG_1
    cout<<" Private Message type ("<<MyMessage->ReadType()<<") sent"<<endl;
#endif

    PList.push_back(MyMessage);
  }
  return RC_OK;
}
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  ReceiveMessage
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : Type and Buffer pointer
**---                                                         
**--- OUTPUT PARAM  : Type 
**---                                                         
**--- RETURN PARAM  : Buffer Length
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : Get Next  Message of Type deleting it from the 
**---                local expert queue.It allocates the Buffer
**---                                                                   
**-------------------------------------------------------------------------
*/ 
unsigned short  ModuleExpert::ReceiveMessage(unsigned short &Type,
					     unsigned char * &Buffer,
					     bool wait ) 
{ 
  int length;

  IncomingMessage = ReceiveMsgType ( Type, false );
  if ( IncomingMessage != NULL ) {
    Type = IncomingMessage->ReadType();
    length = IncomingMessage->ReadPayloadLen();
    Buffer = new unsigned char[length];
    memcpy ( Buffer, IncomingMessage->GetPayload(), length );

    return length;
  }
  else {
    PrivateMessagesList::iterator p = PList.begin();
    while ( p != PList.end() ) {
      IncomingMessage = (DCDT_Msg * ) *p;
      if (( Type == ALL_TYPES ) || ( IncomingMessage->ReadType() == Type )) {
	Type = IncomingMessage->ReadType();
	length = IncomingMessage->ReadPayloadLen();
	
	Buffer = new unsigned char[length];
	
	memcpy ( Buffer, IncomingMessage->GetPayload(), length );
	PList.erase( p );
	delete IncomingMessage;

	return length ;
      }
      p++;
    }
  }

  /* Wait a message */
  if ( wait ) {
    IncomingMessage = ReceiveMsgType ( Type, true );
    assert ( IncomingMessage != NULL );
    Type = IncomingMessage->ReadType();
    length = IncomingMessage->ReadPayloadLen();
    Buffer = new unsigned char[length];
    memcpy ( Buffer, IncomingMessage->GetPayload(), length );

    return length;
  }

  return 0;
}

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  ReceiveLastMessage
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : Type and Buffer pointer
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Buffer Length
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : Get Last Message of Type deleting it from the 
**---                local expert queue.It allocates the Buffer.
**---                It deletes all the messages of type Type received.                                                   
**-------------------------------------------------------------------------
*/
unsigned short ModuleExpert::ReceiveLastMessage ( unsigned short &Type, 
						  unsigned char * &Buffer,
						  bool wait )
{
  int length;

  IncomingMessage = ReceiveLastMsg( Type, false );
  if ( IncomingMessage != NULL ) {
    // There is a new message
    Type = IncomingMessage->ReadType();
    length = IncomingMessage->ReadPayloadLen();
    
    Buffer = new unsigned char[length];
    
    memcpy ( Buffer, IncomingMessage->GetPayload(), length );
    
    return length ;
  }
  else {
    // No new message, so check the private list in reverse order
    // and return the first element deleting the others
    length = 0;
    PrivateMessagesList::iterator p = PList.end();
    while ( p != PList.begin() ) {
      IncomingMessage = ( DCDT_Msg * ) *p;
      if (( Type == ALL_TYPES ) || ( IncomingMessage->ReadType() == Type )) {
	if ( length == 0 ) {
	  Type = IncomingMessage->ReadType();
	  length = IncomingMessage->ReadPayloadLen();
	  Buffer = new unsigned char[length];
	  memcpy ( Buffer, IncomingMessage->GetPayload(), length );
	}
	PList.erase( p );
	//delete IncomingMessage;
      }
      p--;
    }
    return length;
  }

  /* Wait a message */
  if ( wait ) {
    IncomingMessage = ReceiveMsgType ( Type, true );
    assert ( IncomingMessage != NULL );
    Type = IncomingMessage->ReadType();
    length = IncomingMessage->ReadPayloadLen();
    Buffer = new unsigned char[length];
    memcpy ( Buffer, IncomingMessage->GetPayload(), length );

    return length;
  }

  return 0;
}

/*
unsigned short ModuleExpert::ReceiveMessage ( unsigned char * &Buffer, unsigned short &Type )
{
   IncomingMessage = ReceiveMsg( false );
   if ( IncomingMessage == NULL )
    {
      if ( PList.begin() != PList.end() ) {
	IncomingMessage = (DCDT_Msg *) (*(PList.begin()));
	PList.pop_front();
	
	Type = IncomingMessage->ReadType();
	int length = IncomingMessage->ReadPayloadLen();
	Buffer = new unsigned char[length];
	memcpy ( Buffer, IncomingMessage->GetPayload(), length );
	delete IncomingMessage;
	
	return length ;
      }
    }
  else
    {
      Type = IncomingMessage->ReadType();
      int length = IncomingMessage->ReadPayloadLen();
      
      Buffer = new unsigned char[length];
      
      memcpy ( Buffer, IncomingMessage->GetPayload(), length );

      return length ;
    }

  return 0;
}
*/

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  ClearMessages
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : Type (Default All types)
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Result Type (RC_OK, RC_ERR)
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : Clear all messages of type "Type" in Module queue
**---                                                                   
**-------------------------------------------------------------------------
*/ 
int ModuleExpert::ClearMessages(unsigned short Type)
{
  ReceiveLastMsg ( Type, false );

  PrivateMessagesList::iterator p = PList.begin();
  while ( p != PList.end() ) {
    if ( ( Type == ALL_TYPES ) || ( IncomingMessage->ReadType() == Type ) ) {
      delete (DCDT_Msg *) *p;
      p = PList.erase( p );
    }
    else
      p++;
  }

  return RC_OK;
}
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  CreateMemArea
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : int Size,int Name,int Type,int Subtype
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Result Type
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : This is the Memory creation function
**---                                                                   
**-------------------------------------------------------------------------
*/ 
inline int ModuleExpert::CreateMemArea(int Size,
				       int Name,
				       int Type,
				       int SecondType)
{
  return RC_ERROR;
}

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  DestroyMemArea
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : int Name
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Result Type 
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : This is the Memory distruction  function
**---                                                                   
**-------------------------------------------------------------------------
*/ 
inline int ModuleExpert::DestroyMemArea(int Name)
{
  return RC_ERROR; 
}

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  MemCheckIn
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : int Name
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Memory Length
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : This is the Memory check In function
**---                                                                   
**-------------------------------------------------------------------------
*/ 
inline int ModuleExpert::MemCheckIn(int Name)
{
  return RC_ERROR;
}

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  MemCheckOut
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : int Name
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Result Type 
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : This is the Memory check Out function
**---                                                                   
**-------------------------------------------------------------------------
*/ 
inline int ModuleExpert::MemCheckOut(int Name)
{
  return RC_ERROR;
}


/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  MemWrite
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : int Name, void * Buffer 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :  Length written
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : This is the Memory Write function
**---                                                                   
**-------------------------------------------------------------------------
*/ 
inline int ModuleExpert::MemWrite(int Name,void * SourceBuffer)
{
  return RC_ERROR;
}



/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  MemRead
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : int Name, void * Buffer 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  : Length  Red
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : This is the Memory Red function
**---                                                                   
**-------------------------------------------------------------------------
*/ 
inline int ModuleExpert::MemRead(int Name,void * DestBuffer)
{
  return RC_ERROR;
}


/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  GetId
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   Module Expert Identifier
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
int ModuleExpert :: GetId()
{
  return ReadID();
}   

/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  Terminate
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : uSecs for Time Out
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :  Result
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
int ModuleExpert :: Terminate(int usec)
{
  Shutdown ( usec/1000 );
  return RC_OK;
}   

/*----------------------------------------------------------------------
 *-------------------- PRIVATE  METHODS DEFINITIONS --------------------
 *---------------------------------------------------------------------*/
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
void ModuleExpert :: DoYourJob (int wc)
{  
  if (wc)
    {
      return;
    }
#ifdef EXPERTS_TIME //*wiim* commento queste linee di output
  //cout << endl << mModuleName << " return time: " << LocalGetTime() - mExpertTime << endl;
  mExpertTime = LocalGetTime();
#endif

  RunDuty();

#ifdef EXPERTS_TIME
  //cout << endl << mModuleName << " duty cycle time: " << LocalGetTime() - mExpertTime << endl;
  mExpertTime = LocalGetTime();
#endif
} 
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
void ModuleExpert :: Init (void)  
{
  RunInit();
}  
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      :
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
void ModuleExpert :: Close(void)  
{
  RunClose();  
}   
/*---------------------------------------------------------------------- 
**---                                                            
**--- METHOD  NAME  :  IsPrivateReqMessages
**---                                                            
**---------------------------------------------------------------------- 
**---                                                         
**--- INPUT PARAM   : Message Type 
**---                                                         
**--- OUTPUT PARAM  : 
**---                                                         
**--- RETURN PARAM  :   
**---                                                         
**---------------------------------------------------------------------- 
**---                                                                   
**--- OVERVIEW      : Check if This Type has been requested
**---                
**---                                                                   
**-------------------------------------------------------------------------
*/
 
bool ModuleExpert::IsPrivateReqMessages(int Type){
  // Erase Requested Messages from List
  ReqMessagesList::iterator  TypeInList= ReqMessages.begin();
  while (TypeInList!=ReqMessages.end()) {
    if (*TypeInList==Type) {
      return (true);
    }
    TypeInList++;
  }
  return (false);
}
/*----------------------------------------------------------------------
 *-------------------- OPERATOR   DEFINITIONS --------------------------
 *---------------------------------------------------------------------*/

/*----------------------------------------------------------------------
 *----------------  STATIC FUNCTION  DEFINITIONS -----------------------
 *---------------------------------------------------------------------*/
