/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**!!!                                                                   
**!!!                   INFO SOLUTION SPA                     
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                    
**!!! PROJECT       : ROBY                                 
**!!!                                                                    
**!!! CLASS         : ModuleMember
**!!!
**!!! PACKAGE       : Middleware
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
**!!!                                                                   
**!!! FILE NAME     : ModuleExpert.cpp
**!!!                                                                   
**!!! CREATION DATE : 25.10.2001         
**!!!                                                                   
**!!! AUTHOR        : P.M.  Roby  Team                             
**!!!                 Cristian Giussani
**!!!                                                                   
**!!!                                                                   
**!!!___________________________________________________________________
**!!!                                                                   
**!!! SPECS ON DOCS : <Spec. Doc Name A>  Rev.: AA.aa                   
**!!!                                                                     
**!!!                 <Spec. Doc Name B>  Rev.: BB.bb                   
**!!!                ............                                       
**!!!                                                                   
**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*/

#include <string.h>
#include <iostream>
#include "ModuleMember.h"
#include "const.h"

unsigned int LocalGetTime()
{
  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv,&tz);
  return((unsigned int)(tv.tv_sec * 1000000 + tv.tv_usec));
}

ModuleMember::ModuleMember (ModuleDispatch *Dispatch,
			    unsigned int period, string name ):  
  DCDT_Member( Dispatch, (DCDT_TIME)period )
{
  Dispatcher = Dispatch;
  IncomingMessage = NULL;
#ifdef EXPERTS_TIME
  mExpertTime = LocalGetTime();
#endif
  mModuleName = name;
}

   
ModuleMember::~ModuleMember ()
{
}

int ModuleMember::AddMessageRequest(int  Type,
				    RequestType Flag)
{
  switch ( Flag ) {
    
  case LOCAL:
    SubscribeMsgType ( Type, DCDT_LOCAL_MSG );
    break;
    
  case REMOTE:
    SubscribeMsgType( Type, DCDT_REMOTE_MSG );
    break;
    
  default:
    return RC_ERROR;

  }
  
  return RC_OK;  
}

int ModuleMember::RemMessageRequest(int  Type, RequestType Flag)
{
  switch ( Flag ) {
  case LOCAL:
    UnSubscribeMsgType ( Type, DCDT_LOCAL_MSG );
    break;
    
  case REMOTE:
    UnSubscribeMsgType ( Type, DCDT_REMOTE_MSG );
    break;
    
  default:
    return RC_ERROR;
  }
  
  return RC_OK;  
}

int ModuleMember::SendMessage(unsigned char * Buffer, int Length,
			      int Type, int Warranty, char Priority)
{ 
  DCDT_Msg* Message=CreateMsg( Type, Priority );
  Message->SetPayload ( Buffer, Length );
  Message->SetDeliveryWarranty( Warranty );
  ShareMsg(Message);
  
  return RC_OK;
}

int ModuleMember::ReceiveMessage(int Type, unsigned char * &Buffer,
				 bool wait ) 
{ 
  IncomingMessage = ReceiveMsgType ( Type, wait );
  if ( IncomingMessage != NULL ) {
    Type = IncomingMessage->ReadType();
    Buffer = (unsigned char *)IncomingMessage->GetPayload();
    
    return IncomingMessage->ReadPayloadLen();
  }
  
  return 0;
}

int ModuleMember::ReceiveLastMessage(int Type, unsigned char * &Buffer,
				      bool wait )
{
  IncomingMessage = ReceiveLastMsg( Type, wait );
  if ( IncomingMessage != NULL ) {
    Type = IncomingMessage->ReadType();
    Buffer = (unsigned char *)IncomingMessage->GetPayload();

    return IncomingMessage->ReadPayloadLen() ;
  }
  
  return 0;
}

int ModuleMember::ClearMessages(int Type)
{
  return RC_OK;
}

int ModuleMember::GetId()
{
  return ReadID();
}   

int ModuleMember::Terminate(int usec)
{
  Shutdown ( usec/1000 );
  return RC_OK;
}   

void ModuleMember::ResetTimer()
{
#ifdef EXPERTS_TIME
  mExpertTime = LocalGetTime();
#endif
}

void ModuleMember::DoYourJob (int wc)
{  
  if (wc)
    {
      return;
    }
#ifdef EXPERTS_TIME //wiim commento queste linee di output
  //cout << endl << mModuleName << " return time: " << LocalGetTime() - mExpertTime << endl;
  mExpertTime = LocalGetTime();
#endif
  
  RunDuty();
  
#ifdef EXPERTS_TIME
  //cout << endl << mModuleName << " duty cycle time: " << LocalGetTime() - mExpertTime << endl;
  mExpertTime = LocalGetTime();
#endif
}

void ModuleMember::Init (void)  
{
  RunInit();
}

void ModuleMember::Close(void)  
{
  RunClose();  
}
