#include "StringMessageManager.h"
#include <sys/time.h>

StringMessageManager::StringMessageManager(ModuleDispatch* pDispatcher, 
			     long period, 
			     string module_name) : ModuleExpert(pDispatcher,period)
{
  mpDispatcher = pDispatcher;
  mCurrentMessage.str(string());
  
  mModuleName = module_name;
  mMessageId = 0;
  mIsMessageOpened = false;
}

StringMessageManager::~StringMessageManager()
{
  //:TODO: the distruction of the list of messages
}

void StringMessageManager::NewMessage(unsigned int msg_id, string msg_type)
{
  //if we were writing another message, we close it and put it in the message queue
  CloseCurrentMessage();

  struct timeval timestamp;
  gettimeofday(&timestamp,0);
  
  //writing the message header 
  
  mCurrentMessage << "<MESSAGE id = " << msg_type 
		  << " sender = " << mpDispatcher->GetHostName() 
		  << " timestamp = " << "_" << timestamp.tv_sec << "_" << timestamp.tv_usec 
		  << " module = " << mModuleName << ">\n";

  mMessageId = msg_id;
  mIsMessageOpened = true;
}

void StringMessageManager::SendStringMessage(unsigned int message)
{
  //sending of all the messages of type "message" that are queued
  for (unsigned int i = 0; i < mMessages[message].size(); i++)
    {
#ifdef MESSAGE_DEBUG
      cout << "\nSent\n" << mCurrentMessage.str()<<endl;
#endif
      SendMessage((unsigned char*)mMessages[message][i].c_str(),
		  (unsigned short)(mMessages[message][i].size()*sizeof(char)),
		  (unsigned short)message);
    }

  //erasing from the queue all the messages of type "message"
  mMessages[message].clear();

  //delete the message content
  //mCurrentMessage.str(string());
}

void StringMessageManager::SendAllStringMessages()
{
  map<unsigned int, vector<string> >::iterator it;
  for (it = mMessages.begin(); it != mMessages.end(); it++)
    {
      SendStringMessage(it->first);
    }
}

void StringMessageManager::CloseCurrentMessage()
{
  //if we are writing a message, we have to close it and put it in the list of messages
  if ( mIsMessageOpened )
    {
      mCurrentMessage << "</MESSAGE>\n" << '\0';
      mMessages[mMessageId].push_back(mCurrentMessage.str());

      // Empty the message string
      mCurrentMessage.str(string());
      mIsMessageOpened = false;
    }
}

void StringMessageManager::ReceiveStringMessage(unsigned short message)
{
  unsigned char* buffer = 0;
  while (ReceiveMessage(message,buffer))
    {
#ifdef MESSAGE_DEBUG
  cout << "\nReceived\n" << buffer << endl;
#endif 
      mReceivedMessages.push_back(buffer);
    }
}

void StringMessageManager::ReceiveAllStringMessages()
{
  ReqMessagesList::iterator req_it;
  for (req_it = ReqMessages.begin(); req_it != ReqMessages.end(); req_it++)
    {
      ReceiveStringMessage((unsigned short)(*req_it));
    }
}

unsigned char* StringMessageManager::GetStringMessage(unsigned int num)
{
#ifdef MESSAGE_DEBUG
  cout << "\nReading\n" << mReceivedMessages[num] << endl;
#endif 
  return mReceivedMessages[num];
}

unsigned char* StringMessageManager::GetLastStringMessage(unsigned short message)
{
  unsigned char *buffer;
  ReceiveLastMessage( message, buffer );
  return ( buffer );
}

unsigned int StringMessageManager::GetNumberOfReceivedMessages()
{
  return mReceivedMessages.size();
}

void StringMessageManager::CancelAllStringMessages()
{
  for (unsigned int i = 0; i < mReceivedMessages.size(); i++)
    {
      delete[] mReceivedMessages[i];
    }
  mReceivedMessages.clear();
}
