#include <GestioneConfigurazione.h>
#include <sstream>
using namespace std;

string 	GestioneConfigurazione::devSerial;
string 	GestioneConfigurazione::devJoypad;
string 	GestioneConfigurazione::calibJoypad;
string 	GestioneConfigurazione::fileHokuyoL;
string 	GestioneConfigurazione::fileHokuyoR;
string 	GestioneConfigurazione::dcdtConfFile;
string 	GestioneConfigurazione::brianctof;
string 	GestioneConfigurazione::brianshapectof;
string 	GestioneConfigurazione::brianpredicate;
string 	GestioneConfigurazione::brianpredicateactions;
string 	GestioneConfigurazione::briancando;
string 	GestioneConfigurazione::brianbehaviour;
string 	GestioneConfigurazione::brianwant;
string 	GestioneConfigurazione::briansftoc;
string 	GestioneConfigurazione::briansftocshape;
string 	GestioneConfigurazione::audioFile;
string 	GestioneConfigurazione::artoolkitCalibFile;
double 	GestioneConfigurazione::artoolkitTagDimension;
string 	GestioneConfigurazione::matrixInversionParameterFile;
string 	GestioneConfigurazione::absoluteCoordinateTagRelationFile;
string 	GestioneConfigurazione::absoluteCoordinateWheelchairToCameraRelationFile;
string 	GestioneConfigurazione::tagRelationFileTmp;
string 	GestioneConfigurazione::tagRelationFileMeanTmp;
string 	GestioneConfigurazione::logFile;
string 	GestioneConfigurazione::sequencerConfigFile;
string 	GestioneConfigurazione::imageNamePrefix;
string 	GestioneConfigurazione::imageNameSubfix;
string 	GestioneConfigurazione::imageNameFormat;
string 	GestioneConfigurazione::saveImageNamePrefix;
string 	GestioneConfigurazione::saveImageNameSubfix;
string 	GestioneConfigurazione::saveImageNameFormat;
bool GestioneConfigurazione::cameraSimulation;
bool GestioneConfigurazione::saveFrame;
string 	GestioneConfigurazione::spikeConfigFile;
string 	GestioneConfigurazione::matrixInversionParameterFileTmp;
string 	GestioneConfigurazione::matrixSdrAbsoluteWrtTagTmp;
string GestioneConfigurazione::poseConfigFile;
string GestioneConfigurazione::imuDeviceName;
string GestioneConfigurazione::imuConfigFile;
string GestioneConfigurazione::matrixLogParameterFile;
string GestioneConfigurazione::fireIdevice;


int GestioneConfigurazione::init(string path,int ac,char *av[]){
	using namespace po;
	options_description generic("Generic options");
	options_description devConfig("Device Configuration");
	options_description cmdline_options;
	options_description config_file_options;
	options_description visible("Allowed options");
	variables_map vm;	

	try {
		generic.add_options()
		    ("help,h", "produce help message");
	    
		// Declare a group of options that will be 
		// allowed both on command line and in
		// config file
		devConfig.add_options()
		    ("serial-dev,s", 	value<string>(&devSerial)->default_value("/dev/ttyUSB0"), 					"device for wheelchair-pc communication")
		    ("joypad-dev,j", 	value<string>(&devJoypad)->default_value("/dev/input/js0"),         				"device for joypad")
		    ("joypad-calib,c", 	value<string>(&calibJoypad)->default_value("./config/joypad/joypad_calibration"),         	"Calibration file for joypad")
		    ("hokuyo-left,l", 	value<string>(&fileHokuyoL)->default_value("./config/hokuyo/hokuyo_left"),			"file for left  Hokuyo URG laser")
		    ("hokuyo-right,r", 	value<string>(&fileHokuyoR)->default_value("./config/hokuyo/hokuyo_right"),			"file for right Hokuyo URG laser")
		    ("dcdt-cfg-file", value<string>(&dcdtConfFile)->default_value("./config/dcdt/dcdt.conf"),      			"file for dcdt")
		    ("brian-ctof-file", value<string>(&brianctof)->default_value("./config/brian/ctof.txt"),      			"file for configure brian input data")
		    ("brian-ctof-shape-file", value<string>(&brianshapectof)->default_value("./config/brian/shape_ctof.txt"),      	"file for configure brian shape input data")
		    ("brian-predicate-file", value<string>(&brianpredicate)->default_value("./config/brian/Predicate.ini"),      	"file for configure brian predicate")
		    ("brian-predicate-actions-file", value<string>(&brianpredicateactions)->default_value("./config/brian/PredicateActions.ini"),"file for configure brian predicate actions")
		    ("brian-cando-file", value<string>(&briancando)->default_value("./config/brian/Cando.ini"),      			"file for configure brian cando")
		    ("brian-behaviour-file", value<string>(&brianbehaviour)->default_value("./config/brian/behaviour.txt"),      	"file for configure brian behaviour")
		    ("brian-want-file", value<string>(&brianwant)->default_value("./config/brian/want.txt"),      			"file for configure brian want")
		    ("brian-sftoc-file", value<string>(&briansftoc)->default_value("./config/brian/s_ftoc.txt"),      			"file for configure brian output data")
		    ("brian-sftoc-shape-file", value<string>(&briansftocshape)->default_value("./config/brian/s_shape.txt"),      	"file for configure brian output shape")
		    ("audio-config-file",value<string>(&audioFile)->default_value("./config/audio/audioconfig.ini"),      		"file for configure audio ")
		    ("artoolkit-calib-file",value<string>(&artoolkitCalibFile)->default_value("./config/camera/fireI400wide.cal"),      "file for configure camera")
		    ("artoolkit-tag-dimension",value<double>(&artoolkitTagDimension)->default_value(160),      			"tag dimension (mm)")
		    ("matrix-inversion-parameter-file",value<string>(&matrixInversionParameterFile)->default_value("./config/camera/matrixInvParameter.ini"),"parameter for matrix inversion")
		    ("matrix-inversion-parameter-file-tmp",value<string>(&matrixInversionParameterFileTmp)->default_value("./config/camera/matrixInvParameter.tmp"),"parameter for matrix inversion(tmp)")
    		    ("absolute-coordinate-tag-relation-file",value<string>(&absoluteCoordinateTagRelationFile)->default_value("./config/camera/tagRelation.ini"),"tag to absolute")
    		    ("absolute-coordinate-camera-position-file",value<string>(&absoluteCoordinateWheelchairToCameraRelationFile)->default_value("./config/camera/cameraToWheelchair.ini"),"wheelchair to camera")
    		    ("tag-relation-file-tmp",value<string>(&tagRelationFileTmp)->default_value("./config/camera/tagRelation.tmp"),"tag to tag relation (tmp)")
    		    ("tag-relation-file-mean-tmp",value<string>(&tagRelationFileMeanTmp)->default_value("./config/camera/tagRelationMean.tmp"),"tag to tag relation mean(tmp)")
    		    ("matrix-sdr-absolute-wrt-tag-tmp-file",value<string>(&matrixSdrAbsoluteWrtTagTmp)->default_value("./config/camera/absSdrToTagRelation.tmp"),"absolute system to base tag rototraslation")

    		    ("log-file",value<string>(&logFile)->default_value("./log/logFile.log"),						"log file for delta time expert activation")
    		    ("sequencer-config-file",value<string>(&sequencerConfigFile)->default_value("./config/sequencer/sequencer.ini"),	"file config for sequencer")

		    ("camera-simulation",value<string>()->default_value("no"),"specify for simulation from file")
    		    ("image-name-prefix",value<string>(&imageNamePrefix)->default_value("./capture/test-"),	"prefix file image for camera simulation")
    		    ("image-name-subfix",value<string>(&imageNameSubfix)->default_value(""),	"subfix file image for camera simulation")
    		    ("image-name-format",value<string>(&imageNameFormat)->default_value("%0.10d"),	"format for image in camera simulation")

		    ("save-frame",value<string>()->default_value("no"),"specify for save frame captured")
    		    ("save-image-name-prefix",value<string>(&saveImageNamePrefix)->default_value("../../capture/saved-"),	"prefix file image for camera simulation")
    		    ("save-image-name-subfix",value<string>(&saveImageNameSubfix)->default_value(""),	"subfix file image for camera save")
    		    ("save-image-name-format",value<string>(&saveImageNameFormat)->default_value("%0.10d"),	"format for image in camera save")
    		    ("fire-i-device",value<string>(&fireIdevice)->default_value("/dev/video1394/0"),	"dev for camera")

    		    ("spike-config-file",value<string>(&spikeConfigFile)->default_value("./config/spike/mappa.ini"),	"spike config file")
			("pose-config-file",value<string>(&poseConfigFile)->default_value("./config/pose/pose.ini"), "pose config file")
			("imu-device",value<string>(&imuDeviceName)->default_value("./dev/ttyUSB1"), "imu device")
			("imu-config-file",value<string>(&imuConfigFile)->default_value("./config/imu/imu.ini"), "imu file")
			("matrix-log-parameter-file",value<string>(&matrixLogParameterFile)->default_value("./log/logParameter.dat"), "file for log of data from ")
		    ;
	       
		cmdline_options.add(generic).add(devConfig);

		config_file_options.add(devConfig);

		visible.add(generic).add(devConfig);
		
		//store(po::command_line_parser(ac, av).options(cmdline_options).run(), vm);
		store(parse_command_line(ac, av,cmdline_options), vm);
		ifstream ifs(path.c_str());
		store(parse_config_file(ifs, config_file_options), vm);

		notify(vm);
		if(vm.count("camera-simulation")==0){
			cameraSimulation=false;
		}	
		else{
			string str=vm["camera-simulation"].as<string >();
			if(str.compare("yes")==0){
				cameraSimulation=true;
			}
			else {
				if(str.compare("no")==0){
					cameraSimulation=false;
				}
				else{
					cout << "camera-simulation has to be 'yes' or 'no'\n";
					return -1;			
				}
			}			
		}

		if(vm.count("save-frame")==0){
			saveFrame=false;
		}
		else{
			string str=vm["save-frame"].as<string >();
			if(str.compare("yes")==0){
				saveFrame=true;
			}
			else {
				if(str.compare("no")==0){
					saveFrame=false;
				}
				else{
					cout << "save-frame has to be 'yes' or 'no'\n";
					return -1;			
				}
			}			
		}


	}
	catch (exception& e){
		cout << e.what() << "\n";
		return -1;
	}        
        if (vm.count("help")) {
		cout << visible;
	        return 1;
        }

        return 0;
}

string GestioneConfigurazione::toString(){
	ostringstream sd;
	sd<<artoolkitTagDimension;
	return devSerial + "\n" +devJoypad+ "\n" +calibJoypad+ "\n" +fileHokuyoL+ "\n" +fileHokuyoR+ "\n" +dcdtConfFile+ "\n" +brianctof+ "\n" + brianshapectof+ "\n" + brianpredicate + "\n" +brianpredicateactions+ "\n" + briancando + "\n" +brianbehaviour + "\n" +brianwant+ "\n" + briansftoc+ "\n" + briansftocshape+"\n" +audioFile+"\n"+artoolkitCalibFile +"\n"+sd.str() +"\n"+matrixInversionParameterFile+"\n"+ absoluteCoordinateTagRelationFile +"\n" + absoluteCoordinateWheelchairToCameraRelationFile+"\n\n" ;
}

