 /*
	ROBOT CONTROL SOFTWARE
	DONI STEFANO - FERRONI ALESSANDRO
	
	FUZZY CONTROLLER
	BRUSCHIERI STEFANO - CAMPARI PIERPAOLO
	
	PID CONTROLLER AND OTHER CHANGES
	ALEX PASQUALI

	VERSION 3
	LAST UPDATE 20-04-2006
	
*/

#include <stdlib.h>
#include <portb.h>
#include <timers.h>
#include <delays.h>
#include <string.h>
#include <pwm.h>

#include "definitions.h"
#include "controller.h"
#include "rs232.h"
#include "eeprom.h"	

//version of airboard
#define VERSION 3

//Global variables initialization. 

char parsingInputBuffer[DIM_INPUT_BUFFER]; //input buffer for string parsing
unsigned char address=0;// type_var=0;
unsigned char dataReady=0,inputBufferCounter=0, Command=0, Device=0, byteReceived = 0;
int InputData0=0, InputData1=0; //rs232 input data
int FuzzyData=0;
int disabled=1;
unsigned int number_cycles=0;

long total_tick_0=0, total_tick_1=0; //engine tick counters
unsigned char control_counter=0; //control counter 
unsigned int control_kick=0; //control kick conter
unsigned int timerout_kick=0; //timerout kick conter
int setpoint_0=0, setpoint_1=0; //engine 0 and engine 1 setpoints
int current_speed_0=0, current_speed_1=0; //engine 0 and engine 1 current speed

unsigned char debug=DEBUG_DEFAULT; //debug status

unsigned float ticks_to_pwm_0=0.0; //ticks to pwm for engine 0; it has calculated by PWM_MAX/max_ticks
unsigned float ticks_to_pwm_1=0.0; //ticks to pwm for engine 1; it has calculated by PWM_MAX/max_ticks

unsigned float ticks_to_space_0=0.0; //ticks to space for engine 0;
unsigned float ticks_to_space_1=0.0; //ticks to space for engine 1;

unsigned float ticks_to_speed_0=0.0, ticks_to_speed_1=0.0; //ticks to speed for both engines: they are calculated by 100/max_ticks

unsigned int max_tick_0=0, max_tick_1=0; //max ticks for both engines

unsigned char tuning = 0; //tuning counter

//PID controller parameters
unsigned float kp_0=0.0, kp_1=0.0; //proportional constants
unsigned float ki_0=0.0, ki_1=0.0; //integral constants
unsigned float kd_0=0.0, kd_1=0.0; //derivative constants

long already_read_ticks_0=0, already_read_ticks_1=0; //ticks already read

int s=0;
char string_temp[40]; //temporary string for various uses

//Fuzzy controller parameters
unsigned char is_fuzzy;
unsigned char num_of_rules;
int f_incr_max;

//local variable 
unsigned int timerout_kick_delay=0;
unsigned int control_kick_delay=0;
unsigned char control_counter_delay=0;
unsigned char kick_enable=TRUE;
unsigned char re_kick_0=FALSE;
unsigned char re_kick_1=FALSE;


//functions prototype
void setup(void);
void restoreFuzzyData(void);
void restoreData(void);
unsigned char executeCommand(void);
char* ftoa(unsigned float);


/*-------------------------------------------------------------------------------
----------- LOW PRIORITY INTERRUPT - CONTROL ------------------------------------
-------------------------------------------------------------------------------*/

#pragma interrupt control_isr

#pragma code InterruptVectorLow=0x0018

void InterruptVectorLow(void) {								
	_asm
		goto control_isr
	_endasm
}

#pragma code

void control_isr(void) {
	if (PIR1bits.TMR2IF) {
		//timer 2 interrupt
		//Let's see if we must reset kickers
		if ((KICK_0 == KICKER_ON) || (KICK_1 == KICKER_ON)) {
			if (--control_kick == 0) {
				KICK_0 = KICKER_OFF;
				KICK_1 = KICKER_OFF;
				if (debug >= DEBUG_MEDIUM) {
    				//print disabled if debug at low level
   					putrsUSART("\nKICKER_OFF\r");
				}
			}		
		}
		if ((KICK_0 == KICKER_OFF) && (KICK_1 == KICKER_OFF)&&(kick_enable==FALSE)) {
			if (--timerout_kick == 0){
				if (debug >= DEBUG_MEDIUM) {
    				//print disabled if debug at low level
   					putrsUSART("\nKICK ENABLED\r");
				}
				kick_enable=TRUE;
				//disable kickers after contro_kick_delay cicles
				timerout_kick = timerout_kick_delay;
				control_kick = control_kick_delay;

				if ((re_kick_0 == TRUE) || (re_kick_1 == TRUE)){
					if (re_kick_0 == TRUE){
						KICK_0 = KICKER_ON;
						re_kick_0 = FALSE;					
					}
					else if (re_kick_1 == TRUE){
						KICK_1 = KICKER_ON;
						re_kick_1 = FALSE;
					}
					kick_enable=FALSE;
					if (debug >= DEBUG_MEDIUM) {
    					//print disabled if debug at low level
   						putrsUSART("\nRE-KICK\r");
					}
				}
			}
		}

		control_counter++;
		//Let's see if we have to perform control routine
		if (control_counter >= control_counter_delay) {
			engine_control();
			//reset counter
			control_counter = 0;
		}
		
		//reset Timer2 interrupt flag
		PIR1bits.TMR2IF = 0;
	}
}

#pragma code

/*-------------------------------------------------------------------------------
----------- HIGH PRIORITY INTERRUPT - USART -------------------------------------
-------------------------------------------------------------------------------*/

#pragma interrupt rs232_isr

#pragma code InterruptVectorHigh=0x0008
void InterruptVectorHigh(void) {
	_asm
		goto rs232_isr
	_endasm
}

#pragma code
void rs232_isr(void) {
	if (PIR1bits.RCIF) {
			//if receive interrupt is high then read from rs232 bus
			read_rs232();
	}
}
	
#pragma code


/*-------------------------------------------------------------------------------
------------------------- COMMAND EXECUTION -------------------------------------
-------------------------------------------------------------------------------*/
//convert a float number in a string
char* ftoa(unsigned float value) { 
	char string[15];
	unsigned char i;
	
	//reset string
	for (i=0; i<15; i++)
		string[i]='\0';
		
	//check if value is integer
	if (value==(int)value) {
		//value_read is integer
		//print integer part
		itoa(value, string);
	}
	else {
		//value_read if float
		unsigned float float_part=0.0;
		unsigned int int_part = 0;
		unsigned char i=0;
									
		//retrieve integer and float part
		int_part = (unsigned int)value;
		float_part= value-int_part;
									
		//store integer part
		itoa(int_part, string);
		for(i=0; string[i]!='\0'; i++);
		//insert "dot"
		string[i++]='.';
		int_part=i;
									
		//retrieve float part
		for(i=0; i<((1<<EXP)-1); i++) {
			float_part = float_part*10;
			string[int_part++]=(unsigned int)float_part+'0';
			float_part = float_part-(unsigned int)float_part;
		}
		//insert string terminator
		string[int_part]='\0';
	}
	//return string with the value converted
	return string;
}

//retrieve input data from input string where len is the lenght of string and offset is the index of string
unsigned char retrieveInputData(unsigned char input, unsigned char len, unsigned char offset) {
	unsigned char sign, signOffset=1;
	char tempBuffer[9]; //temporary buffer for character storage
	int inputData;
	unsigned char i;
	
	if (len>9)
		return ERROR;
	
	sign = parsingInputBuffer[offset];
	if (sign != '+' && sign != '-') {
		if (sign<'0' || sign>'9')
			return ERROR;
		else
			signOffset=0;
	}
   
   	//retrieve inputdata
   	for (i=0; i<len; i++) {
   	tempBuffer[i] = parsingInputBuffer[offset+i+signOffset];
   	if (tempBuffer[i]<'0' || tempBuffer[i]>'9')
   		return ERROR;
   	}
   	tempBuffer[len] = '\0';
	
 	//convert temp buffer to number
   	inputData = atoi(tempBuffer);
	if (sign == '-')
   		inputData = -inputData;
   
   	//copy value to right input data 
   	if (input == 0) 
   		InputData0 = inputData;
   	else if (input == 1) 
   		InputData1 = inputData;
   	else if (input == 2) 
   		FuzzyData = inputData;
   	else 
   		return ERROR;
   
   	//operation ok
   	return OK;
}

//parse and execute command received from serial bus
unsigned char executeCommand(void) {
	// execute command received
	unsigned char byte_received=0;
	unsigned char error = ERROR;

	unsigned char fuzzy_membership=0;
	unsigned char fuzzy_table_offset=0; // this also mean 'input value' 

	char ok=FALSE;
	int temp_setpoint;
	//reset string_temp
	int i;
	for (i=0; i<40; i++) 
		string_temp[i]='\0';
	
	byte_received = byteReceived; 
   
	//retrieve Command and Device
   Command = parsingInputBuffer[0];
   Device = parsingInputBuffer[1]-'0';
   
   switch (Command) {
   			case 'b':	//disable engines
					error_to_zero(); //azzeramento degli errori nel controllore in fase di disabilitazione (NB se anche in abilitazione fa scattare il robot)
			case 'a': 	//enable engines
   					if (byte_received == 2) {
   						if ((Device != ENGINE_0) && (Device != ENGINE_1) && (Device != BOTH_ENGINES))
							return BAD_DEVC;
						if ((Device == ENGINE_0) || (Device == BOTH_ENGINES)) {
							if (Command == 'a') {
								//enable engine 0
								PIN_ENGINE_0 = ENGINE_ON;
								disabled=0;
							}
							else if (Command == 'b') {
								//disable engine 0
								PIN_ENGINE_0 = ENGINE_OFF;
							}
						}
						if ((Device == ENGINE_1) || (Device == BOTH_ENGINES)) {
							if (Command == 'a') {
								//enable engine 1
								PIN_ENGINE_1 = ENGINE_ON;
								disabled=0;
							}
							else if (Command == 'b') {
								//disable engine 1	
								PIN_ENGINE_1 = ENGINE_OFF;
							}
						}
						//operation ok
						return OK;
					}
					//operation failed
					return BAD_CMD;
					break;
		case 'm':	//set PWM
					if ((Device == ENGINE_0 || Device == ENGINE_1) && byte_received == 6) {
						//retrieve only first input data (inputData0)
						error = retrieveInputData(0,3,2);
						
						if (error != OK)
							return BAD_DATA;
							
						if (Device == ENGINE_0)
							setpoint_0 = (unsigned int) (InputData0*PWM_MAX/100);
						else if (Device == ENGINE_1)
							setpoint_1 = (unsigned int) (InputData0*PWM_MAX/100); // corretto
						else 
							return BAD_DEVC; //ERROR
							
						//operation ok
						return OK;
					}
					else if (Device == BOTH_ENGINES && byte_received == 10) {
						//retrieve both InputData0 and InputData1
						if (!retrieveInputData(0,3,2) || !retrieveInputData(1,3,6))
							return ERROR;
							
						//set new setpoints
						setpoint_0 = (unsigned int) (InputData0*PWM_MAX/100);
						setpoint_1 = (unsigned int) (InputData1*PWM_MAX/100);
	
						//operation ok
						return OK;
					}
					else 
						return BAD_CMD; //operation failed
					break;
		
		case 'e':  	{
							//write (ew) or read (er) data from eeprom
							char operation = Device + '0';
							int var_size; // variable name size 
							char variable[VARIABLE_SIZE];
							char sent_value[VALUE_SIZE];
							char * separator; // points to the equal sign, if exists
							
							//reset variable and sent_value
							for (address=0; address<VARIABLE_SIZE; address++)
								variable[address]=0;
							for (address=0; address<VALUE_SIZE; address++)
								sent_value[address]=0;
							
							//retrieve separator, that is char = (equal)
							separator = strchr((const char *)parsingInputBuffer, '=');
							if (separator != NULL) {
								// we should have a write operation
								if (operation != 'w') 
									return BAD_CMD;
								var_size =	separator - parsingInputBuffer - 2; //  corretta
							} else {
								// we should have a read operation
								if (operation != 'r') 
									return BAD_CMD;
								var_size = byte_received - 2;
							}
							
							if (parsingInputBuffer[2]=='f') var_size=8;
		
							// get variable name
							memcpy((char *)variable, (const char *)(parsingInputBuffer + 2), var_size);
							variable[var_size]='\0';
				
							// get value name			
							if (parsingInputBuffer[2]!='f') {
								//if not_fuzzy param (previous code)
								if (operation == 'w') {
									memcpy((char *)sent_value, (const char *)separator + 1, byte_received - (separator - parsingInputBuffer) - 1);
									sent_value[byte_received - (separator - parsingInputBuffer)-1] = '\0';
								}
							}
							else {
								//if fuzzy param
								if (operation == 'w') {
									//GET THE VALUE
									if (parsingInputBuffer[15]!='&')return BAD_CMD;
									error = retrieveInputData(2,3,11);
	 								if (error != OK)
										return BAD_DATA;
									fuzzy_membership=(unsigned char)FuzzyData;

									error = retrieveInputData(2,3,16);
	 								if (error != OK)
										return BAD_DATA;
									fuzzy_table_offset=(unsigned char)FuzzyData;

								}else{ //operation read
									if (parsingInputBuffer[10]!='&')return BAD_CMD;
									error = retrieveInputData(2,3,11);
	 								if (error != OK)
										return BAD_DATA;
									fuzzy_table_offset=(unsigned char)FuzzyData;
								}
								FuzzyData=0;
							}
																	
							// get variable eeprom address and store it in proper variable
							if (!strcmppgm2ram(variable, (rom char *)"kd0"))  {
								address = KD0_ADDR;
								if (operation == 'w')
									kd_0 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"ki0")) {
								address = KI0_ADDR;
								if (operation == 'w')
									ki_0 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"kp0")) {
								address = KP0_ADDR;
								if (operation == 'w')
									kp_0 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"kd1")) {
								address = KD1_ADDR;
								if (operation == 'w')
									kd_1 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"ki1")) {
								address = KI1_ADDR;
								if (operation == 'w')
									ki_1 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"kp1")) {
								address = KP1_ADDR;
								if (operation == 'w')
									kp_1 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"ct")) {
								address = CT_ADDR;
								if (operation == 'w')
									control_counter_delay = (unsigned char)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"ck")) {
								address = CK_ADDR;
								if (operation == 'w')
									control_kick_delay = (unsigned int)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"maxtc0")) {
								address = MAXTC0_ADDR;
								if (operation == 'w') {
									max_tick_0 = (unsigned int)atof(sent_value);
									//set ticks_to_pwm_0
									ticks_to_pwm_0 = (unsigned float)(PWM_MAX / max_tick_0);
									//set ticks_to_speed_0
									ticks_to_speed_0 = (unsigned float)(100.0 / max_tick_0);
								}
							}
							else if (!strcmppgm2ram(variable, (rom char *)"maxtc1")) {
								address = MAXTC1_ADDR;
								if (operation == 'w') {
									max_tick_1 = (unsigned int)atof(sent_value);
									//set ticks_to_pwm_1
									ticks_to_pwm_1 = (unsigned float)(PWM_MAX /max_tick_1);
									//set ticks_to_speed_1
									ticks_to_speed_1 = (unsigned float)(100.0 / max_tick_1);
								}
							}
							/*else if (!strcmppgm2ram(variable, (rom char *)"ticsp0")) {
								address = TICSP0_ADDR;
								if (operation == 'w')
									ticks_to_space_0 = (unsigned float)atof(sent_value);
							}
							else if (!strcmppgm2ram(variable, (rom char *)"ticsp1"))  {
								address = TICSP1_ADDR;
								if (operation == 'w')
									ticks_to_space_1 = (unsigned float)atof(sent_value);
							}*/
							else if (!strcmppgm2ram(variable, (rom char *)"fuzzymf1"))  {
								if (fuzzy_table_offset>=MF1_SIZE || fuzzy_table_offset<0)return BAD_DATA;
								address = MF1_B_ADDR + fuzzy_table_offset;

							}
							else if (!strcmppgm2ram(variable, (rom char *)"fuzzymf2"))  {
								if (fuzzy_table_offset>=MF2_SIZE || fuzzy_table_offset<0)return BAD_DATA;
								address = MF2_B_ADDR + fuzzy_table_offset;

							}
							else if (!strcmppgm2ram(variable, (rom char *)"fuzrules"))  {
								//here updating of rules
								if (fuzzy_table_offset>=num_of_rules || fuzzy_table_offset<0)return BAD_DATA;
								address = F_RULES_ADDR + fuzzy_table_offset;
							}
							else if (!strcmppgm2ram(variable, (rom char *)"frulenum"))  {
								//here updating number of rules
								address = F_RULES_SIZE_ADDR;
							}
							else if (!strcmppgm2ram(variable, (rom char *)"fisfuzzy"))  {
								//here choice between PID and FUZZY
								address = PID_XOR_FUZZY_ADDR;
							}
							else if (!strcmppgm2ram(variable, (rom char *)"fincrmax"))  {
								//here max voltage increment
								address = F_INCR_MAX_ADDR;
							}
							else if (!strcmppgm2ram(variable, (rom char *)"fsmalout"))  {
								//here small output singleton
								address = SMALL_CONS;
							}
							else if (!strcmppgm2ram(variable, (rom char *)"fhighout"))  {
								//here high output singleton
								address = HIGH_CONS;
							}
							else if (!strcmppgm2ram(variable, (rom char *)"tkd")) {
								address = RE_KICK_DEL_ADDR;
								if (operation == 'w')
									timerout_kick_delay = (unsigned float)atof(sent_value);
							}
                    		else if (!strcmppgm2ram(variable, (rom char *)"nc")) {
								address = NUM_CYC_ADDR;
								if (operation == 'w')
									number_cycles = (unsigned int)atoi(sent_value);
							}
							else 
								return BAD_CMD;
							
							//disable global interrupts //necessaria per ogni scrittura
							INTCONbits.GIE = 0;						
							
							// get/set variable value from/into eeprom
							if (parsingInputBuffer[2]=='f') { //fuzzy param
								if (operation == 'w') {	
									//operation is write
									//store the value in the internal eeprom
									writeByteToIE(address,fuzzy_membership);
									
									//write the result of the operation
									if (debug>=DEBUG_MEDIUM)
										putrsUSART("Write correct\r");

									//update some important parameters also in RAM
									if (address == F_RULES_SIZE_ADDR)
										if (fuzzy_membership + F_RULES_ADDR > 255){
											writeByteToIE(address,255-F_RULES_ADDR);
											putrsUSART("\nMax number of rules overcome\r");
										}
										num_of_rules=readByteFromIE(address);
									if (address == PID_XOR_FUZZY_ADDR)
										is_fuzzy=readByteFromIE(address);
									if (address == F_INCR_MAX_ADDR)
										f_incr_max=readByteFromIE(address);
									if (address == SMALL_CONS || address ==HIGH_CONS || address ==F_RULES_SIZE_ADDR ||(address >=F_RULES_ADDR && address<F_RULES_ADDR+num_of_rules ))
										restoreFuzzyData();
								}
								else {
									//operation is read
									unsigned char mfval_read;

									putsUSART(variable);
									WriteUSART('=');
								
									//retrieve the value from the eeprom
									mfval_read=readByteFromIE(address);
								
									//print the value
									for(s=0;s<40;s++){ string_temp[s]=' ';}
									putsUSART(itoa(mfval_read, string_temp));
									WriteUSART('\r');
								}
							}
							else {  //non-fuzzy param (previous version)

								if (operation == 'w') {
									//operation is write
									//store the value in the internal eeprom
									storeEEPromValue(address,atof(sent_value));
									//write the result of the operation
									if (debug>=DEBUG_MEDIUM)
										putrsUSART("Write correct\r");
								}
								else {
									//operation is read
									unsigned float value_read;

									putsUSART(variable);
									WriteUSART('=');
								
									//retrieve the value from the eeprom
									value_read=loadEEPromValue(address);
									putsUSART(ftoa(value_read));
									WriteUSART('\r');
								}
							}
				
							//enable global interrupts
							INTCONbits.GIE = 1;
							
							return OK;
							break;
					}
		case 'r':	//return distance or speed
					{
						char type;
						unsigned char i;
	
						// let's see if command is correct
						if (byte_received == 2) {
							// command should be "r2"
							if (parsingInputBuffer[1] != '2')
								return BAD_CMD;
						} 
						else if (byte_received == 3) {
							type = parsingInputBuffer[2];			
						} else 
							return BAD_CMD;
						
						// command type: 's' = speed, 'e' = space
						if (byte_received == 2 || type == 'e') {
							//calculate space
							// we have either "r2" of "r0e" or "r1e" or "r2e"  
							
							if ((Device == ENGINE_0) || (Device == BOTH_ENGINES)) {
								//refresh already_read_ticks for timer0
								already_read_ticks_0 = ReadTimer0();
								//reset timer 0
								WriteTimer0(0);
								
								if (DIR_ENGINE_0 == ENGINE_BACKWARD) {
									//if direction is backward I must invert the value
									already_read_ticks_0 = -already_read_ticks_0;
								}
								
								for (s=0;s<40;s++){string_temp[s]=' ';}

								//calculate space in cm
				   				//ltoa((long)((total_tick_0 + already_read_ticks_0) * ticks_to_space_0), string_temp);
								
				   				//calculate space in tick
				   				ltoa((long)(total_tick_0 + already_read_ticks_0), string_temp);
								
								putrsUSART("Tick_0: ");
				   				putsUSART(string_temp);
				  				putrsUSART("\n\r");
			   					
								total_tick_0 = 0;
				   			
				  				ok = TRUE;
				  			}
				   			
					   		if ((Device == ENGINE_1) || (Device == BOTH_ENGINES)) {
						   		//refresh already_read_ticks for timer1
				   				already_read_ticks_1 = ReadTimer1();
				   				//reset timer1
				   				WriteTimer1(0);
				   			
				   				if (DIR_ENGINE_1 == ENGINE_BACKWARD) {
					   				//if direction is backward I must invert the value
									already_read_ticks_1 = -already_read_ticks_1;
								}
								
								for (s=0;s<40;s++)string_temp[s]=0;

								//calculate space in cm
								//ltoa((long)((total_tick_1 + already_read_ticks_1) * ticks_to_space_1), string_temp);
											   				
								//calculate space in tick
				   				ltoa((long)(total_tick_1 + already_read_ticks_1), string_temp);

					   			putrsUSART("Tick_1: ");
				   				putsUSART(string_temp);
				  				putrsUSART("\n\r");

			   					total_tick_1 = 0;
				   			
			   					ok = TRUE;
			   				}						
						} // if (byte_received == 2 || type == 'e') 
						
						if (byte_received == 2 || type == 's') {
				   		//calculate speed
				   		
				   		if ((Device == ENGINE_0) || (Device == BOTH_ENGINES)) {
							for (s=0;s<40;s++)string_temp[s]=0;
				   			itoa(current_speed_0, string_temp);
				   				
				   			putrsUSART("Speed_0: ");
			   				putsUSART(string_temp);
			  				putrsUSART("\n\r");
				   			
				   			ok = TRUE;
				   		}
				   		if ((Device == ENGINE_1) || (Device == BOTH_ENGINES)) {
							for (s=0;s<40;s++)string_temp[s]=0;
				   			itoa(current_speed_1, string_temp);
				   							   						   			
				   			putrsUSART("Speed_1: ");
			   				putsUSART(string_temp);
				   			putrsUSART("\n\r");
				   			ok = TRUE;
				   		}
				   	} // if (byte_received == 2 || type == 's')
				   	
				   	if (ok) return OK;
				   		else return BAD_DEVC;
				   		
				   	break;
				   }
		case 'k':	//enable kickers
					//check if # of bytes is correct
					if (byte_received == 2) {
						if (Device == KICKER_0){
							if (kick_enable==TRUE){	
								//kick
								KICK_0 = KICKER_ON;
								//disable kickers re-activation
								kick_enable=FALSE;
								//disable kickers after contro_kick_delay cicles
								timerout_kick = timerout_kick_delay;
								control_kick = control_kick_delay;
							}else
								re_kick_0=TRUE;
						} else if (Device == KICKER_1){
							if (kick_enable==TRUE){
								//kick
								KICK_1 = KICKER_ON;
								//disable kickers re-activation
								kick_enable=FALSE;
								//disable kickers after control_kick_delay cicles
								timerout_kick = timerout_kick_delay;
								control_kick = control_kick_delay;
							}else 
								re_kick_1=TRUE;
						}	
						else 
							return BAD_DEVC;
					}
					else
						return BAD_CMD;
					return OK;
					break;

		case 'd': 	//enable or disable debug
						/*
							debug values:
							0: no print
							1: print few important message
							2: print some important message
							3: print all messages (all engine debug value)
						*/
						//check if # of bytes is correct
						if (byte_received == 2) {
							if (Device > DEBUG_FULL) 
								return BAD_CMD;
							debug = Device;
						return OK;
						}
						return ERROR;
						break;			
		case 'c': 	//quick-configuration with fixed values
					{		
						char value[VALUE_SIZE];

						//check if # of bytes is correct
						if (byte_received != 2)
								return BAD_CMD;
						
						//disable rs232 receive interrupt
						PIE1bits.RCIE = 0;
						
						if (Device == ENGINE_0 || Device == BOTH_ENGINES || Device == ENGINE_1) {
							if (debug>=DEBUG_MEDIUM)
								putrsUSART("Starting Quick configuration ");
							if (Device == ENGINE_0 || Device == BOTH_ENGINES){
								kp_0=0.5;
								value[0]='0';
								value[1]='.';
								value[2]='5';
								value[3]='\0';
								address=KP0_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								ki_0=0.1;
								value[0]='0';
								value[1]='.';
								value[2]='1';
								value[3]='\0';
								address=KI0_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								kd_0=0.3;
								value[0]='0';
								value[1]='.';
								value[2]='3';
								value[3]='\0';
								address=KD0_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								max_tick_0=240;
								value[0]='2';
								value[1]='4';
								value[2]='0';
								value[3]='\0';
								address=MAXTC0_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								ticks_to_pwm_0 = ((float)PWM_MAX / (float)max_tick_0);
								ticks_to_speed_0 = (100.0 / (float)max_tick_0);
							
								if (debug>=DEBUG_MEDIUM)
									putrsUSART("- - - - ");
							}
							if (Device == ENGINE_1 || Device == BOTH_ENGINES){
								kp_1=0.5;
								value[0]='0';
								value[1]='.';
								value[2]='5';
								value[3]='\0';
								address=KP1_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								ki_1=0.1;
								value[0]='0';
								value[1]='.';
								value[2]='1';
								value[3]='\0';
								address=KI1_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								kd_1=0.3;
								value[0]='0';
								value[1]='.';
								value[2]='3';
								value[3]='\0';
								address=KD1_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								max_tick_1=240;
								value[0]='2';
								value[1]='4';
								value[2]='0';
								value[3]='\0';
								address=MAXTC1_ADDR;
								INTCONbits.GIE = 0;
								storeEEPromValue(address,atof(value));
								INTCONbits.GIE = 1;
							
								ticks_to_pwm_1 = ((float)PWM_MAX / (float)max_tick_1);
								ticks_to_speed_1 = (100.0 / (float)max_tick_1);
							
								if (debug>=DEBUG_MEDIUM)
									putrsUSART("- - - - ");
							}
							
							control_counter_delay=10;
							value[0]='1';
							value[1]='0';
							value[2]='\0';
							address=CT_ADDR;
							INTCONbits.GIE = 0;
							storeEEPromValue(address,atof(value));
							INTCONbits.GIE = 1;
							
							if (debug>=DEBUG_MEDIUM)
									putrsUSART("- ");
							
							number_cycles=0;
							value[0]='0';
							value[1]='\0';
							address=NUM_CYC_ADDR;
							INTCONbits.GIE = 0;
							storeEEPromValue(address,atof(value));
							INTCONbits.GIE = 1;
							
							if (debug>=DEBUG_MEDIUM)
									putrsUSART("- ");
							
							control_kick_delay=200;
							value[0]='2';
							value[1]='0';
							value[2]='0';
							value[3]='\0';
							address=CK_ADDR;
							INTCONbits.GIE = 0;
							storeEEPromValue(address,atof(value));
							INTCONbits.GIE = 1;
							
							if (debug>=DEBUG_MEDIUM)
									putrsUSART("- ");
							
							timerout_kick_delay=200;
							value[0]='2';
							value[1]='0';
							value[2]='0';
							value[3]='\0';
							address=RE_KICK_DEL_ADDR;
							INTCONbits.GIE = 0;
							storeEEPromValue(address,atof(value));
							INTCONbits.GIE = 1;
							
							is_fuzzy=0;
							value[0]='0';
							value[1]='\0';
							address = PID_XOR_FUZZY_ADDR;
							INTCONbits.GIE = 0;
							storeEEPromValue(address,atof(value));
							INTCONbits.GIE = 1;
							
							if (debug>=DEBUG_MEDIUM)
								putrsUSART("End Configuration\r");
							
						}
						else
							return BAD_DEVC;

						//enable rs232 receive interrupt
						PIE1bits.RCIE = 1;

						return OK;
						break;
					}
		case 't':	//auto-tuning engine parameters
					{
						/*This feature allows to set automatically some important 
						parameters, such as max tick number and other constants.*/
						int i;
						
						//check if # of bytes received is correct						
						if (byte_received != 2)
							return BAD_CMD;
							
						//disable rs232 receive interrupt
						PIE1bits.RCIE = 0;
						
						//reset input data
						InputData0 = InputData1 = 0;
							
						//set max setpoints
						if (Device == ENGINE_0 || Device == BOTH_ENGINES) {
							//set setpoint and duty cicle max values for engine0
							setpoint_0 = PWM_MAX;
							SetDCPWM1(PWM_ZERO+PWM_MAX);
							//enable engine 0
							PIN_ENGINE_0 = ENGINE_ON;
						}
						if (Device == ENGINE_1 || Device == BOTH_ENGINES) {
							//set setpoint and duty cicle max values for engine1
							setpoint_1 = PWM_MAX;
							SetDCPWM2(PWM_ZERO+PWM_MAX);
							//enable engine 1
							PIN_ENGINE_1 = ENGINE_ON;
						}
						//wait for 4 sec and after start tuning
						for (i=0; i<16; i++)
							Delay10KTCYx(250); //4 seconds delay
						
						//set tuning value
						tuning = MAX_TEST;
						
						if (Device == ENGINE_0 || Device == BOTH_ENGINES) {
							//set duty cicle at max value for engine 0
							SetDCPWM1(PWM_ZERO+PWM_MAX);
						}
						if (Device == ENGINE_1 || Device == BOTH_ENGINES) {
							//set duty cicle at max value for engine 1
							SetDCPWM2(PWM_ZERO+PWM_MAX);
						}
						
						//wait for tuning value equals to MAX_TEST
						while(tuning > 0);
						if (Device == ENGINE_0 || Device == BOTH_ENGINES) {
							//disable engine 0
							PIN_ENGINE_0 = ENGINE_OFF;
							//set setpoint value to zero
							setpoint_0 = PWM_ZERO;
							//set duty cicle at zero value for engine 0
							SetDCPWM1(PWM_ZERO);
							//set values for engine0
							//set  max tick for engine_0 using InputData0
							max_tick_0 = (unsigned int)(InputData0);
							//save max tick 0 into internal eeprom
							storeEEPromValue(MAXTC0_ADDR, max_tick_0);
							//set ticks_to_pwm_0
							ticks_to_pwm_0 = ((float)PWM_MAX / (float)max_tick_0);
							//set ticks_to_speed_0
							ticks_to_speed_0 = (100.0 / (float)max_tick_0);
							//write result over rs232
							putrsUSART("\n\rMAX TICK_0= ");
							for (s=0;s<40;s++)string_temp[s]=0;
							putsUSART(itoa(max_tick_0, string_temp));
							putrsUSART("\n\r");
						}	
						if (Device == ENGINE_1 || Device == BOTH_ENGINES) {
							//disable engine 1
							PIN_ENGINE_1 = ENGINE_OFF;
							//set setpoint value to zero
							setpoint_1 = PWM_ZERO;
							//set duty cicle at zero value for engine 1
							SetDCPWM2(PWM_ZERO);
							//set values for engine1
							//set max ticks for engine_1 using InputData1
							max_tick_1 = (unsigned int)(InputData1);	
							//save max tick 0 into internal eeprom
							storeEEPromValue(MAXTC1_ADDR,max_tick_1);
							//set ticks_to_pwm_1
							ticks_to_pwm_1 = ((float)PWM_MAX / (float)max_tick_1);
							//set ticks_to_speed_1
							ticks_to_speed_1 = (100.0 / (float)max_tick_1);
							//write result over rs232
							putrsUSART("\n\rMAX TICK_1= ");
							putsUSART(itoa(max_tick_1, string_temp));
							putrsUSART("\n\r");
						}
						//enable serial interrupt
						PIE1bits.RCIE = 1;
											
						return OK;
						break;
					}
		case 'z':	//only for testing purposes
							
							INTCONbits.GIE = 0;			
													
							if(byte_received>9)
								return BAD_DATA;

							error = retrieveInputData(2,3,1);
	 						if (error != OK)
								return BAD_DATA;
							temp_setpoint = FuzzyData;
							
							error = retrieveInputData(2,3,5);
	 						if (error != OK)
								return BAD_DATA;
							
							FuzzyController((temp_setpoint*PWM_MAX)/100,(FuzzyData*PWM_MAX)/100);
							for (s=0;s<40;s++)string_temp[s]=0;
							putrsUSART("\n\r Fuzzy Out(incr): ");
							putsUSART(itoa( FuzzyOut, string_temp));
							putrsUSART("/");
							for (s=0;s<40;s++)string_temp[s]=0;
							putsUSART(itoa(((int)f_incr_max), string_temp));
							putrsUSART("\n\r");

							FuzzyData=0;
							INTCONbits.GIE = 1;	
							return OK;
							break;

		default:		//command error
						return UKNOWN_CMD;
	}
	//restart watchdog
	ClrWdt();


	//reset parsingInputBuffer
	parsingInputBuffer[0] = '\0';

	//reset input buffer
	for (i=0; i<DIM_INPUT_BUFFER; i++) 
		parsingInputBuffer[i]=0;
	
	//reset serial bus variables
	Command = Device = InputData0 = InputData1 = 0;
	
	//reset dataready
	dataReady = 0;
	
	//return success
	return OK;
}

/*-------------------------------------------------------------------------------
----------- PIC SETUP -----------------------------------------------------------
-------------------------------------------------------------------------------*/

//PIC setup
void Setup(void) {
	unsigned int i; //local counter variable
	
	//enable PORTA
	ADCON1 = 0x07; //PORTA all digital
	PORTA = 0x00; //Clear PORTA
	TRISA = 0b00010000; //PORTA=1 is input, PORTA=0 is output
	
	//enable PORTB
	EnablePullups();	// Enable PORTB pullups
	PORTB = 0x00;	//Clear PORTB
	TRISB = 0b00001010;	//PORTB all outputs.
	
	//enable PORTC
	PORTC = 0x00; //Clear PORTC
	TRISC=0b11000001; //PORTC=1 is input, PORTC=0 is output
	
	//configure PWM1 and PWM2 frequence at 20KHz
	/*
		if you change PWM frequence, you must recalculate all constants, such as PWM_MAX.
		PWM frequency at 20KHz allow simple calculations and it is out of the humans audibile frequence
	*/
	OpenPWM1(0x7c);
	OpenPWM2(0x7c);
	
	//configure Timer0: engine 0 tick counter
	/*
		Timer0 interrupt on
		Timer0 at 16 bit
		Timer0 external source
		Timer0 ad edge fall
			Timer0 prescaler 1:1
	*/
	OpenTimer0(TIMER_INT_ON & T0_16BIT & T0_SOURCE_EXT & T0_EDGE_FALL & T0_PS_1_1);
	
	//configure Timer1: engine 1 tick counter
	/*
		Timer1 interrupt on
		Timer1 at 16 bit with less significative bit right
		Timer1 external source
		Timer1 prescaler 1:1
		Timer1 oscillator off
		Timer1 external synch off
	*/
	OpenTimer1(TIMER_INT_ON & T1_16BIT_RW & T1_SOURCE_EXT & T1_PS_1_1 & T1_OSC1EN_OFF & T1_SYNC_EXT_OFF);
	
	//configure Timer2: PWM generator
	/*
		Timer2 interrupt on
		Timer2 prescaler 1:4
		Timer2 postscaler 1:10
	*/
	OpenTimer2(TIMER_INT_ON & T2_PS_1_4 & T2_POST_1_10);
			
	//configure USART 
	/*
		Value to use for SPBRG
		64 for baud rate = 38400
		21 for baud rate = 115200 
	*/
	SPBRG = 21; //baud rate
	TXSTA = 0x24; //transmit register
	RCSTA = 0x90; //receive register
	PIE1bits.RCIE = 1; //enable receive interrupt
	
	//reset input buffer
	for (i=0; i<DIM_INPUT_BUFFER; i++) 
		parsingInputBuffer[i]=0;
		
	//reset serial bus variables 
	dataReady=Command=Device=byteReceived = 0;
	
	//disable kickers
	KICK_0 = KICKER_OFF; //kicker 0 off
	KICK_1 = KICKER_OFF; //kicker 1 off
	
	//disable engines
	PIN_ENGINE_0 = ENGINE_OFF; //engine 0 disabled
	PIN_ENGINE_1 = ENGINE_OFF; //engine 1 disabled
	disabled=1;

	//configure interrupt priorities
	IPR1bits.RCIP = 1; //USART receive interrupt priority high
	IPR1bits.TMR2IP = 0; //TMR2 interrupt priority low
	RCONbits.IPEN = 1; //enable interrupt priorities	
	
	//enable interrupts
	INTCONbits.PEIE = 1;	// Enable peripheral interrupts
	INTCONbits.GIEH = 1; // Enable all interrupts
}


void restoreFuzzyData(void){
	int i;
	num_of_rules=readByteFromIE(F_RULES_SIZE_ADDR);

	for(i=0;i<num_of_rules;i++){
		rules[i]=readByteFromIE(F_RULES_ADDR+i);

   		switch (rules[i]&0x07) {
   			case 0:{
					rules_conseguent[i]=-100;
					break;}
   			case 1:{
					rules_conseguent[i]= 0-readByteFromIE(HIGH_CONS);
					break;}
   			case 2:{
					rules_conseguent[i]= 0-readByteFromIE(SMALL_CONS);
					break;}
   			case 3:{
					rules_conseguent[i]=0;
					break;}
   			case 4:{
					rules_conseguent[i]= readByteFromIE(SMALL_CONS);
					break;}
   			case 5:{
					rules_conseguent[i]=readByteFromIE(HIGH_CONS);
					break;}
   			case 6:{
					rules_conseguent[i]=100;
					break;}
			default:{
					rules_conseguent[i]=0;
					break;}		
		}			
	}
}


//load default robot parameters from internal eeprom
void restoreData(void) {
	//disable global interrupts
	INTCONbits.GIE = 0;
	
	//load pid constants
	kp_0 = loadEEPromValue(KP0_ADDR);
	ki_0 = loadEEPromValue(KI0_ADDR);
	kd_0 = loadEEPromValue(KD0_ADDR);
	kp_1 = loadEEPromValue(KP1_ADDR);
	ki_1 = loadEEPromValue(KI1_ADDR);
	kd_1 = loadEEPromValue(KD1_ADDR);

	//load fuzzy parameters
	//num_of_rules=readByteFromIE(F_RULES_SIZE_ADDR);
	is_fuzzy=readByteFromIE(PID_XOR_FUZZY_ADDR);
	f_incr_max=readByteFromIE(F_INCR_MAX_ADDR);

	restoreFuzzyData();

	//load cycles number of secur timer
	number_cycles= (unsigned int) loadEEPromValue(NUM_CYC_ADDR);

	//load re-kick delay
	timerout_kick_delay= (unsigned int)loadEEPromValue(RE_KICK_DEL_ADDR);

	//load sampling period
	control_counter_delay = (unsigned char)loadEEPromValue(CT_ADDR);
	
	//load kick delay
	control_kick_delay = (unsigned int)loadEEPromValue(CK_ADDR);

	//load max ticks
	max_tick_0 = (unsigned int)loadEEPromValue(MAXTC0_ADDR);
	max_tick_1 = (unsigned int)loadEEPromValue(MAXTC1_ADDR);
	
	//load ticks_to_space
	//ticks_to_space_0 = loadEEPromValue(TICSP0_ADDR);
	//ticks_to_space_1 = loadEEPromValue(TICSP1_ADDR);
	
	//load ticks_to_pwm values
	ticks_to_pwm_0 = ((float)PWM_MAX / (float)max_tick_0);
	ticks_to_pwm_1 = ((float)PWM_MAX / (float)max_tick_1);

	//load ticks_to_speed values
	ticks_to_speed_0 = (100.0 / (float)max_tick_0);
	ticks_to_speed_1 = (100.0 / (float)max_tick_1);
	
	//enable global interrupts
	INTCONbits.GIE = 1;
							
}
	

/*-------------------------------------------------------------------------------
----------- MAIN PROGRAM -----------------------------------------------------------
-------------------------------------------------------------------------------*/

//main program
void main(void) {
	unsigned int i;
	unsigned int secur_timer0=0;
	unsigned int secur_timer1=0;
	unsigned int secur_timer2=0;
	unsigned int secur_timer3=0;
	unsigned int secur_timer4=0;
	unsigned int secur_timer5=0;
	unsigned int secur_timer6=0;
	unsigned int secur_timer7=0;
	unsigned int secur_timer8=0;
	unsigned int secur_timer9=0;

	unsigned int status;
	
	//restart watchdog
 	ClrWdt();
	
	//setup peripherals and software
	Setup();

	//Restore default data from eeprom
	restoreData(); 
		
	//restart watchdog
	ClrWdt();

	// welcome message (da eliminare)
/*	putrsUSART("\nWELCOME IN AIRBOARD v"); 
	putsUSART(itoa(VERSION,string_temp)); 
	putrsUSART("\n\nUSART Ready. Waiting for commands\n\n");
*/
	//set PWM to zero => speed=0
	SetDCPWM1(PWM_ZERO);
	SetDCPWM2(PWM_ZERO);	

	//start control loop
	while(1) {
		
		//Security timer for disable the engine if there is no command for too much time
		if ((number_cycles!=0)
			&&(secur_timer0>=number_cycles)
			&&(secur_timer1>=number_cycles)
			&&(secur_timer2>=number_cycles)
			&&(secur_timer3>=number_cycles)
			&&(secur_timer4>=number_cycles)
			&&(secur_timer5>=number_cycles)
			&&(secur_timer6>=number_cycles)
			&&(secur_timer7>=number_cycles)
			&&(secur_timer8>=number_cycles)
			&&(secur_timer9>=number_cycles)
			&&(disabled==0))
		{
			PIN_ENGINE_0 = ENGINE_OFF;
			PIN_ENGINE_1 = ENGINE_OFF;
			disabled=1;
   			error_to_zero();
			if (debug >= DEBUG_MEDIUM) {
    			//print disabled
   				putrsUSART("\nENGINE DISABLED BY SECUR_TIMER\n\r");
			}
		}

		ClrWdt();

		//check if new command is arrived from serial bus		
		if (dataReady) {
			
			//clear number of cycles with no command
			secur_timer0=0;
			secur_timer1=0;
			secur_timer2=0;
			secur_timer3=0;	
			secur_timer4=0;
			secur_timer5=0;
			secur_timer6=0;
			secur_timer7=0;
			secur_timer8=0;	
			secur_timer9=0;

			//a new command has arrived
			dataReady=0;
			
			// start monitoring procedure length			
			status = executeCommand();

			//reset serial data
			Command = Device = InputData0 = InputData1 = 0;

			if (status != OK) {
				if(INTCONbits.GIE == 0)INTCONbits.GIE = 1;

    			//error has occurred
    			if (debug >= DEBUG_SMALL) {
	    			//print error value
    				putrsUSART("\nERROR ");
					for (s=0;s<40;s++)string_temp[s]=0;
					itoa(status, string_temp);
					putsUSART(string_temp);
					putrsUSART("\n\r");
					putrsUSART("\nDEBUG: ");
					putsUSART(parsingInputBuffer);
					putrsUSART("\n\r");
					//reset parsing input buffer
					for (i=0; i<DIM_INPUT_BUFFER; i++) 
						parsingInputBuffer[i]=0;
    			}
    		}
    		else {
    			//parsing success
    			if (debug >= DEBUG_SMALL) {
	    			//print only if debug value isn't equal to debug small (small informations)
    				putrsUSART("\nOK\n\r");
    			}
    		}
    	}//if (dataready)
		
		if(secur_timer0<=number_cycles)
			secur_timer0++;
		else if (secur_timer1<=number_cycles)
			secur_timer1++;
		else if (secur_timer2<=number_cycles)
			secur_timer2++;
		else if (secur_timer3<=number_cycles)
			secur_timer3++;
		else if (secur_timer4<=number_cycles)
			secur_timer4++;
		else if (secur_timer5<=number_cycles)
			secur_timer5++;
		else if (secur_timer6<=number_cycles)
			secur_timer6++;
		else if (secur_timer7<=number_cycles)
			secur_timer7++;
		else if (secur_timer8<=number_cycles)
			secur_timer8++;
		else if (secur_timer9<=number_cycles)
			secur_timer9++;

	}
}
