/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Panel;
import java.awt.Button;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JLabel;
import javax.swing.Icon;
import javax.swing.ImageIcon;


/**
 * @author Bruschieri Stefano
 */
public class AboutFrame extends Frame implements ActionListener{
	
    final int HEIGHT = 150;
    final int WIDTH  = 300;
	private Button closeButton;
	
	/**
	 *  Simply creates a Frame with some information about authors of
	 *  <code>AIRboardUI</code>.
	 */
	public AboutFrame(){
		super();
		
		Panel ovestPanel = new Panel();	
		Panel eastPanel = new Panel();
		ovestPanel.add(new JLabel(getIconByString("images/pmairLogo.gif")));
		eastPanel.setLayout(new GridLayout(4, 1));
		eastPanel.add(new Label("Bruschieri Stefano"));
		eastPanel.add(new Label("  sb666337@stud.polimi.it"));
		eastPanel.add(new Label("Campari PierPaolo"));
		eastPanel.add(new Label("  pc666272@stud.polimi.it"));
		this.add(eastPanel, "East");
		this.add(ovestPanel, "West");
		
		Panel southPanel = new Panel();
		this.closeButton = new Button("Close");
		this.closeButton.addActionListener(this);
		southPanel.add(closeButton);
		
		this.add(southPanel, "South");
		
		// set dimension of frame
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		setLocation(screenSize.width/2 - WIDTH/2, 
			    screenSize.height/2 - HEIGHT/2);

		setSize(WIDTH, HEIGHT);
		this.setVisible(true);
		this.repaint();
	}
	
    private Icon getIconByString(String path){
    	java.net.URL imgURL = AboutFrame.class.getResource(path);
    	if (imgURL != null) {
    	    return new ImageIcon(imgURL);
    	} else {
    	    System.err.println("Couldn't find file: " + path);
    	    return null;
    	}
        } 
	
    /**
    * Responds to the Close button and closes the window.
    */
    public void actionPerformed(ActionEvent e) {
    	String cmd = e.getActionCommand();
    	
    	//Close this window
    	if(cmd.equals("Close")){
    		this.setVisible(false);
    	}
    }

	

}
