/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */
package airboardUI;

import java.awt.Button;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Vector;
import java.awt.Font;
import javax.comm.SerialPort;
import javax.swing.Timer;
import java.awt.Color;

/**
 * @author Bruschieri Stefano
 */
public class PIDForm extends Frame implements ActionListener {
	
	private static final int KP0 = 0;
	private static final int KI0 = 1;
	private static final int KD0 = 2;
	
	private static final int KP1 = 3;
	private static final int KI1 = 4;
	private static final int KD1 = 5;
	
    private int MAX_ERRORS = 5;
    private int MAX_NO_RESPONSE = 5;
    
	private static final int WIDTH = 300;
	private static final int HEIGHT = 290;
    private SerialConnection connection;
	private AirboardFrame parent;
    
	private Panel m0Panel;
	private Panel m1Panel;
    private Button writeButton;
    private TextField inputOneTextField;
    private TextField inputTwoTextField;
    private TextField inputThreeTextField;
    private TextField inputForTextField;
    private TextField inputFiveTextField;
    private TextField inputSixTextField;
	
    private Vector commands;
  
    ProgressDialog pd;
    Timer timer;
    int i;
	int noResponseCounter = 0;
	int errorsCounter = 0;
    
	public PIDForm(SerialConnection connection, AirboardFrame parent){
		super();
		this.connection = connection;
		this.parent = parent;
		addWindowListener(new CloseHandler(this));	
		
		m0Panel = new Panel();
		m1Panel = new Panel();
		writeButton = new Button("Write");
		writeButton.addActionListener(this);
		inputOneTextField = new TextField(20);
		inputOneTextField.setEditable(true);
		inputTwoTextField = new TextField(20);
		inputTwoTextField.setEditable(true);
		inputThreeTextField = new TextField(20);
		inputThreeTextField.setEditable(true);
		inputForTextField = new TextField(20);
		inputForTextField.setEditable(true);
		inputFiveTextField = new TextField(20);
		inputFiveTextField.setEditable(true);
		inputSixTextField = new TextField(20);
		inputSixTextField.setEditable(true);

		GridBagLayout gridBag = new GridBagLayout();
		GridBagConstraints cons = new GridBagConstraints();
		cons.gridwidth = GridBagConstraints.REMAINDER;
		cons.weightx = 1.0;
		m0Panel.setLayout(new GridLayout(4,2));
		m0Panel.add(new Label("Derivative costant:"));
		m0Panel.add(inputOneTextField);
		m0Panel.add(new Label("Integrative costant:"));
		m0Panel.add(inputTwoTextField);
		m0Panel.add(new Label("Proportional costant:"));
		m0Panel.add(inputThreeTextField);
		Label m0Label = new Label("Motor 1", Label.LEFT);
		m0Label.setFont(new Font("SansSerif", Font.BOLD, 13));
		m0Label.setForeground(new Color(255,0,0));
		gridBag.setConstraints(m0Label , cons);
		add(m0Label);
		gridBag.setConstraints(m0Panel, cons);
		add(m0Panel);
		
		m1Panel.setLayout(new GridLayout(4,2));
		m1Panel.add(new Label("Derivative costant:"));
		m1Panel.add(inputForTextField);
		m1Panel.add(new Label("Integrative costant:"));
		m1Panel.add(inputFiveTextField);
		m1Panel.add(new Label("Proportional costant:"));
		m1Panel.add(inputSixTextField);
		Label m1Label = new Label("Motor 2", Label.LEFT);
		m1Label.setFont(new Font("SansSerif", Font.BOLD, 13));
		m1Label.setForeground(new Color(255,0,0));
		gridBag.setConstraints(m1Label , cons);
		add(m1Label);
		gridBag.setConstraints(m1Panel, cons);
		add(m1Panel);
		add(writeButton, "Center");
		this.setLayout(gridBag);
		
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		this.setLocation(screenSize.width/2 - WIDTH/2, 
			    screenSize.height/2 - HEIGHT/2);
		this.setSize(WIDTH, HEIGHT);
		this.setVisible(true);
		this.repaint();	
	}
		
	/**
	 * The specified String <code>text</code> is converted to a double, if
	 * this number is out of allowed range or its format is not correct, this
	 * function simply return a string with no lenght.
	 * 
	 * @param text The value of parameter 
	 * @param type The type of command defined in this class
	 * @return The command to send to the PIC
	 */
	private String prepareString(String text, int type){
		String toReturn = "";
		
    	/* integer value of text must be >0*/
		try{
			if( (new Double(text)).doubleValue()<0 /*|| 
			(new Double(text)).doubleValue()>255*/ ){
				System.out.println("Errore: value out of range "+Integer.valueOf(text));
				return "";
			}
		} catch(NumberFormatException nfe){
			return "";
		}
    	
    	System.out.println("PIDFOrm.prepareString "+(new Double(text)).doubleValue()+" "+text);
    	
    	if( type == PIDForm.KP0 )
    		toReturn = "ewkd0="+text;
    	if( type == PIDForm.KI0 )
    		toReturn = "ewki0="+text;
    	if( type == PIDForm.KD0 )
    		toReturn = "ewkd0="+text;
    	if( type == PIDForm.KP1 )
    		toReturn = "ewkd1="+text;
    	if( type == PIDForm.KI1 )
    		toReturn = "ewki1="+text;
    	if( type == PIDForm.KD1 )
    		toReturn = "ewkd1="+text;
		return toReturn;
	}
	
    /**
    * Responds to the buttons.
    */
    public void actionPerformed(ActionEvent e) {
    	String cmd = e.getActionCommand();
    
    	if( cmd.equals("Write") ){
    		this.commands = new Vector();
    		String s="";
    		if( this.inputOneTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputOneTextField.getText(), PIDForm.KD0);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}

    		if( this.inputTwoTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputTwoTextField.getText(), PIDForm.KI0);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}
    		
    		if( this.inputThreeTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputThreeTextField.getText(), PIDForm.KP0);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}
    		if( this.inputForTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputOneTextField.getText(), PIDForm.KD0);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}

    		if( this.inputFiveTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputTwoTextField.getText(), PIDForm.KI0);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}
    		
    		if( this.inputSixTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputThreeTextField.getText(), PIDForm.KP0);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}
    		this.setVisible(false);
    		System.out.println("PIDForm.actionPerfomed "+this.commands.size());
    		this.transmitChanges();
    	}
    }
	   
    /**
     *  Sends one command each time through serial port. All of changes are stored
     *  in a vector of Strings and, when user clicks on Apply button, AIRboardUI
     *  reads runs over the vector and, after waiting for the DELAY_TIME, sends
     *  the commands it reads. 
     */
    public void transmitChanges(){    	
    	pd = new ProgressDialog(this.parent, this.commands.size() );
		
		SerialPort sPort = this.connection.getSerialPort();
		OutputStream os;

    	/* timer */
    	timer = new Timer(AirboardFrame.DELAY_TIME, new ActionListener() {		
    		public void actionPerformed(ActionEvent evt) {
    		SerialPort sPort;
    		OutputStream os;
    		InputStream is;
    	    	
    		if(!pd.hasBeenStopped() && 
    				i<=( commands.size()-1)){
    					pd.progressBar.setValue(i);
    			    	pd.setText("Appling changes...");
    	
    					String lastInput = (connection.getLastInputString());
    					
    					if(lastInput.length()>0){
    						if(lastInput.lastIndexOf("ERROR")!=-1 ){
    							System.out.println(i+" An error was occurred ");
        						if(errorsCounter >= MAX_ERRORS) { 
        							pd.setCancelled("AIRLab returns error"); 
        							pd.setCompleted(); 
        						}
    							return;
    						} else { 
    							System.out.println(i+" OK"); 
    							i++; 
    						}    			
    					} else {
    						noResponseCounter++;
    						System.out.println("No response");
    						if(noResponseCounter>=MAX_NO_RESPONSE) { 
    							pd.setCancelled("No response"); 
    							pd.setCompleted(); 
    						}
    						return;
    					} 

    					String toOutput;
    					toOutput = (commands.elementAt((i-1))).toString();
	
    					connection.sendAString(toOutput);
    					System.out.println(toOutput);
    				} else { 
    					pd.setCompleted();
    					timer.stop(); 
    				}}
    		});

    		timer.start();
    }
    
    /**
    * Handles closing down system. Allows application to be closed with window
    * close box.
    */
    class CloseHandler extends WindowAdapter {

	PIDForm pf;
	
	public CloseHandler(PIDForm pf) {
	    this.pf = pf;
	}

	public void windowClosing(WindowEvent e) {
		pf.setVisible(false);
	}
    }
	
}
