/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI.menu;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.OutputStream;

import javax.comm.SerialPort;

import airboardUI.AirboardFrame;
import airboardUI.AlertDialog;
import airboardUI.PIDForm;

/**
 * @author Bruschieri Stefano
 */
public class PIDMenu extends Menu implements ActionListener{

	private MenuItem pidFormItem;
	private MenuItem pidControllerItem;
	private AirboardFrame parent;
	
	public PIDMenu(AirboardFrame parent){
		super("PID");
		this.parent = parent;
		//this.setText("PID");
		pidFormItem = new MenuItem("Insert PID values...");
		pidFormItem.addActionListener(this);
		pidControllerItem = new MenuItem("Use PID controller");
		pidControllerItem.addActionListener(this);
		this.add(pidControllerItem);
		this.add(pidFormItem);
	}
	
    /**
	* Responds to the menu items and buttons.
	*/
	public void actionPerformed(ActionEvent e) {
		String cmd = e.getActionCommand();

		// Set controller type to PID one 
		if(cmd.equals("Use PID controller")){
			if( this.parent.getConnection().getSerialPort() == null ){
				AlertDialog np = new AlertDialog(this.parent, "No port is available",
						"In order to transmit any data "+
						"you must "+
						"activate a port.");
			} else {
				SerialPort sPort = this.parent.getConnection().getSerialPort();
				OutputStream os;
				String request = "ewfisfuzzy=+000&+000";
				try {
					os = sPort.getOutputStream();
					for(int j=0; j<request.length(); j++){
						os.write((int)request.charAt(j));
					} 
					os.write((int)'\n');
				} catch (IOException ioe) {
					sPort.close();
					System.err.println("OutputStream write error: " + ioe);
				}
			}
		}
		
		//Insert PID values: 
		if(cmd.equals("Insert PID values...")){
			if( this.parent.getConnection().getSerialPort() == null ){
				AlertDialog np = new AlertDialog(this.parent, "No port is available",
						"In order to transmit any data"+
						"you must"+
						"activate a port.");
			} else {
					PIDForm pf = new PIDForm(this.parent.getConnection(), this.parent);
			}
			
		}
	}
}
