#include "GianoActuationManager.h"
#include "MotorConstants.h"

#include <math.h>

GianoActuationManager::GianoActuationManager()
{}

GianoActuationManager::~GianoActuationManager()
{}


void GianoActuationManager::ComputeActuations()
{
  mActuationValues.clear();
    
  if (mCommandData.find(TAN_SPEED) == mCommandData.end())
    {
      mCommandData[TAN_SPEED] = 0.0;
    }
  
  if (mCommandData.find(ROT_SPEED) == mCommandData.end())
    {
      mCommandData[ROT_SPEED] = 0.0;
    }
  
  mActuationValues[SPEED_MOTOR_DX] = 1*((int)(-1*mCommandData[TAN_SPEED] - mCommandData[ROT_SPEED]));
  mActuationValues[SPEED_MOTOR_SX] = 1*((int)(-1*mCommandData[TAN_SPEED] + mCommandData[ROT_SPEED]));  

  if (mCommandData.find(KICK) != mCommandData.end())
    {
      mActuationValues[KICK] = (int)mCommandData[KICK];
    }
}

void GianoActuationManager::ComputeOdometry()
{
  float mm_sx = mEncoderValues[ENCODER_MOTOR_SX] * mCoefficientTicTomm_Sx;
  float mm_dx = mEncoderValues[ENCODER_MOTOR_DX] * mCoefficientTicTomm_Dx;
 
 	if(abs(mEncoderValues[ENCODER_MOTOR_SX])>10000){
 		printf("\n\n\n\n\n************ errore encoder sx*%d **********\n\n\n\n\n",mEncoderValues[ENCODER_MOTOR_SX]);
 		mm_sx=0;
 		mm_dx=0;
	}
	if(abs(mEncoderValues[ENCODER_MOTOR_DX])>10000){
 		printf("\n\n\n\n\n****** errore encoder sx*%d **********\n\n\n\n\n",mEncoderValues[ENCODER_MOTOR_DX]);
 		mm_sx=0;
 		mm_dx=0;
	}
 
  float movement = ( mm_sx + mm_dx ) / 2.0;
  //printf("movement: left = %f, right = %f, used: %f [mm]\n", mm_sx, mm_dx, movement );

  float rotation = ( mm_sx - mm_dx ) / mWheelDistance;
  //printf("rotation = %f radians, with wheeldistance = %f mm\n", rotation, mWheelDistance );

  
  mOdometryData[ROBOT_MOVEMENT] = movement; // in mm
  
  mOdometryData[ROBOT_DIRECTION] = rotation / 2; // radiants 
  mOdometryData[ROBOT_ORIENTATION] = rotation * (180/M_PI);   // degrees
  
  mOdometryData[ACT_TAN_SPEED] = (float)(mEncoderValues[SPEED_MOTOR_SX] + mEncoderValues[SPEED_MOTOR_DX]) / 2;
  mOdometryData[ACT_ROT_SPEED] = (float)(mEncoderValues[SPEED_MOTOR_SX] - mEncoderValues[SPEED_MOTOR_DX]) / 2;
  //printf("speed: tan = %f, rot = %f\n", mOdometryData[ACT_TAN_SPEED], mOdometryData[ACT_ROT_SPEED] );
}

void GianoActuationManager::SetWheelDistance( float wheeldistance )
{
   mWheelDistance = wheeldistance;
}

void GianoActuationManager::SetConversionCoefficients( float coefficient_sx, float coefficient_dx )
{
   mCoefficientTicTomm_Sx = coefficient_sx;
   mCoefficientTicTomm_Dx = coefficient_dx;
}

