#include "MetamorphicActuationManager.h"
#include "MotorConstants.h"

#include <cmath>

MetamorphicActuationManager::MetamorphicActuationManager()
{}

MetamorphicActuationManager::~MetamorphicActuationManager()
{}


void MetamorphicActuationManager::ComputeActuations()
{
  mActuationValues.clear();
    
  if (mCommandData.find(X_SPEED) == mCommandData.end())
    {
      mCommandData[X_SPEED] = 0.0;
    }
  
  if (mCommandData.find(Y_SPEED) == mCommandData.end())
    {
      mCommandData[Y_SPEED] = 0.0;
    }

  if (mCommandData.find(OMEGA_SPEED) == mCommandData.end())
    {
      mCommandData[OMEGA_SPEED] = 0.0;
    }
  
  if (mCommandData.find(SHAPE) == mCommandData.end())
    {
      mCommandData[SHAPE] = 50.0;
    }
  
  mActuationValues[X_SPEED] = (int)mCommandData[X_SPEED];
  mActuationValues[Y_SPEED] = (int)mCommandData[Y_SPEED];
  mActuationValues[OMEGA_SPEED] = (int)mCommandData[OMEGA_SPEED];
  mActuationValues[SHAPE] = (int)mCommandData[SHAPE];

  if (mCommandData.find(KICK) != mCommandData.end())
    {
      mActuationValues[KICK] = (int)mCommandData[KICK];
    }
}

void MetamorphicActuationManager::ComputeOdometry()
{
  mOdometryData[ROBOT_MOVEMENT] = sqrt((float) (mEncoderValues[DELTA_X]*mEncoderValues[DELTA_X] 
					  + mEncoderValues[DELTA_Y]*mEncoderValues[DELTA_Y])) / 10; //conversion from mm to cm
  mOdometryData[ROBOT_DIRECTION] = atan2f(mEncoderValues[DELTA_Y],mEncoderValues[DELTA_X]); // radiants
  mOdometryData[ROBOT_ORIENTATION] = mEncoderValues[DELTA_OMEGA]; // degrees

  mOdometryData[ACT_X_SPEED] = (float)mEncoderValues[ACT_X_SPEED];
  mOdometryData[ACT_Y_SPEED] = (float)mEncoderValues[ACT_Y_SPEED];
  mOdometryData[ACT_OMEGA_SPEED] = (float)mEncoderValues[ACT_OMEGA_SPEED];
}
