#include "SerialCommunicationEmmeQuadro.h"

#include "MotorConstants.h"

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>

using namespace std;

// Definizioni del modulo
#define STATUS_OFF 0
#define STATUS_ON 1
#define SX 0
#define DX 1
#define BOTH 2


#define FALSE 0
#define TRUE 1
#define _POSIX_SOURCE 1



SerialCommunicationEmmeQuadro::SerialCommunicationEmmeQuadro()
{
  mBaudRate = B115200;
  mModemDevice = strdup(TTYS0);
  mHardwareFlowControl = true;
  InitCommunication();
  SetMotorStatus(STATUS_ON);
  SetMotorsSpeed(0,0);
  ResetEncoder(BOTH);
  SetKicker(STATUS_OFF,STATUS_OFF);
}

SerialCommunicationEmmeQuadro::~SerialCommunicationEmmeQuadro()
{
  SetMotorsSpeed(0,0);
  SetMotorStatus(STATUS_OFF);
  CloseCommunication();
}


void SerialCommunicationEmmeQuadro::WriteActuations()
{
  SetMotorsSpeed(-mActuationValues[SPEED_MOTOR_SX],-mActuationValues[SPEED_MOTOR_DX]);

  if (mActuationValues.find(KICK) != mActuationValues.end())
    {
      int kick_command = mActuationValues[KICK];
      
      if (kick_command == 2)
	{
	  SetKicker(STATUS_ON,STATUS_ON);
	}
      else if (kick_command == 1)
	{
	  SetKicker(STATUS_ON,STATUS_OFF);
	}
      else if (kick_command == -1)
	{
	  SetKicker(STATUS_OFF,STATUS_ON);
	}
      
      if (kick_command != 0)
	{
	  SetKicker(STATUS_OFF,STATUS_OFF);
	}     
    }
}


// -----------------------------------------------------------------------
//                              ENCODER
// -----------------------------------------------------------------------


  // Reset degli encoders: 0=sx, 1=dx, 2=sx+dx
void SerialCommunicationEmmeQuadro::ResetEncoder(int i)
{
  int d0, d1, d2, d3;

  Write(4, i&0x03, 0, 0);
  Read (&d0, &d1, &d2, &d3);
}


// Legge i valori (32bit) degli encoders:
void SerialCommunicationEmmeQuadro::GetEncoder(long int* sx, long int *dx)
{
  int d0, d1, d2, d3;

  // Campiona l'encoder "sx"
  Write(5, 0, 0, 0);
  Read (&d3, &d2, &d1, &d0);
  *sx = conv_uint(d0) + conv_uint(d1)*256 + conv_uint(d2)*65536 + conv_uint(d3)*16777216;

  // Campiona l'encoder "dx"
  Write(5, 1, 0, 0);
  Read (&d3, &d2, &d1, &d0);
  *dx = conv_uint(d0) + conv_uint(d1)*256 + conv_uint(d2)*65536 + conv_uint(d3)*16777216;
}



// -----------------------------------------------------------------------
//                                MOTORI
// -----------------------------------------------------------------------

// Stato dei motori: false=off, true=on
void SerialCommunicationEmmeQuadro::SetMotorStatus(int i)
{
  int d0, d1, d2, d3;

  Write(1, 0, i&0x01, 0);
  Read (&d0, &d1, &d2, &d3);
}


// Velocit di rotazione dei motori: intervallo di valori [-100, 100]
void SerialCommunicationEmmeQuadro::SetMotorsSpeed(int sx, int dx)
{
  int d0, d1, d2, d3;

  Write(2, 0, sx&0xFF, dx&0xFF);
  Read (&d0, &d1, &d2, &d3);
}


// -----------------------------------------------------------------------
//                                KICKER
// -----------------------------------------------------------------------

// Stato dei kickers: 0=off, 1=on
void SerialCommunicationEmmeQuadro::SetKicker(int dx, int sx)
{
  int d0, d1, d2, d3;

  Write(10, 0, dx&0x01, sx&0x01);
  Read (&d0, &d1, &d2, &d3);
}


// -----------------------------------------------------------------------
//                              BUMPERS
// -----------------------------------------------------------------------

// Stato dei bumpers: 0=off, 1=on
void SerialCommunicationEmmeQuadro::GetBumpers(int *dx, int *sx)
{
  int d0, d1;

  Write(9, 0, 0, 0);
  Read (&d0, &d1, dx, sx);
}



// Scrive un dato sulla seriale
void SerialCommunicationEmmeQuadro::Write(int oda0, int oda1, int oda2, int oda3)
{
  int i;
  char buf[9];

  buf[0]= hexdec[(oda0>>4)&0x0F];
  buf[1]= hexdec[oda0&0x0F];
  buf[2]= hexdec[(oda1>>4)&0x0F];
  buf[3]= hexdec[oda1&0x0F];
  buf[4]= hexdec[(oda2>>4)&0x0F];
  buf[5]= hexdec[oda2&0x0F];
  buf[6]= hexdec[(oda3>>4)&0x0F];
  buf[7]= hexdec[oda3&0x0F];
  buf[8]= 0x0D;
  
  for (i=0; i<9; i++)
    {
      write(fd, &buf[i], 1);
      usleep(5000);
    }
}



// Legge un dato dalla seriale
void SerialCommunicationEmmeQuadro::Read(int* ida0, int* ida1, int* ida2, int* ida3)
{
  int res;
  char buf[255];
  
  res = read(fd, buf, 255);
  *ida0 = base2((int)(h2d(buf[1])&0x0F)|((h2d(buf[0])&0x0F)<<4));
  *ida1 = base2((int)(h2d(buf[3])&0x0F)|((h2d(buf[2])&0x0F)<<4));
  *ida2 = base2((int)(h2d(buf[5])&0x0F)|((h2d(buf[4])&0x0F)<<4));
  *ida3 = base2((int)(h2d(buf[7])&0x0F)|((h2d(buf[6])&0x0F)<<4));
}



