#include "SerialCommunicationTRIBOARD.h"
#include "MotorConstants.h"

#include <assert.h>
#include <sstream>
#include <iostream>
 
using namespace std;

const int WRITING_DELAY = 0;
const int MAX_SPEED = 100;

SerialCommunicationTRIBoard::SerialCommunicationTRIBoard()
{
#ifdef rabbiati
  mBaudRate = B115200;
#else
  mBaudRate = B115200;
#endif
  mModemDevice = strdup("/dev/ttyUSB0");
  mHardwareFlowControl = false;
  InitCommunication();
  StartMotors();
}

SerialCommunicationTRIBoard::~SerialCommunicationTRIBoard()
{
  StopMotors();
  CloseCommunication();
}

void SerialCommunicationTRIBoard::StartMotors()
{
  SetMotorsSpeed(0,0,0);
  usleep(WRITING_DELAY);
}

void SerialCommunicationTRIBoard::StopMotors()
{
  SetMotorsSpeed(0,0,0);
  usleep(WRITING_DELAY);
  Write("b\r");
}

void SerialCommunicationTRIBoard::WriteActuations()
{
  SetMotorsSpeed(mActuationValues[SPEED_MOTOR_0],
		 mActuationValues[SPEED_MOTOR_1],
		 mActuationValues[SPEED_MOTOR_2]);

  if (mActuationValues.find(KICK) != mActuationValues.end())
    {
      usleep(WRITING_DELAY);
      Kick(mActuationValues[KICK]);
    }

}

void SerialCommunicationTRIBoard::ReadEncoders()
{
  int n;
  char c;

  Write("e\r");
  Read(&c);
  assert(c=='e');
  
  Read24bitNumber(&n);
  mEncoderValues[ENCODER_MOTOR_0] = n;
  Read24bitNumber(&n);
  mEncoderValues[ENCODER_MOTOR_1] = n;
  Read24bitNumber(&n);
  mEncoderValues[ENCODER_MOTOR_2] = n;
}

void SerialCommunicationTRIBoard::SetMotorsSpeed(int speed_motor_0, int speed_motor_1, int speed_motor_2)
{
  char abs_speed_motor_0 = (char)speed_motor_0;
  char abs_speed_motor_1 = (char)speed_motor_1;
  char abs_speed_motor_2 = (char)speed_motor_2;

  Write("m");
  usleep(0);
  Write(abs_speed_motor_0);
  usleep(0);
  Write(abs_speed_motor_1);
  usleep(0);
  Write(abs_speed_motor_2);
  usleep(0);
  Write("\r");
}


void SerialCommunicationTRIBoard::Kick(int kicker)
{
  if (kicker == 0)
    {
      Write("k0\r");
    }
  else if (kicker == 1)
    {
      Write("k1\r");
    }
  else if (kicker == -1)
    {
      Write("k2\r");
    }

}
