# -*- coding: UTF-8 -*-

import sys
import getopt
import re

RESET = '\033[0;0m'
BOLD = '\033[1m'
BLACK = '\033[30m'
RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[33m'
BLUE = '\033[34m'
MAGENTA = '\033[35m'
CYAN = '\033[36m'
WHITE = '\033[37m'

def usage():
	print 'This script generates rules for Mr.BRIAN engine starting from templates'
	print 'Version 0.1'
	print '2006-03-17'
	print '\n'
	
	
	print '\t-h, --help\tPrint this message\n'
	
	print '\t-l, --log\tAdd comments to the output file to identify the auto generated rules\n'
	
	print '\t-o, --output\tThe output file generated from the template\n'
	
	print '\t-i, --input\tThe input file containing the template rules\n'
	print 'Example of template (TODO)'
	print '[4]{'
	print 'This is  <D>'
	print 'Angle+1 <D+1>'
	print 'Angle-1 <D-1>'
	print '}'
	
def getFile(fileName):
	try:
		f = open(fileName, "rb")
		text = f.read()
		f.close()
		return text
	except IOError:
		print '\t',RED,BOLD,'ERROR',RESET,' cannot open file ' + fileName
		sys.exit()

def getElement(array,offset):
	# fare il modulo dell'offset 
	if offset < 0:
		offset = offset + len(array) 
	return array[offset]

def shiftArray(array, n):
	for i in range(0,n):
		temp = array[0]
		for j in range(0, len(array)-1):
			array[j] = array[j+1]
		array[len(array)-1] = temp

def buildBehavior(inputFile, outputFile, log):
	allDirections = ['N','NNW','NW','WNW','W','WSW','SW','SSW','S','SSE','SE','ESE','E','ENE','NE','NNE']
	originalDirections = ['N','NNW','NW','WNW','W','WSW','SW','SSW','S','SSE','SE','ESE','E','ENE','NE','NNE']

	print 'Generating from template file',BOLD,inputFile, RESET,'output file',BOLD,outputFile,RESET
	
	try:
		f = open(inputFile)
		f.close()			
	except IOError:
		print '\t',RED,BOLD,'ERROR',RESET,' cannot open input file ' , inputFile , '\n'
		#usage()
		sys.exit()
		
	if outputFile=='':
		print '\t',RED,BOLD,'ERROR',RESET,' please select output file ' , '\n'
		#usage()
		sys.exit()
		
	print '\tBuilding behavior...'
	result = ''
	
	text = getFile(inputFile)
	#print text
	#print "Matching.."
	
	dirRE = '\[(?P<directions>[NSEW\-,\s48(16)]*)\]'
	sepRE = '[\s]*'
	tempRE ='\{(?P<template>[\w\W]*?)\}'
	p = re.compile(dirRE + sepRE + tempRE,re.IGNORECASE)
	iterator = p.finditer(text)
	
	start = 0
	num = 1
	for match in iterator:
		print '\tGenerating section ' , num ,
		#print '--Match found: ' , match.group()
		#print '--Match span: ' , match.span()
		#print '--Match group directions: ' , match.group('directions')
		directions = match.group('directions')
		#print '--Match group body: ' , match.group('template')
		template = match.group('template')
		
		result += text[start:match.start()]
		start = match.end()
		
		directionsToBuild = []
		
		r = re.compile(',',re.IGNORECASE)
		strings = r.split(directions)
		for string in strings:
			#print '--------Match found: \"' , string , '\"'
			if string.strip()=='4' or string.strip()=='8':
				if 'N' not in directionsToBuild:
					directionsToBuild.append('N')
				if 'S' not in directionsToBuild:
					directionsToBuild.append('S')
				if 'E' not in directionsToBuild:
					directionsToBuild.append('E')
				if 'W' not in directionsToBuild:
					directionsToBuild.append('W')
			if string.strip()=='8':
				if 'NW' not in directionsToBuild:
					directionsToBuild.append('NW')
				if 'SW' not in directionsToBuild:
					directionsToBuild.append('SW')
				if 'SE' not in directionsToBuild:
					directionsToBuild.append('SE')
				if 'NE' not in directionsToBuild:
					directionsToBuild.append('NE')
			if string.strip()=='16':
				directionsToBuild = allDirections
			else:
				q = re.compile('\-',re.IGNORECASE)
				iStrings = q.split(string)
				
				if len(iStrings)==1:
					#print '----------------------------Parsing: \"' , iStrings[0] , '\"'
					z = re.compile('[NSEW]+',re.IGNORECASE)
					iDirectionMatch = z.search(iStrings[0])
					
					if iDirectionMatch:
						iDirection = iDirectionMatch.group().strip()
						if iDirection not in allDirections:
							print '\t',RED,BOLD,'ERROR',RESET,' invalid direction \"' , iDirection , '\"'
							sys.exit()
						
						#print '----------------------------Match found: \"' , iDirection , '\"'
						if iDirection not in directionsToBuild:
							#print '----------------------------New direction: \"' , iDirection , '\"'
							directionsToBuild.append(iDirection)
				
				if len(iStrings)==2:
					#print '----------------------------Parsing: \"' , iStrings[0] , "-" , iStrings[1] , '\"'
					z = re.compile('[NSEW]+',re.IGNORECASE)
					
					startingDirectionMatch = z.search(iStrings[0])
					endingDirectionMatch = z.search(iStrings[1])
					
					if startingDirectionMatch:
						sDirection = startingDirectionMatch.group().strip()
						sIndex = allDirections.index(sDirection)
						if sDirection not in allDirections:
							print '\t',RED,BOLD,'ERROR',RESET,' invalid direction \"' , sDirection , '\"'
							sys.exit()
					
						if endingDirectionMatch:
							eDirection = endingDirectionMatch.group().strip()
							eIndex = allDirections.index(eDirection)
							if eDirection not in allDirections:
								print '\t',RED,BOLD,'ERROR',RESET,' invalid direction \"' , eDirection , '\"'
								sys.exit()
					
					
							if sIndex > eIndex:
								print '\t',RED,BOLD,'ERROR',RESET,' invalid direction interval: '
								sys.exit()
							else:
								for i in range(sIndex,eIndex):
									iDirection = allDirections[i]
									#print '----------------------------Match found: \"' , iDirection , '\"'
									if iDirection not in directionsToBuild:
										#print '----------------------------New direction: \"' , iDirection , '\"'
										directionsToBuild.append(iDirection)
		
		#print directionsToBuild
		for i in range(0,len(directionsToBuild)):
			for j in range(i+1,len(directionsToBuild)):
				if originalDirections.index(directionsToBuild[i]) > originalDirections.index(directionsToBuild[j]):
					#print 'prima' ,directionsToBuild[i] , directionsToBuild[j]
					temp = directionsToBuild[i]
					directionsToBuild[i] = directionsToBuild[j]
					directionsToBuild[j] = temp
					#print 'dopo' , directionsToBuild[i] , directionsToBuild[j]
		#print directionsToBuild
					
			
		#print directionsToBuild	
		if log=='true':
			result += '## Start auto generated rule\n'
		#print directionsToBuild ,
		if len(directionsToBuild) > 0:
			num = num + 1
		print directionsToBuild,
		#for direction in directionsToBuild:
		for index in range(0,len(directionsToBuild)):
			#print 'Indice' , index, 'Valore', directionsToBuild[index]
			direction = directionsToBuild[index]			 
			while direction != allDirections[0]:
				shiftArray(allDirections,1)
				#print "shifto, ora è " , allDirections
			# mettere le re per la sostituzione
			behavior = template.replace('<D>',allDirections[0])
			#print allDirections[0]
			#print originalDirections.index(allDirections[0])
			behavior = behavior.replace('<N>',str(originalDirections.index(allDirections[0])))
			for n in range(-15,16):
				# mettere le re per la sostituzione
				behavior = behavior.replace('<D' + str(n) + '>' , getElement(allDirections,n))
				if n>0:
					m = '+' + str(n) 
					# mettere le re per la sostituzione
					behavior = behavior.replace('<D' + str(m) + '>' , getElement(allDirections,n))
			if log=='true':
				result += '\n'
				result += '## Auto generated rule: direction ' + direction + '\n'
			result += behavior
			#result += '\n'
		result += '\n'
		print ''
		if log=='true':
			result += '## End auto generated rule\n'

	result += text[start:]
	
	try:
		flag = 'false'
		try:
			f = open(outputFile, "rb")
			oldResult = f.read()
			f.close()
		except IOError:
			flag = 'true'
			oldResult = ''
			
		if result!=oldResult:
			if flag=='false':
				response = raw_input('\tOverride output file ' + outputFile + '? y/n ')
			else:
				response = 'y'
			if (response.strip().lower()=='y'):
				if flag=='false':
					print '\t',YELLOW,BOLD,'WARNING',RESET,': the file ', outputFile,' will be replaced'
				file = open(outputFile, "w")
				file.write(result)				
				print '\t',GREEN,BOLD,'DONE',RESET,' the file',outputFile,'has been created successfully' ,				
				file.close()
			else:
				print '\t',RED,BOLD,'ERROR',RESET,' the file will NOT be replaced', '\n' ,	
			
		else:
			print '\t',GREEN,BOLD,'DONE',RESET,' the file',outputFile,'is up to date'
	except IOError:
		print '\t',RED,BOLD,'ERROR',RESET,' cannot write to output file ', outputFile ,
		sys.exit()


def main():
	inputFile = ''
	outputFile = ''
	log = ''
	
	try:
		options, arguments = getopt.getopt(sys.argv[1:], 'li:o:h', ['log','inputFile=','outputFile=','help'])
	except getopt.GetoptError:
		print '\t',RED,BOLD,'ERROR',RESET,' error parsing arguments'
		sys.exit()
	log = 'false'
	for opt, arg in options:
		if opt in ("-l", "--log"):
			log = 'true'
		elif opt in ("-i", "--inputFile"):
			inputFile = arg
		elif opt in ("-o", "--outputFile"):
			outputFile = arg
		elif opt in ("-h", "--help"):
			usage()
			sys.exit()
		
	buildBehavior(inputFile,outputFile,log)
