#include "MatrixFunctions.h"

using namespace std;

void print3DMatrix(Param3DRotoTraslation *p,string header,string tailer,FILE *f){
	Matrix3DRotoTraslation *m=toMatrix3DRotoTraslation(p);
	print3DMatrix(m,header,tailer,f);
	delete m;
}

void print3DMatrix(Matrix3DRotoTraslation *m,string header,string tailer,FILE *f){
	fprintf(f,"%s",header.c_str());
	for(int i=0;i<4;i++){
		for(int j=0;j<4;j++){
			fprintf(f,"%0.3f ",m->getValue(i,j));
		}
		fprintf(f,"\n");			
	}	
	fprintf(f,"%s",tailer.c_str());
}

void print3DParam(Matrix3DRotoTraslation *m,string header,string tailer,FILE *f){
	Param3DRotoTraslation *p=toParam3DRotoTraslation(m);
	print3DParam(p,header,tailer,f);
	delete p;
}

void print3DParam(Param3DRotoTraslation *p,string header,string tailer,FILE *f){
	fprintf(f,"%s",header.c_str());
	fprintf(f,"%0.3f ",p->getX());
	fprintf(f,"%0.3f ",p->getY());
	fprintf(f,"%0.3f ",p->getZ());
	fprintf(f,"%0.3f ",p->getA());
	fprintf(f,"%0.3f ",p->getB());
	fprintf(f,"%0.3f ",p->getC());
	fprintf(f,"%s",tailer.c_str());
}

Matrix3DRotoTraslation * toMatrix3DRotoTraslation(Param3DRotoTraslation *p){
	Matrix3DRotoTraslation *m=new Matrix3DRotoTraslation();
	m->setValue(0,0,p->getCosA()*p->getCosB());
	m->setValue(0,1,p->getCosA()*p->getSinB()*p->getSinC()-p->getSinA()*p->getCosC());
	m->setValue(0,2,p->getCosA()*p->getSinB()*p->getCosC()+p->getSinA()*p->getSinC());
	m->setValue(0,3,p->getX());
	m->setValue(1,0,p->getSinA()*p->getCosB());
	m->setValue(1,1,p->getSinA()*p->getSinB()*p->getSinC()+p->getCosA()*p->getCosC());
	m->setValue(1,2,p->getSinA()*p->getSinB()*p->getCosC()-p->getCosA()*p->getSinC());
	m->setValue(1,3,p->getY());
	m->setValue(2,0,-p->getSinB());
	m->setValue(2,1,p->getCosB()*p->getSinC());
	m->setValue(2,2,p->getCosB()*p->getCosC());
	m->setValue(2,3,p->getZ());
	return m;
}

Param3DRotoTraslation * toParam3DRotoTraslation(Matrix3DRotoTraslation *m){
	Param3DRotoTraslation *p=new Param3DRotoTraslation();
	p->setX(	m->getValue(0,3));
	p->setY(	m->getValue(1,3));
	p->setZ(	m->getValue(2,3));
	p->setA(atan2(m->getValue(1,0),m->getValue(0,0)));
	p->setB(atan2(-m->getValue(2,0),sqrt(m->getValue(2,1)*m->getValue(2,1)+m->getValue(2,2)*m->getValue(2,2))));
	p->setC(atan2(m->getValue(2,1),m->getValue(2,2)));	
	return p;
}

Matrix3DRotoTraslation * multMatrix3D(Matrix3DRotoTraslation *m1,Matrix3DRotoTraslation *m2){
	double d;
	Matrix3DRotoTraslation *mr=new Matrix3DRotoTraslation();
	for(int i=0;i<4;i++){
		for(int j=0;j<4;j++){
			d=0;
			for(int k=0;k<4;k++){
				d+=m1->getValue(i,k)*m2->getValue(k,j);
			}
			mr->setValue(i,j,d);
		}
	}
	return mr;
}


Matrix3DRotoTraslation *multMatrix3D(Param3DRotoTraslation *p1,Matrix3DRotoTraslation *m2){
	Matrix3DRotoTraslation *m1=toMatrix3DRotoTraslation(p1);
	Matrix3DRotoTraslation *m3=multMatrix3D(m1,m2);
	delete m1;
	return m3;
}

Matrix3DRotoTraslation *multMatrix3D(Matrix3DRotoTraslation *m1,Param3DRotoTraslation *p2){
	Matrix3DRotoTraslation *m2=toMatrix3DRotoTraslation(p2);
	Matrix3DRotoTraslation *m3=multMatrix3D(m1,m2);
	delete m2;
	return m3;
}

Matrix3DRotoTraslation *multMatrix3D(Param3DRotoTraslation *p1,Param3DRotoTraslation *p2){
	Matrix3DRotoTraslation *m1=toMatrix3DRotoTraslation(p1);
	Matrix3DRotoTraslation *m2=toMatrix3DRotoTraslation(p2);
	Matrix3DRotoTraslation *m3=multMatrix3D(m1,m2);
	delete m1;
	delete m2;
	return m3;
}

Matrix3DRotoTraslation *invMatrix3D(Matrix3DRotoTraslation *m1){
	Matrix3DRotoTraslation *matinv=new Matrix3DRotoTraslation();
	for(int i=0;i<3;i++){
		for(int j=0;j<3;j++){
			matinv->setValue(i,j,m1->getValue(j,i));
		}
	}
	for(int i=0;i<3;i++){
		matinv->setValue(i,3,-(m1->getValue(0,i)*m1->getValue(0,3)+m1->getValue(1,i)*m1->getValue(1,3)+m1->getValue(2,i)*m1->getValue(2,3)));
	}
	return matinv;	
}

Matrix3DRotoTraslation *invMatrix3D(Param3DRotoTraslation *p1){
	Matrix3DRotoTraslation *m1=toMatrix3DRotoTraslation(p1);
	Matrix3DRotoTraslation *m2=invMatrix3D(m1);
	delete m1;
	return m2;
}

