
#include <stdio.h>
#include <stdlib.h>
#include <wiiuse.h>
#include <sys/time.h>
#include <math.h>
#include <unistd.h>

#include <string>
#include <sstream>

#include "wiim.h"

using namespace std;

#define MAX_WIIMOTES				1

#define GR
#define SC  2
#define PI 3.14159265358979323846

wiimote** wiimotes = NULL;

wiimote_t* wm0;

Wiim::Wiim()
{
	int found, connected;
	int i;
	
	fAuto=STARTMODE;
	BlinkRightStatus=0;
	BlinkLeftStatus=0;
	RumbleWStatus=0;
	RumbleSStatus=0;
	pointDiff=0;
	
	printf("\nwiim started\n"); 

	//Initialize an array of wiimote objects.

	wiimotes =  wiiuse_init(MAX_WIIMOTES);
	
	//Find wiimote devices
	 
	found = 0;
	while (!found) {
		found=wiiuse_find(wiimotes, MAX_WIIMOTES,2 );
		printf ("No wiimotes found. Trying again....\n");
	}


	//Connect to the wiimotes

	connected = wiiuse_connect(wiimotes, MAX_WIIMOTES);
	if (connected)
		printf("Connected to %i wiimotes (of %i found).\n", connected, found);
	else {
		printf("Failed to connect to any wiimote.\n");
	}

	usleep(1000000);//stop blinking
	
	//wiiuse_rumble(wiimotes[0], 1);
	for(i=0;i<3;i++){
		wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_1);
		usleep(50000); 
		wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_2);
		usleep(50000); 
		wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_3);
		usleep(50000); 
		wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_4);
		usleep(50000); 
	}
	wiiuse_set_leds(wiimotes[0], 0);
	wiiuse_rumble(wiimotes[0], 0);
	wiiuse_motion_sensing(wiimotes[0], 1);//enable accelerometer		
	usleep(10000);
	wiiuse_set_flags(wiimotes[0],WIIUSE_CONTINUOUS,WIIUSE_ORIENT_THRESH);
	usleep(10000);
	wiiuse_set_ir(wiimotes[0], 1);
	wm0=wiimotes[0];
	if(fAuto==0){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 | WIIMOTE_LED_4);
	}
	
}

Wiim::~Wiim()
{
}

int Wiim::run()
{
	
	int i=0;
		
		if(!wiiuse_poll(wiimotes, MAX_WIIMOTES)){
			return 0;
		}else{
			BlinkRight(BlinkRightStatus);
			BlinkLeft(BlinkLeftStatus);
			WeakRumble(RumbleWStatus);
			StrongRumble(RumbleSStatus);
			return 1;
		}
	
	
}



void Wiim::handle_ctrl_status(struct wiimote_t* wm){
	printf("get status\n");
	printf("--- CONTROLLER STATUS [wiimote id %i] ---\n", wm->unid);
	printf("attachment:      %i\n", wm->exp.type);
	printf("speaker:         %i\n", WIIUSE_USING_SPEAKER(wm));
	printf("ir:              %i\n", WIIUSE_USING_IR(wm));
	printf("leds:            %i %i %i %i\n", WIIUSE_IS_LED_SET(wm, 1), WIIUSE_IS_LED_SET(wm, 2), WIIUSE_IS_LED_SET(wm, 3), WIIUSE_IS_LED_SET(wm, 4));
	printf("battery:         %f %%\n", wm->battery_level);
}

void Wiim::handle_event(struct wiimote_t* wm) {
	
	int i;
			
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_A))		printf("A pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_B))		printf("B pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_UP))		printf("UP pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_DOWN))	printf("DOWN pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_LEFT))	printf("LEFT pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_RIGHT))	printf("RIGHT pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_MINUS))	printf("MINUS pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_PLUS))	printf("PLUS pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_ONE))		printf("ONE pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_TWO))		printf("TWO pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_HOME))	printf("HOME pressed\n");
			
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_A)){
				printf("accellerometer enabled\n");
				wiiuse_motion_sensing(wm, 1);//enable accelerometer				
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_B)){
				printf("accellerometer disabled\n");
				wiiuse_motion_sensing(wm, 0);//disable accelerometer
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_PLUS)){
				printf("ir enabled\n");
				wiiuse_set_ir(wm, 1);//enable ir
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_MINUS)){
				printf("ir disabled\n");
				wiiuse_set_ir(wm, 0);//disable ir
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_HOME)){
				printf("status requested\n");
				wiiuse_status(wm);//request status
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_ONE)){
				printf("position tracking enabled\n");
				wiiuse_motion_sensing(wm, 1);
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_TWO)){
				printf("position tracking disabled\n");
				wiiuse_motion_sensing(wm, 0);
			}
			
			if (WIIUSE_USING_ACC(wm)) {
				printf("wiimote calibration xyz 0g: %i %i %i --- 1g: %i %i %i\t%i\n",
						wm->accel_calib.cal_zero.x,wm->accel_calib.cal_zero.y,wm->accel_calib.cal_zero.z,
						wm->accel_calib.cal_g.x,wm->accel_calib.cal_g.y,wm->accel_calib.cal_g.z);
	      printf("x:%f y:%f z:%f \n",wm->gforce.x,wm->gforce.y,wm->gforce.z);
			}

			if (WIIUSE_USING_IR(wm)) {
				// definisce il numero dei led da rilevare 
				for (i=0; i < 3; i++) {
					if (wm->ir.dot[i].visible){
						printf("IR%i ---: (%u, %u) %u\t\t", i, wm->ir.dot[i].x, wm->ir.dot[i].y,wm->ir.dot[i].size);
					}else{
						printf("IR%i ***: (%u, %u) %u\t\t", i, wm->ir.dot[i].x, wm->ir.dot[i].y,wm->ir.dot[i].size);
					}
				}
				printf("\n");
			}
}

/*
float Wiim::get_X_acceleration() {
	return X_acceleration;
}

float Wiim::get_Y_acceleration() {
	return Y_acceleration;
}

float Wiim::get_Z_acceleration() {
	return Z_acceleration;
}
*/

void Wiim::SetpointDiff(int pd)
{
	pointDiff=pd;
}

void Wiim::BlinkRight(int status)
{
	if (status==0){
		return;
	}
	if(status<239){
		if ((status%32)>=0 && (status%32)<8){
			wiiuse_set_leds(wm0, WIIMOTE_LED_1);
		}
		if ((status%32)>8 && (status%32)<16){
			wiiuse_set_leds(wm0, WIIMOTE_LED_2);
		}
		if ((status%32)>16 && (status%32)<24){
			wiiuse_set_leds(wm0, WIIMOTE_LED_3);
		}
		if ((status%32)>24 && (status%32)<32){
			wiiuse_set_leds(wm0, WIIMOTE_LED_4);
		}
	}
	status++;
	if(status==240){
		showPoint();
	}
	if(status>550){
		status=0;
		wiiuse_set_leds(wm0, 0);
	}
	BlinkRightStatus=status;
	return;
}

void Wiim::BlinkLeft(int status)
{
	if (status==0){
		return;
	}
	if(status<239){
		if ((status%32)>=0 && (status%32)<8){
			wiiuse_set_leds(wm0, WIIMOTE_LED_4);
		}
		if ((status%32)>8 && (status%32)<16){
			wiiuse_set_leds(wm0, WIIMOTE_LED_3);
		}
		if ((status%32)>16 && (status%32)<24){
			wiiuse_set_leds(wm0, WIIMOTE_LED_2);
		}
		if ((status%32)>24 && (status%32)<32){
			wiiuse_set_leds(wm0, WIIMOTE_LED_1);
		}
	}
	status++;
	if(status==240){
		showPoint();
	}
	if(status>550){
		status=0;
		wiiuse_set_leds(wm0, 0);
	}
	BlinkLeftStatus=status;
	return;
}

void Wiim::WeakRumble(int status){
	if (status==0){
		return;
	}
	if (status<15){
		wiiuse_rumble(wm0, 1);
		status++;
	}else{
		wiiuse_rumble(wm0, 0);
		status=0;
	}
	RumbleWStatus=status;
}

void Wiim::StrongRumble(int status){
	if (status==0){
		return;
	}
	if (status<150){
		wiiuse_rumble(wm0, 1);
		status++;
	}else{
		wiiuse_rumble(wm0, 0);
		status=0;
	}
	RumbleSStatus=status;
}

int Wiim::IsAuto()
{
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_PLUS)){ 
		fAuto=1;
		wiiuse_set_ir(wm0, 1);
		wiiuse_set_leds(wm0, 0);
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_MINUS)){ 
		fAuto=0;
		showPoint();
	}

	if(fAuto){
		return 1;
	}else{
		return 0;
	}
}

void Wiim::showPoint()
{	
	switch (pointDiff){
		case 0:
			wiiuse_set_leds(wm0, WIIMOTE_LED_2 | WIIMOTE_LED_3);
		break;
		case 1:
			wiiuse_set_leds(wm0,WIIMOTE_LED_3);
		break;
		case 2:
			wiiuse_set_leds(wm0, WIIMOTE_LED_3 | WIIMOTE_LED_4);
		break;
		case 3:
			wiiuse_set_leds(wm0, WIIMOTE_LED_4);
		break;
		case 4:
			wiiuse_set_leds(wm0,  WIIMOTE_LED_2 | WIIMOTE_LED_3 | WIIMOTE_LED_4);
		break;
		case -1:
			wiiuse_set_leds(wm0, WIIMOTE_LED_2 );
		break;
		case -2:
			wiiuse_set_leds(wm0, WIIMOTE_LED_1 | WIIMOTE_LED_2);
		break;
		case -3:
			wiiuse_set_leds(wm0, WIIMOTE_LED_1);
		break;
		case -4:
			wiiuse_set_leds(wm0,  WIIMOTE_LED_1 | WIIMOTE_LED_2 | WIIMOTE_LED_3);
		break;
		default:
			wiiuse_set_leds(wm0,  WIIMOTE_LED_1 | WIIMOTE_LED_2 | WIIMOTE_LED_3 | WIIMOTE_LED_4);
		break;
	}
}
		
void Wiim::SetCharge(float c)
{
	if(BlinkRightStatus!=0 || BlinkLeftStatus!=0){
		return;
	}
	if(c<1){
		wiiuse_set_leds(wm0, 0);
	}
	if(c>=1 && c<2){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1);
	}else if(c>=2 && c<3){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 |  WIIMOTE_LED_2);
	}else if(c>=3 && c<4){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 |  WIIMOTE_LED_2 | WIIMOTE_LED_3);
	}else if(c>=4){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 |  WIIMOTE_LED_2 | WIIMOTE_LED_3 | WIIMOTE_LED_4);
	}
}	
		

float Wiim::getPitch()
{
	float p;
	p=wm0->orient.pitch;
	//printf("%f\n",p);
	return p;
}



string Wiim::getToMotor()
{
	string buf;
	//controllo manuale
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_ONE)){ //avanti
		buf += "TanSpeed";
		buf += " 0.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_UP)){ //avanti
		buf += "TanSpeed";
		buf += " 10.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_DOWN)){ //indietro
		buf += "TanSpeed";
		buf += " -10.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_RIGHT)){ //destra 
		buf += "RotSpeed";
		buf += " -10.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_LEFT)){ //sinistra
		buf += "RotSpeed";
		buf += " +10.0";
		return buf;
	}
	
	return "";
}

int Wiim::getEmergencyStop()
{
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_TWO)){
		return 1;
	}else{
		return 0;
	}
}

int Wiim::getButtonB()
{
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_B)) { 
		return 1;
	}
	return 0;
}

int Wiim::getJButtonHome()
{
	if (IS_JUST_PRESSED(wm0, WIIMOTE_BUTTON_HOME)){
		return 1;
	}else{
		return 0;
	}
}

int Wiim::getHButtonA()
{
	if (IS_HELD(wm0, WIIMOTE_BUTTON_A)){
		return 1;
	}else{
		return 0;
	}
}
	
float Wiim::GetAcceleration()
{
	float ta;
	ta = sqrt(pow(wm0->gforce.x,2)+pow(wm0->gforce.y,2)+pow(wm0->gforce.z,2));
	//printf (" -- %f\n",ta);
	return ta;
}

ppoint Wiim::getcenter()
{

	int x=0,y=0;
	ppoint p;
	
	//printf("% d\t%d\t%d\n",wm0->ir.num_dots,wm0->ir.ax,wm0->ir.ay);
	if (wm0->ir.num_dots>0){
		if(wm0->ir.ax>=0 && wm0->ir.ax <=1024 && wm0->ir.ay>=0 && wm0->ir.ay <=768){
			p.x=wm0->ir.ax;
			p.y=wm0->ir.ay;
			p.valid=1;
		}else{
			p.valid=0;
		}
	}else{
		p.valid=0;
	}
	return p;
}
