/***************************************************************************
                          triangle_or.h  -  description
                             -------------------
    begin                : Wed Sep 13 2000
    copyright            : (C) 2000 by Halva Giovanni & Giacomo
    email                : invehalv@airlab.elet.polimi.it
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef triangle_or_h
#define triangle_or_h 1

#include "fuzzy_set.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif



/**
* This class inherits from fuzzy_set and specifies it to be a triangle open on the right side.
* @short Class implementing the concept of triangular open on right fuzzy set.
*/

class triangle_or : public fuzzy_set  //## Inherits: open_left_triangle_inheritance%382BDE1C0226
{
 public:
  /**
   * Constructor generated by the UML software used to build Brian. It return an empty triangle_or.
   */  
  triangle_or();

  /**
   * Copy constructor generated by the UML software used to build Brian. It return a copy of the fuzzy set passed.
   */
  triangle_or(const triangle_or &right);

  /**	
   * Constructor that build a triangle_or.
   * @param alabel is the label given to the the fuzzy set
   * @param aa is the first point of the fuzzy set
   * @param ab is the second point of the fuzzy set
   */
  triangle_or (char* alabel, float aa, float ab);

  /**	
   * Destructor destroy a fuzzy set and free its data.
   */
  virtual ~triangle_or();

  /**	
   * Return the first point of the fuzzy set.
   * @return the value of the first point
   */
  const float get_a ();
  /**
   * Modify the value of the first point of the fuzzy set.
   * @param value the new value of the point
   */
  void set_a (float value);

  /**	
   * Return the second point of the fuzzy set.
   * @return the value of the first point
   */
  const float get_b ();
  /**
   * Modify the value of the second point of the fuzzy set.
   * @param value the new value of the point
   */
  void set_b (float value);

  /**	
   * Return the membership of data to the fuzzy set.
   * @see fuzzy_set
   * @return the membership value
   * @param avalue is a crisp data
   */
  const float get_membership_value(float avalue);

 private:
  // Data Members for Class Attributes
  
  float a;
  
  float b;
};

// Class triangle_or 

inline const float triangle_or::get_a ()
{
  return a;
}

inline void triangle_or::set_a (float value)
{
  a = value;
}

inline const float triangle_or::get_b ()
{
  return b;
}

inline void triangle_or::set_b (float value)
{
  b = value;
}


#endif
